#include <rccl.h>
#include <signal.h>
#include <math.h>
#include "manex.260.h"

#define FLAG    2

onSignal()
{
        printf ("Column done\n");
}

#ifdef Vxw_TARGET
#define main hex_260_
#endif

main()
{
        TRSF_PTR e, park, pickup, top, down;
        POS_PTR home, reload, center;
        MANIP *mnp;
        JNTS rcclpark;
        char *robotName;

        float hexsize = 15.0;
        int numRows = 5;
        int numCols = 5;
        int i, j;

        float sin_60;

        int mid;

        rcclSetOptions (RCCL_ERROR_EXIT);
        robotName = getDefaultRobot();
        if (!getRobotPosition (rcclpark.v, "rcclpark", robotName))
         { printf ("position 'rcclpark' not defined for robot\n");
           exit(-1);
         }

        e = allocTransXyz ("E", UNDEF, 0.0, 0.0, TOOLZ);
        park = allocTrans ("PARK", UNDEF);
        pickup = allocTrans ("PICKUP", UNDEF);
        top = allocTrans ("TOP", UNDEF);
        down = allocTrans ("DOWN", UNDEF);

        home = makePosition ("home", T6, e, EQ, park, TL, e);
        reload = makePosition ("reload", T6, e, EQ, pickup, TL, e);
        center = makePosition ("center", T6, e, EQ, park, top, down, TL, e);

        mnp = rcclCreate (robotName, 0);
        rcclStart();

        movej (mnp, &rcclpark);
        waitForCompleted (mnp);

        solveTrans (park, home, park, mnp->here);

        rotToTrsf (pickup, zunit, 90.0);
        multTrsf (pickup, pickup, park);
        multTrsfXyz (pickup, 0.0, 0.0, -100.0);
        multTrsfRot (pickup, xunit, 45.0);

        xyzToTrsf (top, 100.0, -50.0, 50.0);
        multTrsfRot (top, xunit, 90.0);

        setJvelScale (mnp, 2.0);
        setCartVel (mnp, 100.0, 50.0);
        signal (SIGUSR1, onSignal);

        sin_60 = sin(60.0*DEGTORAD);

        for (j=0; j<numCols; j++)
         { setConf (mnp, "n");

           distance (mnp, "dz", -30.0);
           move (mnp, reload);
           mid = move (mnp, reload);
           stop (mnp, 2000.0);
           distance (mnp, "dz", -30.0);
           move (mnp, reload);

           waitForStop (mid);
           OPEN_HAND (mnp);
           delay (500.0);
           CLOSE_HAND (mnp);

           setConf (mnp, "f");

           top->p.x -= 2*hexsize*sin_60;
           identTrsf (down);
           for (i=0; i<numRows-1; i++)
            { traceHex (mnp, center, hexsize, 30.0, UNDEF);
              if (i%2 == 0)
               { multTrsfXyz (down, hexsize*sin_60, 3/2.*hexsize, 0.0);
               }
              else
               { multTrsfXyz (down, -hexsize*sin_60, 3/2.*hexsize, 0.0);
               }
            }
           traceHex (mnp, center, hexsize, 30.0, FLAG);
           setStopSig (mnp, SIGUSR1, FLAG);
         }

        setConf (mnp, "n");
        move (mnp, home);
        stop (mnp, 1000.0);
        waitForCompleted (mnp);
        rcclRelease (YES);
}

traceHex (mnp, p, hexsize, approach, flag)
MANIP *mnp;
POS *p;
float hexsize, approach;
int flag;
{
        TRSF offset;
        int saveMod;
        int mid;
        int theta;

        xyzToTrsf (identTrsf(&offset), 0.0, 0.0, -approach);
        setDistance (mnp, &offset);
        move (mnp, p);

        saveMod = getMod(mnp);
        setMod (mnp, 'c');

        offset.p.z = 0.0;
        for (theta=30; theta<=390; theta+=60)
         { offset.p.x = hexsize * cos(DEGTORAD*theta);
           offset.p.y = hexsize * sin(DEGTORAD*theta);
           setDistance (mnp, &offset);
           move (mnp, p);
         }
        xyzToTrsf (identTrsf(&offset), 0.0, 0.0, -approach);
        setDistance (mnp, &offset);
        if (flag != UNDEF)
         { setMotionFlag (mnp, flag);
         }
        mid = move (mnp, p);
        setMod (mnp, saveMod);
        return (mid);
}
