/*-=-=-=-=-=-=-=-=-=-=-=- RCCL/RCI Notice of Copyright -=-=-=-=-=-=-=-=-=-=-*
 *									    *
 *	Copyright (C) 1987, 1991 by John E. Lloyd.			    *
 *									    *
 *	Information on copyright permissions, as well as a complete	    *
 *	disclaimer of warranty, is given in the file COPYRIGHT.		    *
 *									    *
 *-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=*/

#include "../h/cdefs.h"
#include "../h/IOdevice.h"
#include "../h/eppDevice.h"
#include "../h/errorCodes.h"
#include "../h/tval.h"

#include "../drivers/driErrorMsgs.h"

#define CYCLES_PER_MS	1000		/* rough, for a SUN 4 */
#define DEFAULT_TIMEOUT	10*CYCLES_PER_MS

#define LITTLE_ENDIAN_HOST	1

extern unsigned char *rciIOpage;

#define SEND_SIGNAL()	\
	bitSend ^= EPP_SIGOUT; PUT_WORD(eppCfg2,bitSend); bitTest ^= EPP_SIGIN;

static char *format = "stat=0x%hx ibuf=0x%hx test=0x%hx\n";

#define WAIT_FOR_SIGNAL(x) \
	while (((GET_STAT(eppCfg1) & EPP_SIGIN) == bitTest || \
		(GET_STAT(eppCfg1) & EPP_SIGIN) == bitTest) && --k) \
	   ; \
	if (k == 0) \
	 { errorMsg (0, format, GET_STAT(eppCfg1), DO_READ(&epp->ibuf), bitTest); \
	   errorMsg (ETimeout, driTimeoutMsg, fxnName, (x)); \
	   return (-1); \
	 }

static union { unsigned short s;
	unsigned char c[2];
  } swap;

unsigned short foo;

#if !defined(HAS_MAPPED_IOPORTS)  /* assume intel machine and little-endian*/
#if 0
#define DO_READ(ptr) \
	(foo = inw((int)ptr), printf ("RD %x %x\n", ptr, foo), foo)
#define DO_WRITE(ptr,val) outw((int)ptr,val); printf ("WR %x\n", ptr);
#define GET_STAT(ptr) \
	(foo = inw((int)ptr), printf ("ST %x %x\n", ptr, foo), foo)
#define PUT_WORD(ptr,val) outw((int)ptr,val); printf ("PW %x\n", ptr);
#else
#define DO_READ(ptr) inw((int)ptr)
#define DO_WRITE(ptr,val) outw((int)ptr,val)
#define GET_STAT(ptr) inw((int)ptr)
#define PUT_WORD(ptr,val) outw((int)ptr,val)
#endif
#else
#if defined(LITTLE_ENDIAN_HOST)
#define DO_READ(ptr)		*(unsigned short*)(ptr)
#define DO_WRITE(ptr,val)	*(unsigned short*)(ptr) = (val)
#else
#define	DO_READ(ptr) \
      (swap.c[1]= *(unsigned char*)(ptr), \
       swap.c[0]=((unsigned char*)(ptr))[1], swap.s)
#define	DO_WRITE(ptr,val)
      (swap.s=(val),*(unsigned char*)(ptr)=swap.c[1], \
       ((unsigned char*)(ptr))[1]=swap.c[0])
#endif
#define GET_STAT(ptr) 		*(unsigned short*)(ptr)
#define PUT_WORD(ptr,val)	*(unsigned short*)(ptr) = (val)
#endif

waitForSignal (char *fxnName, int x, volatile unsigned short *eppCfg1,
	       unsigned short bitTest, int k)
{
	while (((GET_STAT(eppCfg1) & EPP_SIGIN) == bitTest || \
		(GET_STAT(eppCfg1) & EPP_SIGIN) == bitTest) && --k) \
	   ; \
	if (k == 0) \
	 { errorMsg (0, "stat=0x%hx test=0x%hx\n",GET_STAT(eppCfg1),bitTest);\
	   errorMsg (ETimeout, driTimeoutMsg, fxnName, (x)); \
	   return (-1); \
	 }
	else
	 { return 1;
	 }
}

static int hardwareInit();
static int interruptEnable();
static int probeTest();

eppRead (dev, buf, size)
IO_DEVICE *dev;
void *buf;
int size;
{
	char *fxnName = "eppRead()";
	Epp *epp = (Epp *)dev->location;

	register volatile unsigned short *eppCfg1; /* control/status reg */
	register volatile unsigned short *eppCfg2; /* control reg */
	register volatile unsigned short *eppIbuf; /* parallel port input buf	*/
	register unsigned short *dataPtr; /* pointer to data buffer	*/
	register unsigned short bitTest;  /* used to test for 'ready'	*/
	register unsigned short bitSend;  /* used to set output bit	*/
	register int count;		    /* word count register	*/
	register int i;

	short myChecksum;		/* computes checksum		*/
	short theirChecksum;
	unsigned long k;		/* timeout counter		*/
	int byteCount;			/* number of bytes read in  	*/

	eppIbuf = &(epp->ibuf);
	eppCfg1 = &(epp->cfg1stat);
	eppCfg2 = &(epp->cfg2);

	bitTest = (GET_STAT(eppCfg1) & EPP_SIGIN);	/* get initial csr value */
	bitSend = (GET_STAT(eppCfg1) & EPP_SIGOUT_R) ? EPP_SIGOUT : 0;

	DO_WRITE(&epp->obuf, EPP_READ_READY);

	k = dev->blocktime * CYCLES_PER_MS * 5;
#if 0
	while (DO_READ(&epp->ibuf) != EPP_WRITE_READY && --k)
	   ;
	if (k == 0)
	 { errorMsg (ETimeout, driTimeoutMsg, fxnName, 0);
	   DO_WRITE (&epp->obuf, 0);
	   PUT_WORD (eppCfg2, 0); /* clear SIGOUT */
	   return (-1);
	 }
#endif
	WAIT_FOR_SIGNAL(1);
	DO_WRITE (&epp->obuf, 0);
	count = (int)DO_READ(eppIbuf);
	dev->byteCount = count*sizeof(short);

	/* ----- check for bogus count: negative, or too big ----- */

	if (count < 0 || size < dev->byteCount)
	 { errorMsg (ECount, driCountMsg, fxnName, count);
	   PUT_WORD (eppCfg2, 0); /* clear SIGOUT */
	   return (-1);
	 }

	/* ----- now read in all the words ----- */

	k = (count+2)*dev->timeout;

	dataPtr = (unsigned short *)buf;
	myChecksum = 0;
	i = count;
	while (i--)
         { SEND_SIGNAL();
#if 0
	   WAIT_FOR_SIGNAL(2);
#else
	   if (waitForSignal (fxnName, 2, eppCfg1, bitTest, k) < 0)
	    { int j;
	      errorMsg (0, "left=%d count=%d\n", i, count);
	      for (j=0; j<count-i; j++)
	       { rciPrintf ("%2d: 0x%4x\n", j, ((unsigned short*)buf)[j] );
	       }
	      return -1;
	    }
#endif
	   *dataPtr = DO_READ(eppIbuf);
	   myChecksum += *dataPtr++;
	 }

	/* ----- now get their checksum ----- */

	SEND_SIGNAL();
	WAIT_FOR_SIGNAL(3);

	theirChecksum = DO_READ(eppIbuf);
	
	SEND_SIGNAL();
	WAIT_FOR_SIGNAL(4);

	/* input buffer is now clear */

	if (myChecksum != theirChecksum)
	 { errorMsg (ETransmit, driChecksumMsg, 
				fxnName, theirChecksum, myChecksum);
	   rciPrintf ("eppRead: bad checksum. Count = %d\n", count);
	   for (i=0; i<count; i++)
	    { rciPrintf ("%2d: 0x%4x\n", i, ((unsigned short*)buf)[i] );
	    }
	   return (-1);
	 }

	return (dev->byteCount);
}

eppWrite (dev, buf, size)
IO_DEVICE *dev;
void *buf;
int size;
{
	char *fxnName = "eppWrite()";
	Epp *epp = (Epp *)dev->location;

	register volatile unsigned short *eppCfg1; /* control/status reg */
	register volatile unsigned short *eppCfg2; /* control reg */
	register volatile unsigned short *eppObuf; /* parallel port output buf	*/
	register unsigned short *dataPtr;   /* pointer to data buffer	*/
	register unsigned short bitTest;    /* used to test for 'ready'	*/
	register unsigned short bitSend;    /* used to send output bit	*/
	register int count;		    /* word count register	*/

	short checksum;			    /* computes checksum	*/
	unsigned long k;		    /* timeout counter		*/
	int byteCount;			    /* number of bytes read in  */

	eppObuf = &(epp->obuf);
	eppCfg1 = &(epp->cfg1stat);
	eppCfg2 = &(epp->cfg2);

        count = (size+1)/sizeof(short);
#if 0
	if (size % sizeof(short))
	 { count++;
	 }
#endif
        if (count < 0)
         { errorMsg (ECount, driCountMsg, fxnName, count);
	   DO_WRITE (eppObuf, 0);
	   PUT_WORD (eppCfg2, 0);	/* clear SIGOUT */
	   return (-1);
	 }

	DO_WRITE(&epp->obuf, EPP_WRITE_READY);
	
	k = dev->blocktime * CYCLES_PER_MS;
	while (DO_READ(&epp->ibuf) != EPP_READ_READY && --k)
	   ;
	if (k == 0)
	 { errorMsg (ETimeout, driTimeoutMsg, fxnName, 0);
	   DO_WRITE (eppObuf, 0);
	   PUT_WORD (eppCfg2, 0);	/* clear SIGOUT */
	   return (-1);
	 }
        DO_WRITE(eppObuf,count);
	bitTest = (GET_STAT(eppCfg1) & EPP_SIGIN) ^ EPP_SIGIN;	/* get signal bit */
	bitSend = (GET_STAT(eppCfg1) & EPP_SIGOUT_R) ? EPP_SIGOUT : 0;

	k = (count+2)*dev->timeout;
	SEND_SIGNAL();
	WAIT_FOR_SIGNAL(1);

	dataPtr = (unsigned short *)buf;
	
	/* ----- send words down the chute ----- */

	checksum = 0;
	while (count--)
	 { checksum += *dataPtr;
	   DO_WRITE (eppObuf, *dataPtr++);	/* load the epp */

	   SEND_SIGNAL();
#if 0
	   WAIT_FOR_SIGNAL(2);
#else
	   if (waitForSignal (fxnName, 2, eppCfg1, bitTest, k) < 0)
	    { errorMsg (0, "left=%d count=%d\n", 
			    count, (size+1)/sizeof(short));
	      return -1;
	    }
#endif
         }

	DO_WRITE (eppObuf, checksum);
	SEND_SIGNAL();
	WAIT_FOR_SIGNAL(3);

	DO_WRITE (eppObuf, 0);
	SEND_SIGNAL();

	/* make sure signal bit is clear before we leave routine */

#if 1
	if (GET_STAT(eppCfg1) & EPP_SIGOUT_R)
	 { WAIT_FOR_SIGNAL(4);
	   SEND_SIGNAL();
	 }
#else
	k = 100;
	while (--k)
	   ;
	PUT_WORD (eppCfg2, 0); 	/* clear SIGOUT */
#endif
	/* k = 1000; */
	k = 20000;
	while (--k) {
	  /*	  if ( k == 500 )
	    rciPrintf ("k=%d\n", k);
	  */
	}
	return (size);
}

eppProbe (dev)
IO_DEVICE *dev;
{
	unsigned char *location;
	Epp *epp;

#if HAS_MAPPED_IOPORTS
	location = rciIOpage + dev->address;
#else
	location = (unsigned char *)dev->address;
#endif
	if (!probeTest (location))
	 { return (-1);
	 }
	epp = (Epp *)(location);

	dev->location = location;
	dev->timeout = DEFAULT_TIMEOUT;
	dev->blocktime = EPP_DEFAULT_BLOCKTIME;
	dev->size = sizeof (Epp);

	hardwareInit (dev, epp);

	return (dev->size);
}

eppAttach (dev)
IO_DEVICE *dev;
{
	char *fxnName = "eppAttach()";
	Epp *epp = (Epp *)dev->location;
	unsigned long k;
	unsigned short *obuf;

	if (DO_READ(&epp->ibuf) != EPP_LISTEN)
	 { errorMsg (EConnect, driListeningMsg, fxnName);
	   return (-1);
	 }
	obuf = &epp->obuf;
	DO_WRITE(obuf, EPP_LISTEN);

	PUT_WORD (&epp->cfg2, 0);
	PUT_WORD (&epp->cfg2, EPP_SIGOUT);

	k = dev->timeout*100;
	while (DO_READ (&epp->ibuf) && --k)
	   ;
	if (k == 0)
	 { errorMsg (ETimeout, driTimeoutMsg, fxnName, 1);
	   rciPrintf ("ibuf=0x%x\n", DO_READ (&epp->ibuf));
	   return (-1);
	 }

	PUT_WORD (&epp->cfg2, 0); /* clear SIGOUT */
	DO_WRITE (obuf, 0);

#if 1
	k = dev->timeout*100;
	while ((GET_STAT (&epp->cfg1stat) & EPP_SIGIN) && --k)
	   ;
	if (k == 0)
	 { errorMsg (ETimeout, driTimeoutMsg, fxnName, 1);
	   return (-1);
	 }
#else
	for (k=1000; k>0; k--)	/* XXX Give other device 1.5 msec. to notice */
	   ;
#endif

	return (1);	
}

eppDetach (dev)
IO_DEVICE *dev;
{
	Epp *epp;

	if (dev->location)
	 { epp = (Epp *)dev->location;

	   PUT_WORD (&epp->cfg1stat, (EPP_DIOAEN|EPP_DIOBEN|EPP_PULSE1));
	   DO_WRITE (&epp->obuf, 0);
	   return (1);
	 }
	else
	 { return (0);
	 }
}

eppSignal (dev, n)
IO_DEVICE *dev;
int n;
{
	char *fxnName = "eppSignal()";
	Epp *epp = (Epp *)(rciIOpage + (dev->address));
       	int k;
	int retval = 1;

	unsigned short *ibuf;
	unsigned short *obuf;

	ibuf = &epp->ibuf;
	obuf = &epp->obuf;

	if (n < 1 || n > 2)
	 { errorMsg (ECount, driCountMsg, fxnName, n);
	   return (-1);
	 }
	if (DO_READ(ibuf) == EPP_SIGNAL_ACK)
	 { errorMsg (ESync, driSyncMsg, fxnName);
	   return (-1);
	 }
	DO_WRITE (obuf, n);

	PUT_WORD (&epp->cfg1stat, dev->args[0] | EPP_INTR_SEND); /* send interrupt */
	k = dev->timeout * 50;
	while (DO_READ(ibuf) != EPP_SIGNAL_ACK && --k)
	   ;
	if (k == 0)
	 { errorMsg (ETimeout, driTimeoutMsg, fxnName, 1);
	   retval = -1;
	 }
	PUT_WORD (&epp->cfg1stat, dev->args[0]);	/* clear interrupt */
	DO_WRITE (obuf, 0);

	k = dev->timeout * 100;
	while (DO_READ(ibuf) == EPP_SIGNAL_ACK && --k)
	   ;
	if (k == 0)
	 { errorMsg (ETimeout, driTimeoutMsg, fxnName, 2);
	   return (-1);
	 }

	return (retval);
}

eppIoctl (dev, com, data)
IO_DEVICE *dev;
int com;
void *data;
{
	char *fxnName = "eppIoctl()";
	Epp *epp = (Epp *)dev->location;

	switch (com)
	 { case IOC_SIGINTER_ENABLE:
	    { dev->args[0] |= EPP_INTEN1;
	      PUT_WORD (&epp->cfg1stat, dev->args[0]);
	      break;
	    }
	   case IOC_SIGINTER_DISABLE:
	    { dev->args[0] &= ~EPP_INTEN1;
	      PUT_WORD (&epp->cfg1stat, dev->args[0]);
	      break;
	    }
	   case IOC_TIMEOUT:
	    { dev->timeout = *(int *)data;
	      break;
	    }
	   case IOC_GET_BLOCKTIME:
	    { *(int *)data = dev->blocktime;
	      break;
	    }
	   case IOC_SET_BLOCKTIME:
	    { dev->blocktime = *(int *)data;
	      break;
	    }
	   default:
	    { errorMsg (EBadIoctl, driIoctlMsg, fxnName, com);
	      return (-1);
	    }
	 }
	return (1);
}

eppTestRead (dev)
IO_DEVICE *dev;
{
	unsigned short *ibuf;

	ibuf = &((Epp *)(dev->location))->ibuf;
	if (DO_READ(ibuf) == EPP_WRITE_READY)
	 { return (1);
	 }
 	else
	 { return (0);
	 }
}

eppTestWrite (dev)
IO_DEVICE *dev;
{
	unsigned short *ibuf;

	ibuf = &((Epp *)(dev->location))->ibuf;
	if (DO_READ(ibuf) == EPP_READ_READY)
	 { return (1);
	 }
 	else
	 { return (0);
	 }
}

static hardwareInit (dev, epp)
IO_DEVICE *dev;
Epp *epp;
{
	PUT_WORD (&epp->cfg1stat, EPP_LRESET1);
	PUT_WORD (&epp->cfg2, EPP_LRESET2);
	PUT_WORD (&epp->cfg3, 0);
	PUT_WORD (&epp->cfg1stat, 0);
	PUT_WORD (&epp->cfg2, 0);
/*	epp->cfg4 = 1; */
	PUT_WORD (&epp->dmaclr1, 0);
	PUT_WORD (&epp->dmaclr2, 0);
	PUT_WORD (&epp->cntintclr, 0);

	/* Now set the directions */

	PUT_WORD (&epp->cfg3, (EPP_WRITEC|EPP_WRITED));

	/* use dev->args[0] to shadow cfg2 */
	dev->args[0] = (EPP_DIOAEN|EPP_DIOBEN|EPP_PULSE1);
	PUT_WORD (&epp->cfg1stat, dev->args[0]);
}

static probeTest (loc)
unsigned char *loc;
{
	Epp *epp;

#if HAS_MAPPED_IOPORTS
	/* first, make sure the thing is there */
	if (rciBadAddr (loc, sizeof(short)))
	 { return (0);
	 }
#endif
	epp = (Epp*)loc;

	/* reset the stupid thing */
	PUT_WORD (&epp->cfg1stat, EPP_LRESET1);
	PUT_WORD (&epp->cfg2, EPP_LRESET2);
	PUT_WORD (&epp->cfg3, 0);
	PUT_WORD (&epp->cfg1stat, 0);
	PUT_WORD (&epp->cfg2, 0);

	/* write EPP_SETACK2 and see if it shows up as EPP_ACK2 */
	PUT_WORD (&epp->cfg2, EPP_SETACK2);
	if (!(GET_STAT (&epp->cfg1stat) & EPP_ACK2))
	 { return (0);
	 }
	PUT_WORD (&epp->cfg2, 0);
	if (GET_STAT (&epp->cfg1stat) & EPP_ACK2)
	 { return (0);
	 }
	return (1);
}
