/* 
 * pid.h -- 
 * 
 * Stewart Kingdon,
 * Copyright (c) 1998,
 * Laboratory for Computational Intelligence,
 * University of British Columbia.
 */

/*
 * $Author: kingdon $
 * $Date: 1998/12/18 19:56:30 $
 * $Revision: 1.2 $
 *
 * $Log: pid.h,v $
 * Revision 1.2  1998/12/18 19:56:30  kingdon
 * [1] added support for all 8 channels
 * [2] added interpolation between set points
 * [3] added control rate
 *
 * Revision 1.1  1998/10/31 01:47:48  kingdon
 * Initial revision
 *
 */

/* $Id: pid.h,v 1.2 1998/12/18 19:56:30 kingdon Exp kingdon $ */

#define PID_MAGIC	0xCAFEBABE
#define PID_VERSION	0x00000003
#define PID_NCHAN	8

typedef struct
{
      int       mode;		    /* 0x250n0: 0 = stop */
      int       goal;		    /* 0x250n1: current set point */
      int       posn;		    /* 0x250n2: current position */
      int       kp;		    /* 0x250n3: P term of pid controller */
      int       kd;		    /* 0x250n4: D term of pid controller */
      int       ki;		    /* 0x250n5: I term of pid controller */
      int       imax;		    /* 0x250n6: maximum integration value */
      int       ival;		    /* 0x250n7: current integration value */
      int       max_dgoal;	    /* 0x250n8: max posn error */
      unsigned  reserved[7];	    /* 0x250n9-0x250nF: unused */
} PID;

typedef struct
{
      unsigned  magic;		    /* 0x25000: should be 0xCAFEBABE */
      unsigned  version;	    /* 0x25001: should be 0x00000003 */
      unsigned  nchan;		    /* 0x25002: should be 8 */
      unsigned  interpolate;	    /* 0x25003: number ticks for interp. */
      unsigned  ctrl_rate;	    /* 0x25004: controller rate (in Hz) */
      unsigned  ctrl_count;	    /* 0x25005: incremented by one/cycle */
      unsigned  reserved[2+8];	    /* 0x25006-0x2500F: unused */
      PID       channel[PID_NCHAN]; /* 0x25010-0x2508F: motor channels */
} PIDHEADER;

#endif
