#ifndef _PMC_PID_H_WRAP_
#define _PMC_PID_H_WRAP_



/* 
 * pid.h -- 
 * 
 * Stewart Kingdon,
 * Copyright (c) 1998,
 * Laboratory for Computational Intelligence,
 * University of British Columbia.
 */

/*
 * $Author: kingdon $
 * $Date: 1998/10/31 01:47:48 $
 * $Revision: 1.1 $
 *
 * $Log: pid.h,v $
 * Revision 1.1  1998/10/31 01:47:48  kingdon
 * Initial revision
 *
 */

/* $Id: pid.h,v 1.1 1998/10/31 01:47:48 kingdon Exp kingdon $ */

#define PID_SERVO_ENABLE	0x1

#define PID_MAGIC	0xCAFEBABE
#define PID_VERSION	0x00000003
#define PID_NCHAN	8

typedef struct
{
      int       mode;		    /* 0x250n0: 0 = stop */
      int       goal;		    /* 0x250n1: current set point */
      int       posn;		    /* 0x250n2: current position */
      int       kp;		    /* 0x250n3: P term of pid controller */
      int       kd;		    /* 0x250n4: D term of pid controller */
      int       ki;		    /* 0x250n5: I term of pid controller */
      int       imax;		    /* 0x250n6: maximum integration value */
      int       ival;		    /* 0x250n7: current integration value */
      int       max_dgoal;	    /* 0x250n8: max posn error */
      unsigned  reserved[7];	    /* 0x250n9-0x250nF: unused */
} PID;

typedef struct
{
      unsigned  magic;		    /* 0x25000: should be 0xCAFEBABE */
      unsigned  version;	    /* 0x25001: should be 0x00000003 */
      unsigned  nchan;		    /* 0x25002: should be 8 */
      unsigned  interpolate;	    /* 0x25003: number ticks for interp. */
      unsigned  ctrl_rate;	    /* 0x25004: controller rate (in Hz) */
      unsigned  ctrl_count;	    /* 0x25005: incremented by one/cycle */
      unsigned  reserved[2+8];	    /* 0x25006-0x2500F: unused */
      PID       channel[PID_NCHAN]; /* 0x25010-0x2508F: motor channels */
} PIDHEADER;

extern PID rciPmcPid;
extern PIDHEADER rciPmcPidHeader;

#define PID_MEM_BASE	0x25000

#define PID_HEADER_OFF(f) \
	(int)((int*)&rciPmcPidHeader.f-(int*)&rciPmcPidHeader)

#define PID_CHAN_OFF(n,f) \
	((int)((int*)&rciPmcPidHeader.channel[n]-(int*)&rciPmcPidHeader) + \
	 (int)((int*)&rciPmcPid.f-(int*)&rciPmcPid))

#define PID_CHAN_BASE(n) \
	(PID_MEM_BASE + \
	((int)((int*)&rciPmcPidHeader.channel[n]-(int*)&rciPmcPidHeader)))
	
#define PID_MAGIC_ADDR          (PID_MEM_BASE+PID_HEADER_OFF(magic))
#define PID_VERSION_ADDR        (PID_MEM_BASE+PID_HEADER_OFF(version))
#define PID_NCHAN_ADDR          (PID_MEM_BASE+PID_HEADER_OFF(nchan))
#define PID_INTERPOLATE_ADDR	(PID_MEM_BASE+PID_HEADER_OFF(interpolate))
#define PID_CTRL_RATE_ADDR	(PID_MEM_BASE+PID_HEADER_OFF(ctrl_rate))

#define PID_MODE_ADDR(n)	(PID_MEM_BASE+PID_CHAN_OFF(n,mode))
#define PID_GOAL_ADDR(n)	(PID_MEM_BASE+PID_CHAN_OFF(n,goal))
#define PID_POSN_ADDR(n)	(PID_MEM_BASE+PID_CHAN_OFF(n,posn))
#define PID_KP_ADDR(n)		(PID_MEM_BASE+PID_CHAN_OFF(n,kp))
#define PID_KD_ADDR(n)		(PID_MEM_BASE+PID_CHAN_OFF(n,kd))
#define PID_KI_ADDR(n)		(PID_MEM_BASE+PID_CHAN_OFF(n,ki))
#define PID_IMAX_ADDR(n)	(PID_MEM_BASE+PID_CHAN_OFF(n,imax))
#define PID_IVAL_ADDR(n)	(PID_MEM_BASE+PID_CHAN_OFF(n,ival))

#endif
