/* Program to peek and poke some of the registers of the S16D */

#ifdef X86sol_TARGET
#include "../h/cdefs.h"
#include <stdio.h>
#include "../h/ctree.h"

#include <unistd.h>
#include <sys/types.h>
#include <sys/mman.h>
#include <fcntl.h>

#include "s16dreg.h"
#include "s16dioctl.h"
#include "rciioctl.h"

char **argvec;
char key_buf [256];

S16dRegs *s16d;
int s16dFd;

void setConfig (u_short v)
{
	if (ioctl (s16dFd, S16D_SET_CONFIG, &v) < 0)
	 { perror ("ioctl S16D_SET_CONFIG");
	   exit (-1);
	 }
}

u_short getConfig ()
{
	u_short ushort;

	if (ioctl (s16dFd, S16D_GET_CONFIG, &ushort) < 0)
	 { perror ("ioctl S16D_GET_CONFIG");
	   exit (-1);
	 }
	return (ushort);
}

u_short getCmd ()
{
	u_short ushort;

	if (ioctl (s16dFd, S16D_GET_COMMAND, &ushort) < 0)
	 { perror ("ioctl S16D_GET_COMMAND");
	   exit (-1);
	 }
	return (ushort);
}

main(int argc, char **argv)
{
	char *devname = NULL;
	char *keytree;
	u_short ushort;
	int value;

	/* Variables used to map the device registers into memory */

	unsigned long regAddr = 0;
	int pageSize;
	off_t pageStart;
	caddr_t vpageStart;
	u_long pageOffset;

	argvec = argv;
	argc--; argv++;
	while (argc > 0)
	 { if (ARGUMENT("-x"))
	    {
	    }
	   else
	    { devname = *argv;
	    }
	   argc--; argv++;
	   continue;
usage:
	   fprintf (stderr, 
"Usage: %s <s16dDeviceName> \n", argvec[0]);
	   exit (1);
	 }

	if (devname == NULL)
	 { devname = "/dev/RCIs16dmem0";
	 }

	if ((s16dFd = open (devname, O_RDWR)) < 0)
	 { perror ("open");
	   exit (1);
	 }

	/* map in the device registers */

#ifdef __STDC__
    	pageSize = sysconf(_SC_PAGESIZE);
#else
    	pageSize = getpagesize();
#endif
	pageStart = (regAddr / pageSize) * pageSize;
	pageOffset = regAddr % pageSize;
	vpageStart = mmap((caddr_t)0, sizeof(S16dRegs), PROT_READ|PROT_WRITE,
			   MAP_SHARED, s16dFd, pageStart);
	if (vpageStart == (caddr_t)(-1))
	 { perror ("mmap");
	   exit (1);
	 }

	/* map has been obtained. Save it. */

	s16d = (S16dRegs*)vpageStart + pageOffset;

	keytree = tree_match_parse("                                    \
(       quit                                                            \
	show								\
	peek (status data config cmd)					\
	poke ([cmd data config] (%x", &value, "))			\
)",0);
	if (keytree == 0)
	 { printf ("ERROR/Parse error at %.30s\n", tree_match_err_at());
	   exit (-1);
	 }
	do
 	 { tree_match (keytree, "PROMPT> ", key_buf);

	   if (COMMAND("show"))
	    { printf ("cmd:    0x%04hx\n", getCmd());
	      printf ("status: 0x%04hx\n", s16d->csr);
	      printf ("config: 0x%04hx\n", getConfig());
	      printf ("data:   0x%04hx\n", s16d->data);
	    }
	   else if (COMMAND ("peek cmd"))
	    { printf ("cmd:    0x%04hx\n", getCmd());
	    }
	   else if (COMMAND ("peek status"))
	    { printf ("status: 0x%04hx\n", s16d->csr);
	    }
	   else if (COMMAND ("peek data"))
	    { printf ("data:   0x%04hx\n", s16d->data);
	    }
	   else if (COMMAND ("peek config"))
	    { printf ("config: 0x%04hx\n", getConfig());
	    }
	   else if (COMMAND ("poke cmd"))
	    { s16d->csr = value;
	    }
	   else if (COMMAND ("poke data"))
	    { s16d->data = value;
	    }
	   else if (COMMAND ("poke config"))
	    { setConfig (value);
	    }
	 }
	while (!(COMMAND("quit")));
}
#endif

