/* mc8prim.h - include file for MC8-DSP-ISA communication primitives */

#ifndef MC8PRIM_H
#define MC8PRIM_H

#include "mc8const.h"

/*************/
/* functions */
/*************/
mc8_struct *mc8Open(long lAddr, int wIrq, int wForce);
int mc8Close(mc8_struct *mc8Info);

int mc8ReadReg(mc8_struct *mc8Info, int wReg, short *wData);
int mc8ReadRegs(mc8_struct *mc8Info, int wReg, short *rgwData, int cRegs);
int mc8ReadCtrl(mc8_struct *mc8Info, short *wCtrl);
int mc8WriteReg(mc8_struct *mc8Info, int wReg, short wData);
int mc8WriteRegs(mc8_struct *mc8Info, int wReg, short *rgwData, int cRegs);
int mc8WriteCtrl(mc8_struct *mc8Info, short wCtrl);

int mc8ReadDSPReg(mc8_struct *mc8Info, int wReg, long *lData);
int mc8WriteDSPReg(mc8_struct *mc8Info, int wReg, long lData);

/* Memory access: If accessed data is 3 column, (lAddr & MEMADDR) != 0 */
int mc8ReadMem(mc8_struct *mc8Info, long lAddr, short *rgwData, int cWords);
int mc8WriteMem(mc8_struct *mc8Info, long lAddr, short *rgwData, int cWords);
/* Fixed PC word length (longs) - others later? */
int mc8ReadMemL(mc8_struct *pmc8Info, long lAddr, short *rgsData, int cWords);
int mc8WriteMemL(mc8_struct *pmc8Info, long lAddr, short *rgsData, int cWords);

/* DSP Control functions */
int mc8DSPReset(mc8_struct *mc8Info);
int mc8DSPHardReset(mc8_struct *mc8Info);
int mc8DSPStart(mc8_struct *mc8Info);
int mc8DSPInit(mc8_struct *mc8Info);

/*************/
/* macros    */
/*************/
#define mc8IsExtAddr(lAddr) (lAddr >= 0x400000L)

/*************/
/* registers */
/*************/
#define MC8_PIT			0x00
#define MC8_PITL		0x00
#define MC8_PITH		0x01
#define MC8_WD			0x02
#define MC8_GATE		0x04
#define MC8_WDCTRL		0x06
#define MC8_SYNCCTRL	0x06

#define MC8_DAC0		0x08
#define MC8_DAC1		0x09
#define MC8_DAC2		0x0a
#define MC8_DAC3		0x0b
#define MC8_DAC4		0x0c
#define MC8_DAC5		0x0d
#define MC8_DAC6		0x0e
#define MC8_DAC7		0x0f
#define MC8_DAC8		0x10
#define MC8_DAC9		0x11
#define MC8_DAC10		0x12
#define MC8_DAC11		0x13
#define MC8_DAC12		0x14
#define MC8_DAC13		0x15
#define MC8_DAC14		0x16
#define MC8_DAC15		0x17

#define MC8_ADC0		0x18
#define MC8_ADC1		0x19
#define MC8_ADC2		0x1a
#define MC8_ADC3		0x1b
#define MC8_ADC4		0x1c
#define MC8_ADC5		0x1d
#define MC8_ADC6		0x1e
#define MC8_ADC7		0x1f
                            
#define MC8_DIGIN0		0x20
#define MC8_DIGIN0L		0x20
#define MC8_DIGIN0H		0x21
#define MC8_DIGIN1		0x22
#define MC8_DIGIN1L		0x22
#define MC8_DIGIN1H		0x23
#define MC8_DIGIN2		0x24
#define MC8_DIGIN2L		0x24
#define MC8_DIGIN2H		0x25
#define MC8_DIGOUT0		0x26
#define MC8_DIGOUT0L	0x26
#define MC8_DIGOUT0H	0x27

#define MC8_DMASK0		0x28
#define MC8_DMASK0L		0x28
#define MC8_DMASK0H		0x29
#define MC8_DMASK1		0x2a
#define MC8_DMASK1L		0x2a
#define MC8_DMASK1H		0x2b
#define MC8_DMASK2		0x2c
#define MC8_DMASK3		0x2d
#define MC8_DACT		0x2e
#define MC8_DACTL		0x2e
#define MC8_DACTH		0x2f

#define MC8_DMFLG		0x30
#define MC8_OEDIG		0x32
#define MC8_INTFLGHOST	0x34
#define	MC8_INTENAHOST	0x35
#define MC8_INTFLGDSP	0x36
#define MC8_INTENADSP	0x37

#define MC8_DSPRESET	0x38
#define MC8_SOFTSYNC	0x3e

/* Quadrature Decoders */
#define MC8_QDCOUNT		0x40
#define MC8_QDCOUNT0	0x40
#define MC8_QDCOUNT1	0x42
#define MC8_QDCOUNT2	0x44
#define MC8_QDCOUNT3	0x46
#define MC8_QDCOUNT4	0x48
#define MC8_QDCOUNT5	0x4a
#define MC8_QDCOUNT6	0x4c
#define MC8_QDCOUNT7	0x4e

#define MC8_QDMODE		0x50
#define MC8_QDMODE0		0x50
#define MC8_QDMODE1		0x52
#define MC8_QDMODE2		0x54
#define MC8_QDMODE3		0x56
#define MC8_QDMODE4		0x58
#define MC8_QDMODE5		0x5a
#define MC8_QDMODE6		0x5c
#define MC8_QDMODE7		0x5e

/* Other useful values */
#define MC8_AUTOINCR	0x80
#define MC8_DSPREG		0x100
#define MC8_ADCBUSY		0x8000
	/* Beginning of internal memory */
#define MC8_INTADDR		0x20000L
	/* Beginning of external memory */
#define MC8_EXTADDR		0x400000L


/* Excerpt from def21060.h - SHARC IOP registers */
#ifndef SYSCON
#define SYSCON			0x00
#define VIRPT			0x01
#define WAIT			0x02
#define SYSTAT			0x03
#define EPB0			0x04
#define EPB1			0x05
#define EPB2			0x06
#define EPB3			0x07
#define DMAC6			0x1c
#define DMAC7			0x1d
#define DMAC8			0x1e
#define DMAC9			0x1f

#define DMASTAT			0x37

#define II6				0x40
#define IM6				0x41
#define C6				0x42
#define CP6				0x43
#define GP6				0x44

#define II7				0x48
#define IM7				0x49
#define C7				0x4a
#define CP7				0x4b
#define GP7				0x4c
#define EI7				0x4d
#define EM7				0x4e
#define EC7				0x4f

#endif /* SYSCON */

/* Some useful DMAC values */
#define MC8DMA_ENA		0x0001
#define MC8DMA_IW32		0x0040
#define MC8DMA_IR32		0x0044
#define MC8DMA_IW48		0x00a0
#define MC8DMA_IR48		0x00a4
/* External read/write reverses sense of TRAN */
#define MC8DMA_EXTR		0x0220
#define MC8DMA_EXTW		0x0224
/* Flushes existing accesses and disables DMA */
/* *MUST* do this before setting new params   */
#define MC8DMA_FLUSH	0x2000

/* Masks for DMAC status bits */
/* Both should be 0 if access is finished */
#define MC8DMA_FSMSK	0xc000
#define MC8DMA_PSMSK	0x0018

#endif /* MC8PRIM_H */
