#ifdef HAS_X11_MOTIF
/* Support for running the Ctree matcher within an X11 Motif Text widget. */

#include <stdio.h>
#include <unistd.h>
#include <fcntl.h>
#include <Xm/Text.h>

#include "../h/ctree.h"
#include "../h/ctreeMotif.h"

#define CTM_COBUFSIZE	256
#define CTM_HISTORY	 12

typedef struct {
	Widget w;
	int rows;
	int cols;
	int ibuf;
	int ibufSet;
	unsigned char obuf[CTM_COBUFSIZE+1];
	int ocnt;
	int bellSet;
	int rowNum;
	int history;
	void (*fxn) PROTO((char*,void*,FILE*));
	void *arg;
	char *prompt;
	char *tree;
	char *keybuf;
	int inputFds[2];
	int outputFds[2];
	FILE *outputFp;
	FILE *inputFp;
} XmTextCtreeBuf;

static XmTextCtreeBuf *ctmData;

static void ctmDataInit (XmTextCtreeBuf *cbuf, Widget w)
{
	short rows, cols;

	cbuf->w = w;
	XtVaGetValues (w, XmNrows, &rows, XmNcolumns, &cols, NULL);
	cbuf->rows = rows;
	cbuf->cols = cols;
	cbuf->ibufSet = 0;
	cbuf->bellSet = 0;
	cbuf->ocnt = 0;
	cbuf->rowNum = 1;
	cbuf->history = CTM_HISTORY;
}

static int ctmGetChar (XmTextCtreeBuf *cbuf)
{
	if (cbuf->ibufSet)
	 { cbuf->ibufSet = 0;
	   return (cbuf->ibuf);
	 }
	else
	 { return (EOF);
	 }
}

static int ctmFlush (XmTextCtreeBuf *cbuf)
{
	char *str, *cp;

	if (cbuf->bellSet)
	 { XBell (XtDisplay(cbuf->w), 100);
	   cbuf->bellSet = 0;
	 }
	if (cbuf->ocnt != 0)
	 { str = XmTextGetString (cbuf->w);
	   if (cbuf->ocnt > 0)
	    { cbuf->obuf[cbuf->ocnt] = '\0';
	      XmTextInsert (cbuf->w, strlen(str), (char*)cbuf->obuf);
	      
	      if (cbuf->obuf[cbuf->ocnt-1] == '\n')
	       { if (cbuf->rowNum < cbuf->history)
		  { cbuf->rowNum++;
		  }
		 else
		  { for (cp=str; *cp != '\n'; cp++)
		       ;
		    XmTextReplace (cbuf->w, 0, cp-str+1, "");
		  }
	       }
	      XtFree (str);	
	    }
	   else
	    { /* put cp at the farthest point we can delete back to */
	      for (cp = str+strlen(str)-1; *cp != '\n' && cp != str-1; cp--)
	       { if (cbuf->ocnt++ == 0)
		  { break;
		  }
	       }
	      cp++;
	      if (cp != str+strlen(str))
	       { XmTextReplace (cbuf->w, cp-str, strlen(str)+1, "");
	       }
	    }
	   cbuf->ocnt = 0;
	 }
	return (1);
}

static int ctmPutChar (XmTextCtreeBuf *cbuf, int c)
{
	if (cbuf->ocnt == CTM_COBUFSIZE)
	 { ctmFlush (cbuf);
	 }
	switch (c)
	 { case '\b':
	   case '\0177':
	    { cbuf->ocnt--;
	      break;
	    }
	   case '\07':
	    { cbuf->bellSet = 1;
	      break;
	    }
	   case '\r':
	    { c = '\n';
	      /* no break */
	    }
	   default:
	    { cbuf->obuf[cbuf->ocnt++] = c;
	      break;
	    }
	 }
	if (c == '\n')
	 { ctmFlush (cbuf);
	 }
	return (1);
}

doflush ()
{ 
	char buf[80];
	int i, n;

	fflush (ctmData->outputFp);
	do
	 { n = read (ctmData->outputFds[0], buf, 80);
	   for (i=0; i<n; i++)
	    { ctmPutChar (ctmData, buf[i]);
	    }
	 }
	while (n != 0);
	ctmFlush (ctmData);
}

static void ctmConfig(Widget w, XEvent *event, String *strs, Cardinal *num)
{
	short rows, cols;
	XtVaGetValues (w,
	   XmNrows, &rows,
	   XmNcolumns, &cols,
	   NULL);
	ctmData->rows = rows;
	ctmData->cols = cols;
}

static void ctmKeyhit(Widget w, XEvent *event, String *strs, Cardinal *num)
{
	char buf[16];
	KeySym sym;
	XComposeStatus stat;
	int n;

	n = XLookupString (&event->xkey, buf, 16, &sym, &stat);
	buf[n] = '\0';
	if (n == 1)
	 { switch (buf[0])
	    { case '\r':
	       { buf[0] = '\n';
		 break;
	       }
	      default:
	       { break;
	       }
	    }
	   write (ctmData->inputFds[1], buf, 1);
	   if (tree_match_nb (ctmData->tree, ctmData->prompt,
			      ctmData->keybuf, (int*)0))
	    { doflush();
	      (*ctmData->fxn)(ctmData->keybuf, ctmData->arg,
			      ctmData->outputFp);
	    }
	   /* call C-tree matcher again just to print prompt */
	   if (tree_match_nb (ctmData->tree, ctmData->prompt,
			      ctmData->keybuf, (int*)0))
	    { fprintf (stderr, "Error! False positive from tree_match_nb\n");
	      exit (1);
	    }
	 }
}

static int getCharFxn(fp)
FILE *fp;
{
	if (fp == stdin)
	 { return (ctmGetChar (ctmData));
	 }
	else
	 { return (getc (fp));
	 }
}

static int putCharFxn(c)
int c;
{
	return (ctmPutChar (ctmData, c));
}

static int flushFxn()
{
	return (ctmFlush (ctmData));
}


Widget tree_match_motif_widget (name, parent, app)
char *name;
Widget parent;
XtAppContext app;
{
	static XtActionsRec actionTab[] = {
		{"ctm-keyhit", ctmKeyhit},
		{"ctm-config", ctmConfig},
	};

	static char transSpec[] = 
		"<Key>: ctm-keyhit()\n\
		 <Configure>: ctm-config()";

	XtTranslations transTab;
	Widget w;

	transTab = XtParseTranslationTable(transSpec);
	XtAppAddActions (app, actionTab, XtNumber(actionTab));

	w = XtVaCreateManagedWidget (name, xmTextWidgetClass, parent,
	   XmNeditMode, XmMULTI_LINE_EDIT,
	   XmNcolumns, 80,
	   XmNrows, 10, 
	   XmNtranslations, transTab,
	   XmNwordWrap, True,
	   NULL);
	ctmData = (XmTextCtreeBuf *)malloc(sizeof(XmTextCtreeBuf));
	ctmDataInit (ctmData, w);

	if (pipe (ctmData->inputFds) < 0)
	 { free (ctmData);
	   return (NULL);
	 }
	if (pipe (ctmData->outputFds) < 0)
	 { close (ctmData->inputFds[0]);
	   close (ctmData->inputFds[1]);
	   free (ctmData);
	   return (NULL);
	 }
	ctmData->inputFp = fdopen (ctmData->inputFds[0], "r");
	ctmData->outputFp = fdopen (ctmData->outputFds[1], "w");
#if !defined(FNDELAY) && defined(O_NONBLOCK)
	fcntl (ctmData->outputFds[0], F_SETFL, O_NONBLOCK);
#else
	fcntl (ctmData->outputFds[0], F_SETFL, FNDELAY);
#endif
	tree_match_set_input_stream (ctmData->inputFp);
	tree_match_set_output_stream (ctmData->outputFp);
	tree_match_set_flush_routine (doflush);
	tree_match_set_flags (CTREE_F_NOTTY);

	return (w);
}

void tree_match_widget_init (tree, prompt, keybuf, fxn, arg)
char *tree;
char *prompt;
char *keybuf;
void (*fxn)();
void *arg;
{
	ctmData->fxn = fxn;
	ctmData->arg = arg;
	ctmData->prompt = prompt;
	ctmData->keybuf = keybuf;
	ctmData->tree = tree;

	/* call C-tree matcher just to print prompt */
	if (tree_match_nb (ctmData->tree, ctmData->prompt,
	 		   ctmData->keybuf, (int*)0))
	 { fprintf (stderr, "Error! False positive from tree_match_nb\n");
	   exit (1);
	 }
}

#endif
