#ifdef HAS_X11_MOTIF
#include <stdio.h>

#define ATHENA	0

#if ATHENA
#include <X11/StringDefs.h>
#include <X11/Intrinsic.h>
#include <X11/Xaw/Text.h>
#include <X11/Xaw/AsciiText.h>
#else
#include <Xm/Text.h>
#endif

char *prompt = "PROMPT> ";
int myedit = 0;
int cursorPos;


#define XTC_COBUFSIZE	256
#define XTC_HISTORY	 12

typedef struct {
	Widget w;
	int rows;
	int cols;
	int ibuf;
	int ibufSet;
	unsigned char obuf[XTC_COBUFSIZE+1];
	int ocnt;
	int bellSet;
	int rowNum;
	int history;
} XmTextCtreeBuf;

XmTextCtreeBuf *wbuf;

void textCtreeBufInit (XmTextCtreeBuf *cbuf, Widget w)
{
	short rows, cols;

	cbuf->w = w;
	XtVaGetValues (w, XmNrows, &rows, XmNcolumns, &cols, NULL);
	cbuf->rows = rows;
	cbuf->cols = cols;
	cbuf->ibufSet = 0;
	cbuf->bellSet = 0;
	cbuf->ocnt = 0;
	cbuf->rowNum = 1;
	cbuf->history = XTC_HISTORY;
}

int textCtreeGetChar (XmTextCtreeBuf *cbuf)
{
	if (cbuf->ibufSet)
	 { cbuf->ibufSet = 0;
	   return (cbuf->ibuf);
	 }
	else
	 { return (EOF);
	 }
}

void textCtreePutChar (XmTextCtreeBuf *cbuf, int c)
{
	if (cbuf->ocnt == XTC_COBUFSIZE)
	 { textCtreeFlush (cbuf);
	 }
	switch (c)
	 { case '\b':
	   case '\0177':
	    { if (cbuf->ocnt > 0)
	       { cbuf->ocnt--;
	       }
	      break;
	    }
	   case '\07':
	    { cbuf->bellSet = 1;
	      break;
	    }
	   case '\r':
	    { c = '\n';
	      /* no break */
	    }
	   default:
	    { cbuf->obuf[cbuf->ocnt++] = c;
	      break;
	    }
	 }
	if (c == '\n')
	 { textCtreeFlush (cbuf);
	 }
}

textCtreeFlush (XmTextCtreeBuf *cbuf)
{
	char *str, *cp;

	if (cbuf->bellSet)
	 { XBell (XtDisplay(cbuf->w), 100);
	   cbuf->bellSet = 0;
	 }
	str = XmTextGetString (cbuf->w);
	cbuf->obuf[cbuf->ocnt] = '\0';
	XmTextInsert (cbuf->w, strlen(str), cbuf->obuf);
	
	if (cbuf->obuf[cbuf->ocnt-1] == '\n')
	 { if (cbuf->rowNum < cbuf->history)
	    { cbuf->rowNum++;
	    }
	   else
	    { for (cp=str; *cp != '\n'; cp++)
		 ;
	      XmTextReplace (cbuf->w, 0, cp-str+1, "");
	    }
	 }
	XtFree (str);	
	cbuf->ocnt = 0;
}

void myConfig(Widget w, XEvent *event, String *strs, Cardinal *num)
{
	short rows, cols;
	XtVaGetValues (w,
	   XmNrows, &rows,
	   XmNcolumns, &cols,
	   NULL);
	printf ("%d %d\n", rows, cols);
}

void myInsert(Widget w, XEvent *event, String *strs, Cardinal *num)
{
	char buf[16];
	KeySym sym;
	XComposeStatus stat;
	int n;

	n = XLookupString (&event->xkey, buf, 16, &sym, &stat);
	buf[n] = '\0';
	if (n == 1)
	 { textCtreePutChar (wbuf, buf[0]);
	 }
}

void mycallback(Widget w, XtPointer clientData, XtPointer callData)
{
	char *str;
	XmTextVerifyPtr info = (XmTextVerifyPtr)callData;
	int *myeditp = (int*)clientData;
	int i;

	str = XmTextGetString(w);
	printf ("len=%d\n", strlen(str));
	for (i=0; i<strlen(str); i++)
	 { putchar (str[i]);
	   putchar (',');
	 }
	putchar ('\n');
	info->doit = TRUE;
}

#if ATHENA == 0
static XtCallbackRec modifyVerifyCallback[] = {
  { mycallback, (caddr_t)&myedit },
  { NULL, NULL}
};
#endif

static XtActionsRec actionTable[] = {
	{"my-insert", myInsert},
	{"my-config", myConfig},
};

static char transSpec[] = 
	"<Key>: my-insert()\n\
	 <Configure>: my-config()";

XtTranslations transTab;



main(argc, argv)
int argc;
char **argv;
{
	Widget toplevel;
	Widget textw;
	XtAppContext app;

	toplevel = XtVaAppInitialize (&app, "Demos", 
	   NULL, 0, &argc, argv, NULL, NULL);

	transTab = XtParseTranslationTable(transSpec);
	XtAppAddActions (app, actionTable, XtNumber(actionTable));
#if ATHENA
	XtVaCreateManagedWidget ("text.", textWidgetClass, toplevel,
/*
	   XtNstring, prompt,
	   XtNlength, 20,
	   XtNdisplayPosition, strlen(prompt),
*/
	   NULL);
#else
	cursorPos = strlen(prompt);	
	textw = XtVaCreateManagedWidget ("text.", xmTextWidgetClass, toplevel,
	   XmNvalue, prompt, 
	   XmNeditMode, XmMULTI_LINE_EDIT,
	   XmNcursorPosition, cursorPos,
	   XmNcolumns, 20, 
	   XmNrows, 10, 
	   XmNtranslations, transTab,
	   NULL);
	wbuf = (XmTextCtreeBuf *)malloc(sizeof(XmTextCtreeBuf));
	textCtreeBufInit (wbuf, textw);
#if 0
	XtAddCallback (textw, XmNmodifyVerifyCallback,  
	   mycallback, (caddr_t)&myedit);
#endif
	XmTextSetMaxLength (textw, 20);
	printf ("0x%x\n", &myedit);
#endif
	XtRealizeWidget(toplevel);
	XtAppMainLoop(app);
}
#endif

