#include <math.h>

#include "../h/stringDefs.h"

#include "../h/cdefs.h"
#include "../h/robotMath.h"
#include "../h/rciTask.h"
#include "../h/rciMonitor.h"
#include "../h/rciMacros.h"
#include "../h/robotControl.h"
#include "../h/robotMath.h"
#include "../h/IOdevice.h"

#include "playback.h"

tgenTrigger (rbt)
RCI_RBT *rbt;
{
	char *fxnName = "tgenTrigger() -- ";

	if ((*rbt->wakeup)(rbt->wakeupArg) < 0)
	 { rciAbort (0, "%s Can't signal robot\n", fxnName);
	 }
}

tgenStart (td)
RCI_DESC *td;
{
	TgenBuf *tbuf;
	RCI_RBT *rbt;
	int j, numJ;

	char *pfx = "tgenStart() -- ";

	if ((tbuf = (TgenBuf*)getSharedMemory ("tbuf")) == NULL)
	 { rciAbort (0, "%s can't get shared memory\n", pfx);
	   return (-1);
	 }
	td->tcb->sharedMem = (void*)tbuf;

	/* ---- Robot interface code ---- */

	if ((rbt = td->rbts[0]) == NULL)
	 { rciAbort (0, "%s no robot defined for task\n", pfx);
	   return (-1);
	 }
	if ((*rbt->startup)(td, rbt) < 0)
	 { return (-1);
	 }
	switch (td->tcb->scheduleDisp)
	 { 
	   case ON_TRIGGER:
	    { rciWakeupOn (td, rbt->ready, rbt->readyArg);
	      rciSetTrigger (tgenTrigger, (int)rbt);
	      break;
	    }
	   case ON_FUNCTION:
	    { rciWakeupOn (td, rbt->ready, rbt->readyArg);
	      break;
	    }
	   case ON_SIGNAL:
	    { rciWakeupOn (td, simRobotsReady, (int)td);
	      simSetTimer (td->tcb->interval, (struct timeval*)0);
	      break;
	    }
	 }
	return (1);
}

tgenDrive (td)
RCI_DESC *td;
{
	char *fxnName = "tgenDrive() --";

	RCI_RBT *rbt = td->rbts[0];
	TgenBuf *tbuf;
	register int j;
	int numJ;
	float interval;
	struct timeval tv0;

	char *pfx = "tgenDrive() --";
	
	if (td->tcb->scheduleDisp == ON_SIGNAL)
	 { getSystemTime(&tv0);
	 }

	numJ = rbt->jls->numJ;

	tbuf = (TgenBuf*)td->tcb->sharedMem;

	/* ---- begin robot interface code ---- */

	if (td->tcb->scheduleDisp == ON_CLOCK)
	 { if (ROBOT_WAKEUP (rbt) < 0)
	    { rciAbort (0, "%s can't wakeup robot\n", fxnName);
	      return (-1);
	    }
	 }
	if ((*rbt->input)(td, rbt) < 0)
	 { return (-1);
	 }
 	if (td->tcb->cycleCount > 1 && rbt->how->power)
	 { 
	   JOINT_LEVEL_CONTROL(rbt);
	   if (tbuf->go)
	    { for (j=0; j<numJ; j++)
	       { MOVE_ANGLE (rbt, DEGTORAD*tbuf->buf[tbuf->idx++], j);
	       }
	      if (tbuf->idx/numJ == tbuf->go)
	       { tbuf->go = 0;
		 tbuf->idx = 0;
	       }
	    }
	 }
	if ((*rbt->output)(td, rbt) < 0)
	 { return (-1);
	 }
	if (td->tcb->scheduleDisp == ON_TRIGGER)
	 { rciSetTrigger (tgenTrigger, (int)rbt);
	 }
	if (td->tcb->scheduleDisp == ON_SIGNAL)
	 { simSetTimer (td->tcb->interval, &tv0);
	 }
	return (1);
}

tgenRelease(td, arg, code)
RCI_DESC *td;
int arg;
int code;
{
	RCI_RBT *rbt = td->rbts[0];
	int rcode;
	
	rciSetTrigger (NULLF, 0);	/* clear any trigger function */
	if (code == RELEASE_REQ)
	 { rcode = RBT_NORMAL;
	 }
	else
	 { rcode = RBT_HARDERROR;
	 }
	if ((*rbt->release)(td, rbt, rcode) < 0)
	 { return (-1);
	 }
	return (1);
}
