/*
  Source Code for ATI ISA FT adapter
 
  Adapted for RCCL from source provided by ATI, written by Jerry Dahl
*/

#ifdef X86sol_TARGET
#include <atireg.h>
#include <atiioctl.h>
#include "../drivers/inout.h"

#if MAPPED_IO
#define OUTB(p,v)	regs->p = (v)
#define OUTW(p,v)	regs->p = (v)
#define INB(p)		regs->p
#define INW(p)		regs->p
#else
#define OUTB(p,v)	outb(info->iobase+ATI_##p,v)
#define OUTW(p,v)	outw(info->iobase+ATI_##p,v)
#define INB(p)		inb(info->iobase+ATI_##p)
#define INW(p)		inw(info->iobase+ATI_##p)
#endif

static void atiReadData (AtiForceRec *rec, AtiSoftc *info)
{
#if MAPPED_IO
	AtiRegs *regs = info->regs;
#endif
	int i;

	OUTW (addr, 0x100);
	rec->dword = INW (data);
	for (i=0; i<7; i++)
	 { rec->fdata[i] = INW (data);
	 }
}

static int atiReadCom (int com, int mode, void *ptr, int n,
			AtiSoftc *info)
{
#if MAPPED_IO
	AtiRegs *regs = info->regs;
#endif
	unsigned short *dbuf = (unsigned short *)ptr;
	int retcode;

	if (atiWriteCom (com, mode, NULL, 0, info) < 0)
	 { return (-1);
	 }
	OUTW(addr,0);	  /* reset address */
	retcode = INW(data);
	if (mode != ATI_NO_MODE)
	 { INW(data);
	 }
	while (n--)
	 { *dbuf++ = INW(data);
	 }
	return (retcode);
}

static int atiWriteCom (int com, int mode, void *ptr, int n,
			AtiSoftc *info)
{
#if MAPPED_IO
	AtiRegs *regs = info->regs;
#endif
	unsigned short *dbuf = (unsigned short *)ptr;
	unsigned dword, retcode;
	int spin;

	/*
	   write command block to DPRAM at offset 0x000 
	 */
	OUTW(addr,0); 		/* set address register to command_block */
	OUTW(data,com);
	if (mode != ATI_NO_MODE)
	 { OUTW(data,mode);
	 }
	if (dbuf)
	 { while (n--)
	    { OUTW(data,*dbuf++);
	    }
	 }
	/*
	   read MailBox at DPRAM 0x3FE to clear host interrupt flag
	 */
	OUTW(addr,0x3FE);	     	/* set mailbox address */
	INW(data);		   	/* reset interrupt flag */
	OUTW(addr,0x3FE) ;	    	/* set mailbox address */
	OUTW(data,com);		    	/* clear acknowledgment */

	/*
	   write command word to Mailbox at DPRAM 0x3FF to set DSP intr flag
	 */
	OUTW(addr,0x3ff) ;
	OUTW(data,com) ;

	/*
	   waits for command completion (interrupt flag) from DSP
	 */
	retcode = 1;
	for (spin = ATI_SPIN_TIMEOUT; spin; spin--)
	 {
	   dword = INW(status);
	   // wait for interrupt flag
	   if((dword & 0x0010)==0)     		/* interrupt set */
	    {
	      /* read mailbox location at 0x3FE to reset interrupt flag */
	      OUTW(addr,0x3FE) ;	 	/* write dpram address */
	      retcode = INW(data) ;	 	/* reset interrupt flag */

	      /* test mailbox for command acknowledgment */
	      if(retcode & 0x00ff)		/* test for ack */
	       { break;
	       }
	    }
	 }
	if (spin)
	 { return retcode;
	 }
	else
	 { return -1;
	 }
}

static int atiHardwareReset (AtiSoftc *info)
{
	/* make the device quiescent. Called at attach and detach time */

	/* get card info */
	if (atiReadCom (ATI_CARD_DATA_CMD, ATI_READ, &info->cdata,
			sizeof(info->cdata)/2, info) < 0)
	 { return (-1);
	 }
	return (atiHardwareIntrEnable (info, 0));
}

static int atiHardwareClose(AtiSoftc *info)
{
	/* make the device quiet when it is closed */

	/* disable interrupts */
	return (atiHardwareIntrEnable (info, 0));
}

static int atiHardwareOpen (AtiSoftc *info)
{
	/* initialize the device when it is openned */	

	if (atiReadCom (ATI_CARD_DATA_CMD, ATI_READ, &info->cdata,
			sizeof(info->cdata)/2, info) < 0)
	 { return (-1);
	 }
	return (0);
}

static int atiHardwareIntrEnable (AtiSoftc *info, int enable)
{
	/* enable/disable device interrupts */

	if (enable != info->cdata.intrEnable)
	 { info->cdata.intrEnable = enable;
	   if (atiWriteCom (ATI_CARD_DATA_CMD, ATI_WRITE, &info->cdata,
			    sizeof(info->cdata)/2, info) < 0)
	    { return (-1);
	    }
	 }
	return (1);
}
#endif

