/*-=-=-=-=-=-=-=-=-=-=-=- RCCL/RCI Notice of Copyright -=-=-=-=-=-=-=-=-=-=-*
 *									    *
 *	Copyright (C) 1991 by John E. Lloyd.				    *
 *									    *
 *	Information on copyright permissions, as well as a complete	    *
 *	disclaimer of warranty, is given in the file COPYRIGHT.		    *
 *									    *
 *-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=*/
/*#include <stdio.h>*/

#include "../h/stringDefs.h"

#include "../h/cdefs.h"
#include "../h/rciTask.h"
#include "../h/robotControl.h"
#include "../h/errorCodes.h"
#include "../h/errorStack.h"
#include "../h/moper_com.h"
#include "../h/puma_kynvar.h"
#include "../h/robot_jls.h"
#include "../h/priDevice.h"
#include "../h/IOdevice.h"
#include "../h/robotMath.h"
#include "../h/rciMonitor.h"
#include "../h/rciMacros.h"
#include "../h/tval.h"
#include "../h/packByte.h"

#undef STARTUP_COUNT
#define STARTUP_COUNT	3

/* XXX Should move these into an include file XXX */

#define RCI_MSG_CODE      0x50  /* RCI message signature */
#define RCI_INIT_GATHER   0x51  /* Collect initial state info */
#define RCI_INIT_COMMANDS 0x52  /* Dispatch initial commands */
#define RCI_GATHER        0x53  /* Collect state info */
#define RCI_COMMANDS      0x54  /* Dispatch commands */
#define RCI_RELEASE       0x55  /* Release control */

extern int rciCpu;
extern char *pumaErrorString();

/* ---- error codes and directives to the wrapup function ---- */

#define ACTIVATED	1	/* robot must be ready to allow a write */
#define NO_NOTIFY	2	/* robot error; don't bother telling robot */
#define DOING_IO	3	/* doing I/O; can;t tell robot anything */

#define READ_BYTE(b,bp)		((b) = *bp++)
#define READ_SHORT(s,bp)	(s = unpackShort(bp), bp+=2)
#define READ_UBYTE(b,bp)	((b) = (unsigned char)*bp++)
#define READ_USHORT(s,bp)	(s = (unsigned short)unpackShort(bp), bp+=2)
#define READ_INT(i,bp)		(i = unpackInt(bp), bp+=4)
#define READ_INT3(i,bp)		(i = unpack3byteInt(bp), bp+=3)

static int rawForceRead; /* XXX shouldn't be a static */

crs_input(td, rbt)
RCI_DESC *td;
RCI_RBT *rbt;
{
	/* If we are running under the simulator, use the puma functions,
	   because the simulator controller emulates an old Unimation
	   controller. We are running under the simulator if the 
	   scheduling discipline is ON_SIGNAL */
	if (td->tcb->scheduleDisp == ON_SIGNAL ||
	    td->tcb->scheduleDisp == ON_CLOCK_SIM)
	 { return (puma_input (td, rbt));
	 }
	else
	 { return (crsplus_input (td, rbt));
	 }
}

crsZebraComputeForce (force, raw)
float *force;
int *raw;
{
	double accum;
	int useReducedForce = 0;
	
/* Matrix is calibrated to output Newtons and Newton-mm, with an input
   in the range [-1,1] */

	static float matrix[6][6] = {
#if 0
	   -3.3085,  54.5880,  -5.7139,  62.9138,   2.5368,-126.8194,
	    2.7955, -97.0946,   0.1177, 105.5729,  -0.3318,  -1.2085,
	  102.2091,  -1.9253, 108.9549,   4.8621, 102.3375,  -1.1971,

	   -1269.6,   -296.9,  -1515.8,    242.6,   2576.1,      1.1,
	    2207.4,   -157.5,  -2306.9,   -189.7,    -74.3,    374.9,
	    -141.7,   1359.3,   -173.3,   1624.3,    147.9,   1564.8 };
#endif
#if 0
/* Jan 20, 1996 */
	   -8.7662,  57.6528,  -3.4797,  67.9528,   3.2694,-130.2337,
	   13.1334,-101.3809,   2.3179, 101.3617,  -0.7891,   3.5634,
	   94.1888,  -3.3325, 110.1356,   7.6601, 100.4196,  -5.7662,

	   -1432.6,   -318.7,  -1433.8,    354.6,   2529.3,    -95.2,
	    2399.8,   -108.6,  -2313.6,   -312.4,    -89.8,    468.7,
	     105.3,   1684.2,     -8.1,   1524.3,    -31.7,   1658.8 };
#endif
#if 1
/* Aug  8, 1996 */
   -7.0243,   62.7964,   -6.8676,   64.1884,   -1.4510, -128.9461,
    6.4169, -104.1935,    1.6789,  106.2509,    1.3776,    0.1503,
  104.8312,   -3.8633,  111.3726,    3.7942,   98.1773,    0.9973,

   -1314.3,     -46.6,   -1433.5,     209.3,    2476.9,    -181.2,
    2096.3,      69.1,   -2353.6,    -253.4,     -62.5,     123.6,
    -179.2,    1601.5,     -53.4,    1597.8,     116.6,    1448.8 };
#endif
#if 0
   -00608.0,   00668.0,  -00034.0,  00098.0,    00129.0,  -00255.0,
    00613.0,  -00716.0,   00013.0,  00097.0,    00250.0,  -00255.0,
    15710.0,  -15629.0,   00039.0, -00060.0,    00557.0,  -00619.0,

    065410.0,  -065120.0, -023770.0,  022970.0, 239880.0, -239560.0,
    096180.0,  -095840.0, -008370.0,  009820.0, 206790.0, -208510.0,
   -005490.0,   007070.0, -000640.0,  002220.0, 009140.0, -007430.0 };
#endif


	/* This is a special matrix which eliminates the reading from
	   the second strain gauge and computes only fx,fy,fz and tx,tz
	   in a hand frame given by a z translation of 100 mm from the
	   sensor center. */
	static float rmatrix[6][6] = { 
	   21.3616,      0.0, -22.5817,  -0.0789,  -0.6506,   0.0870,
	  -41.0847,      0.0,  30.1203, 217.6167,   5.3376,-226.9344,
	  101.3390,      0.0, 109.5498,   7.0838, 102.4499,  -5.6730,

	   -1403.8,      0.0,  -1424.0,    585.3,   2593.4,   -689.2,
	    2136.2,      0.0,  -2258.2,     -7.9,    -65.1,      8.7,
	     472.6,      0.0,   -593.3,     55.7,     68.6,   4725.0 };

	int i, j;

	if (useReducedForce)
	 { 
	   for (i=0; i<6; i++)
	    { accum = 0;
	      for (j=0; j<6; j++)
	       { accum += rmatrix[i][j]*raw[j]/512.0;
	       }
	      force[i] = accum;
	    }
	 }
	else
	 { for (i=0; i<6; i++)
	    { accum = 0;
	      for (j=0; j<6; j++)
	       { accum += matrix[i][j]*raw[j]/512.0;
	       }
	      force[i] = accum;
	    }
	 }
}

static Tval timeCheck;
extern Tval rciWallClock();

crsplus_input(td, rbt)
RCI_DESC *td;
RCI_RBT *rbt;
{
 	register HOW *how = rbt->how;
 	register JLS *jls = rbt->jls;
	RCI_TCB *tcb = td->tcb;
	IO_DEVICE *dev = (IO_DEVICE*)rbt->dev;

	char *pfx = "rbtInput() --";

	/* ---- read stuff in from the robot and load the 'how' struct ---- */

	 { unsigned char dataBuf[MAX_DATABUF];
	   float ang[6];
	   int   enc[6];
	   register unsigned char *bp = dataBuf;
	   register int n, j, k;
	   register float rate;
	   int readFlags;
	   int stateFlags;

	   rbt->notice = DOING_IO;
	   if ((n = (*dev->read) (dev, dataBuf, MAX_DATABUF)) < 0)
	    { errorMsg (getErrorCode(0), "%s error reading from robot\n",pfx);
	      rciAbortErrors();
	      return (-1);
	    }
	   rbt->notice = ACTIVATED;

	   if (n == sizeof(int))
	    { int error;
	      READ_INT(error,bp);
	      rciAbort (ERobotError, "controller error 0x%x:\n%s\n",
			error, pumaErrorString (error) );
	      rbt->notice = NO_NOTIFY;
	      return (-1);
	    }
	   READ_INT3(enc[0],bp);
	   READ_INT3(enc[1],bp);
	   READ_INT3(enc[2],bp);
	   READ_INT3(enc[3],bp);
	   READ_INT3(enc[4],bp);
	   READ_INT3(enc[5],bp);

	   /* ---- compute angle and velocity ---- */

	   (*jls->encoderToAngle)(ang, enc, jls);	   

	   how->time = rciTime();

	   if (tcb->cycleCount == 1)
	    { 
	      how->angVel[0] = 0.0;
	      how->angVel[1] = 0.0;
	      how->angVel[2] = 0.0;

	      how->angVel[3] = 0.0;
	      how->angVel[4] = 0.0;
	      how->angVel[5] = 0.0;

	      how->encVel[0] = 0.0;
	      how->encVel[1] = 0.0;
	      how->encVel[2] = 0.0;

	      how->encVel[3] = 0.0;
	      how->encVel[4] = 0.0;
	      how->encVel[5] = 0.0;

	      how->angSet[0] = ang[0];
	      how->angSet[1] = ang[1];
	      how->angSet[2] = ang[2];

	      how->angSet[3] = ang[3];
	      how->angSet[4] = ang[4];
	      how->angSet[5] = ang[5];

	      how->encSet[0] = enc[0];
	      how->encSet[1] = enc[1];
	      how->encSet[2] = enc[2];

	      how->encSet[3] = enc[3];
	      how->encSet[4] = enc[4];
	      how->encSet[5] = enc[5];
	    }
	   else
	    { rate = 1000.0/(float)(tcb->interval);

	      how->angVel[0] = (ang[0] - how->ang[0]) * rate;
	      how->angVel[1] = (ang[1] - how->ang[1]) * rate;
	      how->angVel[2] = (ang[2] - how->ang[2]) * rate;

	      how->angVel[3] = (ang[3] - how->ang[3]) * rate;
	      how->angVel[4] = (ang[4] - how->ang[4]) * rate;
	      how->angVel[5] = (ang[5] - how->ang[5]) * rate;

	      how->encVel[0] = (enc[0] - how->enc[0]) * rate;
	      how->encVel[1] = (enc[1] - how->enc[1]) * rate;
	      how->encVel[2] = (enc[2] - how->enc[2]) * rate;

	      how->encVel[3] = (enc[3] - how->enc[3]) * rate;
	      how->encVel[4] = (enc[4] - how->enc[4]) * rate;
	      how->encVel[5] = (enc[5] - how->enc[5]) * rate;
	    }

	   if (k = how->loadSettings)
	    { if (how->controlLevel == MOTOR_LEVEL)
	       { 
		 if (k & 0x01)
		  { how->encSet[0] = enc[0];
		  }
		 if (k & 0x02)
		  { how->encSet[1] = enc[1];
		  }
		 if (k & 0x04)
		  { how->encSet[2] = enc[2];
		  }
   
		 if (k & 0x08)
		  { how->encSet[3] = enc[3];
		  }
		 if (k & 0x10)
		  { how->encSet[4] = enc[4];
		  }
		 if (k & 0x20)
		  { how->encSet[5] = enc[5];
		  }
	       }
	      else
	       { 
		 if (k & 0x01)
		  { how->angSet[0] = ang[0];
		  }
		 if (k & 0x02)
		  { how->angSet[1] = ang[1];
		  }
		 if (k & 0x04)
		  { how->angSet[2] = ang[2];
		  }
   
		 if (k & 0x08)
		  { how->angSet[3] = ang[3];
		  }
		 if (k & 0x10)
		  { how->angSet[4] = ang[4];
		  }
		 if (k & 0x20)
		  { how->angSet[5] = ang[5];
		  }
	       }
	      how->loadSettings = 0;		    
	    }

	   how->ang[0] = ang[0];
	   how->ang[1] = ang[1];
	   how->ang[2] = ang[2];

	   how->ang[3] = ang[3];
	   how->ang[4] = ang[4];
	   how->ang[5] = ang[5];

	   how->enc[0] = enc[0];
	   how->enc[1] = enc[1];
	   how->enc[2] = enc[2];

	   how->enc[3] = enc[3];
	   how->enc[4] = enc[4];
	   how->enc[5] = enc[5];

	   if (how->checking & CHK_MAXPOS)
	    { int jnts = 0;
	      if (ang[0] < jls->jmin[0] || ang[0] > jls->jmax[0])
	       { jnts |= (1<<0);
	       }
	      if (ang[1] < jls->jmin[1] || ang[1] > jls->jmax[1])
	       { jnts |= (1<<1);
	       }
	      if (ang[2] < jls->jmin[2] || ang[2] > jls->jmax[2])
	       { jnts |= (1<<2);
	       }
	      if (ang[3] < jls->jmin[3] || ang[3] > jls->jmax[3])
	       { jnts |= (1<<3);
	       }
	      if (ang[4] < jls->jmin[4] || ang[4] > jls->jmax[4])
	       { jnts |= (1<<4);
	       }
	      if (ang[5] < jls->jmin[5] || ang[5] > jls->jmax[5])
	       { jnts |= (1<<5);
	       }
	      if (jnts)
	       { rciAbort (EMaxPos,
		"joint limit exceeded for joints 0x%x\n", jnts);
		 return (-1);
	       }
	    }

	   if (how->checking & CHK_MAXVEL)
	    { int jnts = 0;

	      if (FABS (how->angVel[0]) > jls->maxVel[0])
	       { jnts |= (1<<0);
	       }
	      if (FABS (how->angVel[1]) > jls->maxVel[1])
	       { jnts |= (1<<1);
	       }
	      if (FABS (how->angVel[2]) > jls->maxVel[2])
	       { jnts |= (1<<2);
	       }

	      if (FABS (how->angVel[3]) > jls->maxVel[3])
	       { jnts |= (1<<3);
	       }
	      if (FABS (how->angVel[4]) > jls->maxVel[4])
	       { jnts |= (1<<4);
	       }
	      if (FABS (how->angVel[5]) > jls->maxVel[5])
	       { jnts |= (1<<5);
	       }

	      if (jnts)
	       { int j;

		 rciAbort (EMaxVel,
		"velocity limit exceeded for joints 0x%x\n", jnts);

		 for (j=0; j<6; j++)
		  { if (jnts & (1<<j))
		     { break;
		     }
		  }
		 rciPrintf("\nangVel[%d]=%g ang[%d]=%g rate=%g timediff=%g\n",
			    j, how->angVel[j]*RADTODEG,
			    j, how->ang[j]*RADTODEG,
			    (double)rate,
			    tvalToDouble(subTval(rciWallClock(),timeCheck)));
		 return (-1);
	       }
	      timeCheck = rciWallClock();
	    }
	   how->state 	    = 0;

	   READ_UBYTE(stateFlags, bp);
	   how->power 	    = ((stateFlags & POWER_ON) != 0);
	   how->calibrated  = ((stateFlags & IS_CALIBRATED) != 0);

	   READ_UBYTE(readFlags, bp);

	   for (k=0; k<how->numPioIn; k++)
	    { READ_USHORT (how->pioInput[k], bp);
	    }
	   for (k=0; k<how->numPioOut; k++)
	    { READ_USHORT (how->pioOutput[k], bp);
	    }

	   if (readFlags & MARK_FOUND)
	    { READ_UBYTE (how->markFound, bp);
	    }

	   if (how->torqueOK = (readFlags & TORQUE_OK))
	    { for (j=0; j<6; j++)
	       { short indata;

		 READ_SHORT(indata, bp);
		 how->mtorque[j] = jls->currentToMtorque[j] * indata;
	       }

	      if (how->checking & CHK_MAXTOR)
	       { int jnts = 0;
		 if (FABS (how->mtorque[0]) > jls->maxMtorque[0])
		  { jnts |= (1<<0);
		  }
		 if (FABS (how->mtorque[1]) > jls->maxMtorque[1])
		  { jnts |= (1<<1);
		  }
		 if (FABS (how->mtorque[2]) > jls->maxMtorque[2])
		  { jnts |= (1<<2);
		  }
		 if (FABS (how->mtorque[3]) > jls->maxMtorque[3])
		  { jnts |= (1<<3);
		  }
		 if (FABS (how->mtorque[4]) > jls->maxMtorque[4])
		  { jnts |= (1<<4);
		  }
		 if (FABS (how->mtorque[5]) > jls->maxMtorque[5])
		  { jnts |= (1<<5);
		  }
		 if (jnts)
		  { rciAbort (EMaxTor,
		   "maximum motor torque exceeded for joints 0x%x\n", jnts);
		    return (-1);
		  }
	       }
	    }
	   how->potOK = 0;
#if 0
	   if (how->pendantOK = (readFlags & PENDNT_OK))
	    { char buf[8];

	      for (k=0; k<6; k++)
	       { READ_BYTE(buf[k], bp);
	       }
	      buf[6] = buf[7] = 0;
	      how->pendantInput[0] = unpackInt(&buf[0]);
	      how->pendantInput[1] = unpackInt(&buf[4]);
	    }
#endif
	   /* The next set of bytes contain gripper info */

	   READ_SHORT(how->hand, bp);
	   READ_SHORT(how->handForce, bp);

	   if (how->forceOK = (readFlags & FORCE_OK))
	    { int raw[6];

	      READ_SHORT(raw[0], bp);
	      READ_SHORT(raw[1], bp);
	      READ_SHORT(raw[2], bp);

	      READ_SHORT(raw[3], bp);
	      READ_SHORT(raw[4], bp);
	      READ_SHORT(raw[5], bp);

	      if (!rawForceRead)
	       { crsZebraComputeForce (how->forceInput, raw);
	       }
	      else
	       { for (k=0; k<6; k++)
		  { how->forceInput[k] = raw[k];
		  }
	       }
	    }
	   else
	    { zeroVf ((float*)how->forceInput, 6);
	    }
	   how->jstickOK = 0;
	   how->teleopOK = 0;

	   if (readFlags & TEST_OK)
	    { int skip;

	      skip = *bp++;
	      bp += 2*skip;
	    }

	   how->adc = 0;
	   how->slu = 0;
	 }

	return (1);
}

#define PUSH_COM(c)    if (bp + 1 > top) goto overflow; \
			*bp++ = (c);

#define PUSH_COM_BYTE(c,x) if (bp + 2 > top) goto overflow; \
			*bp++ = (c); \
			*bp++ = (x);

#define PUSH_COM_SHORT(c,s) if(bp + 3 > top) goto overflow; \
			*bp++ = (c); \
			packShort(s,bp); \
			bp+=2;

#define PUSH_COM_INT3(c,i) if(bp + 4 > top) goto overflow; \
			*bp++ = (c); \
			pack3byteInt(i,bp); \
			bp+=3;

crs_output(td, rbt)
RCI_DESC *td;
RCI_RBT *rbt;
{
	/* See comment in crs_input() */
	if (td->tcb->scheduleDisp == ON_SIGNAL ||
	    td->tcb->scheduleDisp == ON_CLOCK_SIM)
	 { return (puma_output (td, rbt));
	 }
	else
	 { return (crsplus_output (td, rbt));
	 }
}

crsplus_output(td, rbt)
RCI_DESC *td;
RCI_RBT *rbt;
{
	unsigned char dataBuf[MAX_DATABUF];

	RCI_TCB *tcb = td->tcb;
	IO_DEVICE *dev = (IO_DEVICE*)rbt->dev;
	HOW *how = rbt->how;
	JLS *jls = rbt->jls;
	register CHG *chg = rbt->chg;
	register unsigned char *bp = dataBuf + 2;
	unsigned char *top = &dataBuf[MAX_DATABUF];

	char *pfx = "rbtOutput() --";

	/* ---- do the control commands first ---- */

	if (tcb->newRate)
	 { PUSH_COM_BYTE (SET_RATE_C, MIN(tcb->interval,255));
	 }
	   
	if (chg->ctrlcom)
  	 { if (chg->paramset)
	    { rciAbort (ENotSupported,
	"setting servo parameters not supported on CRS robots\n");
	      return (-1);
	    }

	   if (chg->integration.com)
	    { /* just ignore this for now */
	      chg->integration.com = 0;
	    }

	   if (chg->calibMode.com)
	    { rciAbort (ENotSupported,
	"SET_CALIBMODE() not supported on CRS robots\n");
	      return (-1);
	    }

	   if (chg->calibflag.com)
	    { rciAbort (ENotSupported,
	"SET_CALIBFLAG() not supported on CRS robots\n");
	      return (-1);
	    }

	   if (chg->power.com)
	    { if (chg->power.val.i)
	       { PUSH_COM (ON_POWER_C);
	       }
	      else
	       { PUSH_COM (OFF_POWER_C);
	       }
	      chg->power.com = 0;
	    }
	    
	   if (chg->maintCom.com)
	    { rciAbort (ENotSupported,
	"SET_MAINT_COM() not supported on CRS robots\n");
	      return (-1);
	    }
	    
	   if (chg->checking.com)
	    { how->checking = chg->checking.val.l & CHK_RBTALL;
	      chg->checking.com = 0;
	    }

	   if (chg->powerOnRelease.com)
	    { how->powerOnRelease = !!chg->powerOnRelease.val.i;
	      chg->powerOnRelease.com = 0;
	    }
	    
	   chg->ctrlcom = 0;
	 }


	/* ----- now do the motion control stuff ----- */

	 { unsigned long newTorqueMode;
	   unsigned long modeChange;
	   float deltaT;
	   register int j;
	   char *illCmd;
	
	   deltaT = tcb->interval / 1000.;

	   /* check for a change in the control level */

	   if (chg->controlLevel.com)
	    { how->controlLevel = chg->controlLevel.val.i;
	      chg->controlLevel.com = 0;
	    }

	   /* look for the stop command first */

	   if (chg->stop.com)
	    { how->torqueMode = 0;
	      how->loadSettings = 0x3f;

	      for (j=0; j<PUMAJ; j++)
	       { chg->motion[j].com = 0;
		 how->mtorqueSet[j] = 0.0;
		 how->jtorqueSet[j] = 0.0;
	       }
	      PUSH_COM (STOP_ALL_C);
	      chg->stop.com = 0;
	    }

	   if (how->controlLevel == JOINT_LEVEL)
	    { 
	      int newEnc[PUMAJ];
	      float newMtorque[PUMAJ];
	      float newMtorqueBias[PUMAJ];
	      int biasSet = NO;
	      char com[PUMAJ];
   
	      newTorqueMode = how->torqueMode;

	      for (j=0; j<PUMAJ; j++)
	       { switch (com[j] = chg->motion[j].com)
		  { case 0:
		     { break;
		     }
		    case CHG_ANGLE:
		     { float ang = chg->motion[j].val.f;
   
		       if (how->checking & CHK_REQPOS)
			{ if (ang < jls->jmin[j] || ang > jls->jmax[j])
			   { rciAbort (EReqPos,
	"%s requested angle %g exceeds limits, joint %d\n",pfx,
			     ang*RADTODEG, j+1);
			     return (-1);
			   }
			}
		       if (how->checking & CHK_REQVEL)
			{ if (FABS(ang - how->angSet[j]) > jls->maxVel[j]*deltaT)
			   { rciAbort (EReqVel,
	"%s requested velocity %g exceeds velocity limits, joint %d\n", pfx,
			     RADTODEG*FABS(ang - how->angSet[j])/deltaT, j+1);
rciPrintf ("j = %d how->angSet = %8.3f deltaT = %g ang = %8.3f\n",
			j+1, (double)how->angSet[j]*RADTODEG, (double)deltaT, 
			ang*RADTODEG);
			     return (-1);
			   }
			}
		       how->angSet[j] = ang;
		       how->jtorqueSet[j] = 0.0;
		       newTorqueMode &= ~(1<<j);
		       break;
		     }
		    case CHG_JTORQUE:
		     {
		       how->jtorqueSet[j] = chg->motion[j].val.f;
		       newTorqueMode |= (1<<j);
		       break;
		     }
		    case CHG_STOP:
		     {
		       how->jtorqueSet[j] = 0.0;
		       newTorqueMode &= ~(1<<j);
		       break;
		     }
		    case CHG_ANGLE_VALUE:
		     {
		       rciAbort (EBadCmd,
	"%s SET ANGLE VALUE not currently supported for PUMAs\n");
		       return (-1);
		       break;
		     }
		    case CHG_ENCODER:
		     { illCmd = "SET ENCODER";
		       break;
		     }
		    case CHG_ENCODER_VALUE:
		     { illCmd = "SET ENCODER VALUE";
		       break;
		     }
		    case CHG_MTORQUE:
		     { illCmd = "SET MTORQUE";
		       rciAbort (EBadCmd,
     "%s illegal joint level motion command %s\n", pfx, illCmd);
		       return (-1);
		       break;
		     }
		    default:
		     { rciAbort (EBadCmd,
     "%s illegal joint level motion command (%d)\n", pfx, com[j]);
		       return (-1);
		       break;
		     }
		  }
		 chg->motion[j].com = 0;

		 switch (chg->bias[j].com)
		  { case 0:
		     { break;
		     }
		    case CHG_JTORQUE_BIAS:
		     { how->jtorqueBias[j] = chg->bias[j].val.f;
		       biasSet = YES;
		       break;
		     }
		    case CHG_MTORQUE_BIAS:
		     { illCmd = "SET MTORQUE BIAS";
		       rciAbort (EBadCmd,
     "%s illegal joint level bias command %s\n", pfx, illCmd);
		       return (-1);
		       break;
		     }
		    default:
		     { rciAbort (EBadCmd,
     "%s illegal joint level bias command (%d)\n", pfx, chg->bias[j].com);
		       return (-1);
		       break;
		     }
		  }		    
		 chg->bias[j].com = 0;
	       }

	      modeChange = newTorqueMode ^ how->torqueMode;
	      how->torqueMode = newTorqueMode;
			  
	      (*jls->angleToEncoder)(newEnc, how->angSet, jls);
	      if (how->torqueMode)
	       { (*jls->jtorqueToMtorque)(newMtorque, how->jtorqueSet, jls);
	       }
	      if (biasSet)
	       { (*jls->jtorqueToMtorque)(newMtorqueBias,how->jtorqueBias,jls);
	       }
	      
	      /* now go and dispatch the commands */
   
	      for (j=0; j<PUMAJ; j++)
	       { float newBias;

		 if (how->torqueMode & (1<<j))
		  { float mtor = newMtorque[j];
		    if (mtor != how->mtorqueSet[j] || modeChange & (1<<j))
		     { if (how->checking & CHK_REQTOR)
			{ if (FABS(mtor) > jls->maxMtorque[j])
			   { rciAbort (EReqTor,
      "%s requested motor torque %f exceeds bounds, joint %d\n",
				       pfx,(double)mtor,j+1);
			     return (-1);
			   }
			}
		       PUSH_COM_SHORT (MOVE_CUR_C | j, 
				     mtor*jls->mtorqueToCurrent[j] +
				     jls->currentBias[j]);
		     }
		    how->encSet[j] = newEnc[j];
		    how->mtorqueSet[j] = mtor;
		  }
		 else
		  { if (newEnc[j] != how->enc[j] || modeChange & (1<<j))
		     { PUSH_COM_INT3 (MOVE_ENC_C | j, newEnc[j]);
		     }
		    how->encSet[j] = newEnc[j];
		    how->mtorqueSet[j] = how->torqueMode ? newMtorque[j]:0.0;
		  }
		 if (biasSet)
		  { rciAbort (ENotSupported,
	"SET_JTORQUE_BIAS() not supported on CRS robots\n");
		    return (-1);
		  }
	       }
	      how->loadSettings |= newTorqueMode;
	    }
	   else 	/* control level == MOTOR LEVEL */
	    { int newEnc[PUMAJ];
	      float newAng[PUMAJ];
	      float newMtorque[PUMAJ];
	      int biasSet = NO;
	      int encValueSet = NO;
	      char com[PUMAJ];
   
	      newTorqueMode = how->torqueMode;

	      for (j=0; j<PUMAJ; j++)
	       { switch (com[j] = chg->motion[j].com)
		  { case 0:
		     { newEnc[j] = how->encSet[j];
		       newMtorque[j] = how->mtorqueSet[j];
		       break;
		     }
		    case CHG_ENCODER:
		     { newEnc[j] = chg->motion[j].val.i;
		       newMtorque[j] = 0.0;
		       newTorqueMode &= ~(1<<j);
		       break;
		     }
		    case CHG_MTORQUE:
		     { newEnc[j] = how->encSet[j];
		       newMtorque[j] = chg->motion[j].val.f;
		       newTorqueMode |= (1<<j);
		       break;
		     }
		    case CHG_STOP:
		     { newEnc[j] = how->encSet[j];
		       newMtorque[j] = 0.0;
		       newTorqueMode &= ~(1<<j);
		       break;
		     }
		    case CHG_ENCODER_VALUE:
		     { newEnc[j] = chg->motion[j].val.i;
		       newMtorque[j] = 0.0;
		       newTorqueMode &= ~(1<<j);

		       how->enc[j] = chg->motion[j].val.i;
		       how->encSet[j] = chg->motion[j].val.i;
		       encValueSet = YES;
		       break;
		     }
		    case CHG_ANGLE:
		     { illCmd = "SET ANGLE";
		       break;
		     }
		    case CHG_ANGLE_VALUE:
		     { illCmd = "SET ANGLE VALUE";
		       break;
		     }
		    case CHG_JTORQUE:
		     { illCmd = "SET JTORQUE";
		       rciAbort (EBadCmd,
     "%s illegal motor level motion command %s\n", pfx, illCmd);
		       return (-1);
		       break;
		     }
		    default:
		     { rciAbort (EBadCmd,
     "%s illegal motor level motion command (%d)\n", pfx, com[j]);
		       return (-1);
		       break;
		     }
		  }

		 switch (chg->bias[j].com)
		  { case 0:
		     { break;
		     }
		    case CHG_MTORQUE_BIAS:
		     { rciAbort (ENotSupported,
	"SET_MTORQUE_BIAS() not supported on CRS robots\n");
		       return (-1);
		     }
		    case CHG_JTORQUE_BIAS:
		     { illCmd = "SET JTORQUE BIAS";
		       rciAbort (EBadCmd,
     "%s illegal motor level bias command %s\n", pfx, illCmd);
		       return (-1);
		       break;
		     }
		    default:
		     { rciAbort (EBadCmd,
     "%s illegal motor level bias command (%d)\n", pfx, chg->bias[j].com);
		       return (-1);
		       break;
		     }
		  }		    
		 chg->bias[j].com = 0;
	       }

	      if (encValueSet)
	       { (*jls->encoderToAngle)(how->ang, how->enc, jls);
		 (*jls->encoderToAngle)(how->angSet, how->encSet, jls);
	       }
	      modeChange = newTorqueMode ^ how->torqueMode;
	      how->torqueMode = newTorqueMode;
   
	      (*jls->encoderToAngle)(newAng, newEnc, jls);
	      if (how->torqueMode)
	       { (*jls->mtorqueToJtorque)(how->jtorqueSet, newMtorque, jls);
	       }
	      else
	       { how->jtorqueSet[0] = 0.0;
		 how->jtorqueSet[1] = 0.0;
		 how->jtorqueSet[2] = 0.0;

		 how->jtorqueSet[3] = 0.0;
		 how->jtorqueSet[4] = 0.0;
		 how->jtorqueSet[5] = 0.0;
	       }
	      if (biasSet)
	       { (*jls->mtorqueToJtorque)(how->jtorqueBias, 
					  how->mtorqueBias, jls);
	       }

	      for (j=0; j<PUMAJ; j++)
	       { 		 
		 if (how->checking & CHK_REQPOS)
		  { if (newAng[j] < jls->jmin[j] || newAng[j] > jls->jmax[j])
		     { rciAbort (EReqPos,
	"%s requested angle %g exceeds limits, joint %d\n",pfx,
		       RADTODEG*newAng[j],j+1);
		       return (-1);
		     }
		  }
		 if (how->checking & CHK_REQVEL)
		  { if (FABS(newAng[j] - how->angSet[j]) > jls->maxVel[j]*deltaT)
		     { rciAbort (EReqVel,
	"%s requested velocity %g exceeds limits, joint %d\n", pfx,
		       RADTODEG*FABS(newAng[j] - how->angSet[j])/deltaT, j+1);
		       return (-1);
		     }
		  }

		 if (how->torqueMode & (1<<j))
		  { float mtor = newMtorque[j];
		    if (mtor != how->mtorqueSet[j] || modeChange&(1<<j))
		     { if (how->checking & CHK_REQTOR)
			{ if (FABS(mtor) > jls->maxMtorque[j])
			   { rciAbort (EReqTor,
  "%s requested motor torque %f exceeds bounds, joint %d\n",
				       pfx,(double)mtor,j+1);
			     return (-1);
			   }
			}
		       PUSH_COM_SHORT (MOVE_CUR_C | j, 
				  mtor*jls->mtorqueToCurrent[j] +
				  jls->currentBias[j]);
		     }
		  }
		 else
		  { if (com[j] == CHG_ENCODER_VALUE)
		     { PUSH_COM_INT3 (LOAD_ENC_C | j, chg->motion[j].val.i);
		       how->calibMode &= ~(1<<j);
		     }
		    else if (com[j] == CHG_STOP)
		     { how->loadSettings |= (1<<j);
		       PUSH_COM (STOP_JNT_C | j);
		     }
		    else if (newEnc[j] != how->enc[j] || modeChange&(1<<j))
		     { PUSH_COM_INT3 (MOVE_ENC_C | j, newEnc[j]);
		     }
		  }
		 how->mtorqueSet[j] = newMtorque[j];
		 how->encSet[j] = newEnc[j];
		 how->angSet[j] = newAng[j];
		 chg->motion[j].com = 0;
	       }
	      how->loadSettings |= how->torqueMode;
	    }
	
#if 1
	   if (chg->hand.com)
	    { switch (chg->hand.com)
	       { case CHG_HANDPOS:
		  { PUSH_COM_BYTE (HAND_POS_C, chg->hand.val.i);
		    PUSH_COM (HAND_POS_C); /* remove when servo if fixed */
		    break;
		  }
		 case CHG_HANDFORCE:
		  { PUSH_COM_BYTE (HAND_FORCE_C, chg->hand.val.i);
		    PUSH_COM (HAND_FORCE_C); /* remove when servo if fixed */
		    break;
		  }
		 default:
		  { rciAbort (EBadCmd, 
	"illegal command code (%d) specified for hand\n", chg->hand.com);
		    return (-1);
		  }
	       }
	      chg->hand.com = 0;
	    }
#endif
	 }
       
	if (chg->readcom)
 	 {
	   if (chg->torque.com)
	    { PUSH_COM_BYTE (TORQUE_READ_C, chg->torque.val.i);
	      chg->torque.com = 0;
	    }

	   if (chg->pot.com)
	    { rciAbort (ENotSupported,
	"SET_POT_READ() not supported on CRS robots\n");
	      return (-1);
	    }

	   if (chg->pendant.com)
	    { rciAbort (ENotSupported,
	"SET_PENDANT_READ() not supported on CRS robots\n");
	      return (-1);
	    }

	   if (chg->force.com)
	    { PUSH_COM_BYTE (FORCE_READ_C, chg->force.val.i);
	      rawForceRead = (chg->force.val.i > 1);
	      chg->force.com = 0;
	    }

	   if (chg->teleop.com)
	    { rciAbort (ENotSupported,
	"SET_TELEOP_READ() not supported on CRS robots\n");
	      return (-1);
	    }

	   if (chg->jstick.com)
	    { rciAbort (ENotSupported,
	"SET_JSTICK_READ() not supported on CRS robots\n");
	      return (-1);
	    }

	   if (chg->adc.com)
	    { rciAbort (ENotSupported,
	"SET_ADC_CHAN() not supported on CRS robots\n");
	      return (-1);
	    }

	   if (chg->slu.com)
	    { rciAbort (ENotSupported,
	"SET_SLU_CHAN() not supported on CRS robots\n");
	      return (-1);
	    }

	   if (chg->setTest.com)
	    { PUSH_COM_BYTE (TEST_READ_C, chg->setTest.val.i);
	      chg->setTest.com = 0;
	    }

	   chg->readcom = 0;
	 }

	if (chg->writecom)
  	 { int n;

	   for (n=0; n<how->numPioOut; n++)
	    { int cmd;
	      if (cmd = chg->pioOutput[n].com)
	       { rciAbort (ENotSupported,
	"PUT,SET,CLEAR_PIO() not supported on CRS robots\n");
		 return (-1);
	       }
	    }

	   for (n=0; n<how->numDac; n++)
	    { if (chg->dacOutput[n].com)
	       { rciAbort (ENotSupported,
	"PUT_DAC() not supported on CRS robots\n");
		 return (-1);
	       }
	    }

	   for (n=0; n<how->numSlu; n++)
	    { if (chg->sluOutput[n].com)
	       { rciAbort (ENotSupported,
	"PUT_SLU() not supported on CRS robots\n");
		 return (-1);
	       }
	    }
   
	   if (chg->pendantMsg.com)
	    { rciAbort (ENotSupported,
	"PUT_PENDANT_MSG() not supported on CRS robots\n");
	      return (-1);
	    }

	   if (chg->putTest.com)
	    { PUSH_COM_SHORT (PUT_TEST_C, chg->putTest.val.i);
	      chg->putTest.com = 0;
	    }

	   chg->writecom = 0;
	 }

	if (chg->putVectorSet)
	 { rciAbort (ENotSupported,
	"PUT_VECTOR() not supported on CRS robots\n");
	   return (-1);
	 }

	 { int n = bp - (dataBuf+2);
	   int k;

	   dataBuf[0] = RCI_COMMANDS;
	   dataBuf[1] = n;
	   rbt->notice = DOING_IO;
	   if ((*dev->write) (dev, dataBuf, n+2) != n+2)
  	    { errorMsg (getErrorCode(0), "%s error writing to robot\n",pfx);
	      rciAbortErrors();
	      return (-1);
	    }
	   rbt->notice = ACTIVATED;
	 }

	return (1);

overflow:
	rciAbort (ETooMany, "too many commands sent to robot\n");
	return (-1);
}

#define WAKEUP_SIG	1
#define RELEASE_SIG	2

crs_signal (dev)
IO_DEVICE *dev;
{
	char comBuf[2];
	
	comBuf[0] = RCI_GATHER;
	comBuf[1] = 0;
	(*dev->signal)(dev, 0);	/* just arms the interrupt */
	if ((*dev->write) (dev, comBuf, 2) != 2)
	 { errorMsg (getErrorCode(0), "rbtSignal() -- can't signal robot\n");
	   rciAbortErrors();
	   return (-1);
	 }
	return (1);
}

crs_startup(td, rbt)
RCI_DESC *td;
RCI_RBT *rbt;
{
	/* See comment in crs_input() */
	if (td->tcb->scheduleDisp == ON_SIGNAL ||
	    td->tcb->scheduleDisp == ON_CLOCK_SIM)
	 { return (puma_startup (td, rbt));
	 }
	else
	 { return (crsplus_startup (td, rbt));
	 }
}

crsplus_startup (td, rbt)
RCI_DESC *td;
RCI_RBT *rbt;
{
	int n;
	unsigned char dataBuf[MAX_DATABUF];

	RCI_TCB *tcb = td->tcb;
	IO_DEVICE *dev = (IO_DEVICE*)rbt->dev;
 	register HOW *how = rbt->how;
	JLS *jls = rbt->jls;
	register unsigned char *bp = dataBuf;
	CHG *chg = rbt->chg;

	char *pfx = "rbtStartup() --";

	if ((*dev->attach)(dev) < 0)
	 { errorMsg (getErrorCode(0), "%s cannot connect to robot\n",pfx);
	   rciAbortErrors();
	   return (-1);
	 }

	/* set any necessary commands */

	if (chg->checking.com)
	 { how->checking = chg->checking.val.l & CHK_RBTALL;
	   chg->checking.com = 0;
	 }

	if (chg->powerOnRelease.com)
	 { how->powerOnRelease = !!chg->powerOnRelease.val.i;
	   chg->powerOnRelease.com = 0;
	 }

	/* interrupt the robot and get startup information */

	dataBuf[0] = RCI_INIT_GATHER;
	dataBuf[1] = 0;
	rbt->notice = DOING_IO;
	if ((*dev->write)(dev, dataBuf, 2) != 2)
	 { errorMsg (getErrorCode(0), 
"%s cannot write INIT_GATHER to robot\n",pfx);
	   rciAbortErrors();
	   return (-1);
	 }
	if ((n = (*dev->read) (dev, dataBuf, MAX_DATABUF)) < 0)
	 { errorMsg (getErrorCode(0), "%s error reading from robot\n",pfx);
	   rciAbortErrors();
	   return (-1);
	 }	 

	if (n == sizeof(int))
	 { int error;
	   READ_INT(error,bp);
	   rciAbort (ERobotError, "controller error 0x%x:\n%s\n",
		     error, pumaErrorString (error) );
	   return (-1);
	 }
	else if (n != STARTUP_COUNT)
	 { rciAbort (ECount, "%s received %d vs. %d bytes from robot\n", 
		     pfx, n, STARTUP_COUNT);
	   return (-1);
	 }

	rbt->notice = ACTIVATED;

	READ_BYTE(how->robotID, bp);
	READ_BYTE(how->controllerID, bp);
	READ_BYTE(how->options, bp);

	how->numAdc = 0;
	how->numDac = 0;
	how->numPioIn = 0;
	how->numPioOut = 0;
	how->numSlu = 0;
	how->integration = 0;
 	how->calibMode = 0;
	how->markFound = 0;

	bp = dataBuf + 2;

	*bp++ = SET_RATE_C;
	*bp++ = MIN(tcb->interval,255);

	n = (int)(bp - (dataBuf+2));

	dataBuf[0] = RCI_INIT_COMMANDS;
	dataBuf[1] = n;
	rbt->notice = DOING_IO;
	if ((*dev->write) (dev, dataBuf, n+2) != n+2)
  	 { errorMsg (getErrorCode(0), "%s error writing to robot\n", pfx);
	   rciAbortErrors();
	   return (-1);
	 }

	rbt->notice = ACTIVATED;

	rbt->ready = dev->testRead;
	rbt->readyArg = (int)dev;

	rbt->wakeup = crs_signal;
	rbt->wakeupArg = (int)dev;

	return (1);
}

crs_release(td, rbt, code)
RCI_DESC *td;
RCI_RBT *rbt;
int code;
{
	/* See comment in crs_input() */
	if (td->tcb->scheduleDisp == ON_SIGNAL ||
	    td->tcb->scheduleDisp == ON_CLOCK_SIM)
	 { return (puma_release (td, rbt, code));
	 }
	else
	 { return (crsplus_release (td, rbt, code));
	 }
}

crsplus_release (td, rbt, code)
RCI_DESC *td;
RCI_RBT *rbt;
int code;
{
	/* Send the release signal to the robot.
	   Send down release commands.
	   Read back information, if any.
	*/

	RCI_TCB *tcb = td->tcb;
	IO_DEVICE *dev = (IO_DEVICE*)rbt->dev;
	HOW *how = rbt->how;

	char *pfx = "rbtRelease() --";

	unsigned char dataBuf[MAX_DATABUF];
	unsigned char *bp = dataBuf + 2;
	int retval = 0;
	int n;

	/* clear how->powerOnRelease if code is not normal release request */
	
	if (code == RBT_HARDERROR)
	 { how->powerOnRelease = NO;
	 }

	if ((rbt->notice == DOING_IO) ||
	    (rbt->notice == NO_NOTIFY) ||
	    (rbt->notice == 0) ||
	    (dev->location == 0))
	 { goto releaseEnd;
	 }
	*bp++ = STOP_ALL_C;
	if (how->powerOnRelease)
	 { how->powerOnRelease = NO;
	 }
	else
	 { *bp++ = OFF_POWER_C;
	 }

	n = (int)(bp - (dataBuf+2));
	dataBuf[0] = RCI_RELEASE;
	dataBuf[1] = n;
	if ((*dev->write) (dev, dataBuf, n+2) < n+2)
	 { errorMsg (getErrorCode(0), "%s error writing to robot\n",pfx);
	   rciAbortErrors();
	   retval = -1;
	   goto releaseEnd;
	 }
releaseEnd:
	rbt->notice = 0;
	(*dev->detach)(dev);
	BZERO ((char*)rbt->chg, sizeof(CHG));
	return (retval);
}
