
//------------------------------------------------------------------
// Source Code for ATI ISA FT adapter
//
// 	This file contains sample 'C' Source code routines for interfacing the
//	ATI ISA/FT card.  It is intended for use by non-Windows applications
//	in interfacing the card.
//
//  Win 95 applications are expected to use the ATI ActiveX control
//  which provides Properties, Methods and Event object interfaces.
//	
//  Windows 3.1 (16-bit) applications are expected to use the ATI
//	Dynamic Link Library (DLL).    
//
//  This code should compile without errors using Borland C++ V4.51 or later.
//
//  Auther: Jerry Dahl, ATI Industrial Automation
//
//------------------------------------------------------------------

#ifdef X86sol_TARGET
#include "inout.h"

#define outport(p,v) 	outw(p,v)
#define inport(p)	inw(p)

#define  FALSE		0
#define  TRUE		1


unsigned addr_reg 		= 0x280;			// base I/O address of FT Card
unsigned data_reg 		= 0x282;    	// addr_reg + 2
unsigned status_reg 		= 0x284;     	// addr_reg + 4
unsigned irq_num			= 7;        	// interrupt number
unsigned data_available	= FALSE;			// flag from interrupt handler
unsigned interrupt_mode	= FALSE;     	// polling vs interrupt mode


//------------------------------------------------------------------
// (C) Copyright 1997,  by ATI Industrial Automation
//
// function:  send_command
//
// Uses global registers:
// addr_reg	PC I/O address of F/T card's address register
//	data_reg	PC I/O address of F/T card's data register
//
// Calling parameters:
//	unsigned *bptr 	= pointer to command structure
//	int numwords	= length of command
//
// Return Value: 	1 	if successful
//						0 	if abort
//
//------------------------------------------------------------------
int send_command(unsigned *bptr, int numwords)
{
	unsigned dword, command, retcode, loop ;

	//---------------------------------------
	// write command block to DPRAM at offset 0x000
	//---------------------------------------
	command = *bptr;		    	// save command word
	outport(addr_reg,0); 		// set address register to command_block
	for(;numwords;numwords--)	// write command block to dual port RAM
		{
		dword = *bptr ;
		outport(data_reg,dword) ;
		bptr ++ ;
		}
	//---------------------------------------
	// read MailBox at DPRAM 0x3FE to clear host interrupt flag
	//---------------------------------------
	outport(addr_reg,0x3FE);	     	// set mailbox address
	inport(data_reg);		   	    	// reset interrupt flag

	outport(addr_reg,0x3FE) ;	    	// set mailbox address
	outport(data_reg,command);	    	// clear acknowledgment

	//---------------------------------------
	// write command word to Mailbox at DPRAM 0x3FF to set
	// DSP interrupt flag
	//---------------------------------------
	outport(addr_reg,0x3ff) ;
	outport(data_reg,command) ;

	//---------------------------------------
	//  waits for command completion (interrupt flag) from DSP
	//---------------------------------------
	retcode = TRUE ;
	for(loop=TRUE;loop;)
		{
		dword = inport(status_reg);

		// wait for interrupt flag
		if((dword & 0x0010)==0)       		// interrupt set
			{

			// read mailbox location at 0x3FE to reset interrupt flag
			outport(addr_reg,0x3FE) ;	 	// write dpram address
			retcode = inport(data_reg) ;	 	// reset interrupt flag

			// test mailbox for command acknowledgment
			if(retcode & 0x00ff) loop=FALSE;	// test for ack
			}
		}
	return (retcode);
}

//------------------------------------------------------------------
// (C) Copyright 1997,  by ATI Industrial Automation
// function:  receive_command
// reads numwords words from ft card starting at fifo offset 0
//
// global registers:
//    addr_reg	PC I/O address of F/T card's address register
//	data_reg	PC I/O address of F/T card's data register
//
//
//  Calling parameters:
//	unsigned *bptr 	= pointer to command structure
//	int numwords	= length of command
//
//------------------------------------------------------------------
void  receive_command(unsigned *bptr, int numword)
{
	unsigned  dword;

	outport(addr_reg,0);	  // reset address
	for(;numword;numword--)
		{
		dword = inport(data_reg);
		*bptr = dword ;
		bptr ++ ;
		}
}


//-------------------------------------------------
// sensor_data command structure
//-------------------------------------------------
struct
	{
	unsigned  	cmd_rc;			// 0x0100
	unsigned  	mode;
	unsigned  	matrix[96];  		// 6 x 8 32-bit words
	unsigned		sbias[8];		// hardware bias array
	signed  		Mm[8];
	signed  		Bm[8];
	signed  		Mb[8];
	signed  		Bb[8];
	unsigned  	number_of_gages;
	unsigned  	counts_per_force;
	unsigned  	counts_per_torque;
	unsigned  	max_force;
	unsigned  	max_torque;
	unsigned  	maxtemperature;
	unsigned  	min_temperature;
	unsigned  	units;
	unsigned  	serial_number;
	} sensor_data;

//------------------------------------------------------------------
// (C) Copyright 1997,  by ATI Industrial Automation
//
// function: read_sensor_info()
//	Sends Sensor_Command to F/T card
//	Returns sensor_data
//
// global registers:
//    addr_reg	PC I/O address of F/T card's address register
//	data_reg	PC I/O address of F/T card's data register
//  Return Value: 	1 	if successful
//				0 	if abort
//
//------------------------------------------------------------------
int read_sensor_info(void)
{
	sensor_data.cmd_rc = 0x0100 ;		// sendor data command word
	sensor_data.mode   = 0x0000 ;		// mode=read
	if(!send_command(&sensor_data,2)) return(FALSE);
	receive_command(&sensor_data,sizeof(sensor_data)) ;
	return (TRUE);
}


//------------------------------------------------------------------
// (C) Copyright 1997,  by ATI Industrial Automation
//
// read_ft_data()
//
// global registers:
//    addr_reg	PC I/O address of F/T card's address register
//	data_reg	PC I/O address of F/T card's data register
//	interrupt_mode	0=use polling node
//				1=use interrupt mode
//
// polls data_reg[0] for new data
//------------------------------------------------------------------
void read_ft_data(void)
	{
	int  	loop, i;
	int  	dword ;
	int	force_counts[3], torque_counts[3];

	//-----------------------------------
	// read sensor info to get
	// counts_per_force, counts_per_torque, units
	//----------------------------------
	read_sensor_info();

	printf ("units=%d\n", sensor_data.units);

	for(loop=TRUE;loop;)
		{
		if(interrupt_mode)
			{
			//-----------------------------------
			// interrupt handler sets data_available to
			// indicate that new data is availble.
			//-----------------------------------
			if(!data_available) continue;
			outport(addr_reg,0x100);	// write ram starting addr
			dword = inport(data_reg);	// read first word[0]
			}
		else
			{
			//-----------------------------------
			// polling mode, test for new data
			//-----------------------------------
			outport(addr_reg,0x100);	// write ram starting addr
			dword = inport(data_reg);	      // read first word[0]
			if(!dword) continue;
			}

		//-----------------------------------------
		// new data available.
		// first check monitoring conditions
		//-----------------------------------------
		dword = inport(data_reg);      // read 2nd word=monitor cond
		if(dword)
			{
			outport(addr_reg,0x101);  // monitor condition
			outport(data_reg,0x00);	  // clear monitor condition
			//-----------------------------------
			// mon_cond_status
			// 1 = Alarm state
			// 2 = Out of Alarm State
			// 3 = Expired
			//------------------------------------
			}

		//-----------------------------------------
		// get the new F/T data in counts
		// to get actual forces and torque, the application
		// should divide the counts by the
		// counts_per_force and counts_per_torque
		// conversion factors
		//-----------------------------------------
		for(i=0;i<3;i++)
		   { force_counts[i] = inport(data_reg);
		     printf ("%10.4f ", 
			((signed short)force_counts[i])/
			 (double)sensor_data.counts_per_force);
		   }
		for(i=0;i<3;i++)
		   { torque_counts[i] = inport(data_reg);
		     printf ("%10.4f ", 
			((signed short)torque_counts[i])/
			 (double)sensor_data.counts_per_torque);
		   }
		printf ("\n");
		sleep(1);

		// acknowledge data
		outport(addr_reg,0x100);   // write ram starting addr
		outport(data_reg,0x00);	   // acknowledge data
		}

	}
#endif

