#ifndef _CARTBLEND_H_
#define _CARTBLEND_H_

/*-=-=-=-=-=-=-=-=-=-=-=- RCCL/RCI Notice of Copyright -=-=-=-=-=-=-=-=-=-=-*
 *									    *
 *	Copyright (C) 1991 by John E. Lloyd.				    *
 *									    *
 *	Information on copyright permissions, as well as a complete	    *
 *	disclaimer of warranty, is given in the file COPYRIGHT.		    *
 *									    *
 *-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=*/

/*---------------------------------------------------------------------------
 *
 * DESCRIPTION:	Defines the structure which holds the coefficients for
 * 		blending together two Cartesian path segments. 
 * 
 * AUTHOR/DATE:	John Lloyd, October 1986
 *
 * MODIFICATION HISTORY:
 *
 *--------------------------------------------------------------------------*/

typedef struct {
	TRSF R1_0;		/* 1st rotation, path 1 */
	TRSF R2_0;		/* 1st rotation, path 2 */
	QTRN q1_0;		/* quaternion rep. of R1_0 */
	QTRN q2_0;		/* quaternion rep. of R2_0 */
	QTRN q0;		/* quaternion rep. of inv(R2_0)*R1_0 */
	VECT vAxis;		/* rotation axis for initial velocity */
	VECT cAxis;		/* rotation axis for velocity correction */
	float vTheta;		/* initial velocity displacement angle */
	float cTheta;		/* velocity correction angle */
	VECT P0;		/* Coefficients for 5th order Hermite */
	VECT P1;		/*    translation correction */
	VECT P3;		/*    ...    */
	VECT P4;
	VECT P5;  
} CARTESIAN_BLEND_PARAMS;

#endif
