/* These functions return JLS and KYN structures that can be accessed
   ny an RCI task. */

#include "../h/rciTask.h"
#include "../h/errorCodes.h"
#include "../h/errorStack.h"
#include "../rci/loadRobot.h"
#include "../h/robot_kynvar.h"

extern RCI_RBT *rciRbtDescList;
extern int rciTaskId;

extern ROBOT_DESC *createRobotDesc();

/* ----- Error Messages. Defined externally to save space ----- */

static char *robotNotSetupMsg = 
	"%s -- rciJlsKynInit() not yet called for robot %s\n";

JLS *rciGetJls(name)
const char *name;
{
	char *fxnName = "rciGetJls()";
	RCI_RBT *rbt;

	/* search the rbtDescList for name */
	for (rbt=rciRbtDescList; rbt; rbt=rbt->next)
	 { if (strcmp(rbt->name, name) == 0)
	    { break;
	    }
	 }
	if (rbt)
	 { return (rbt->jls);
	 }
	else
	 { errorMsg (ENotDefined, robotNotSetupMsg, name);
	   return (NULL);
	 }
}

KYN *rciGetKyn(name, sizep)
const char *name;
int *sizep;
{
	char *fxnName = "rciGetKyn()";
	RCI_RBT *rbt;

	/* search the rbtDescList for name */
	for (rbt=rciRbtDescList; rbt; rbt=rbt->next)
	 { if (strcmp(rbt->name, name) == 0)
	    { break;
	    }
	 }
	if (rbt)
	 { if (sizep)
	    { *sizep = rbt->kynSize;
	    }
	   return (rbt->kyn);
	 }
	else
	 { errorMsg (ENotDefined, robotNotSetupMsg, name);
	   return (NULL);
	 }
}
