/* Start up an RCI task and time it's execution */

#include <stdio.h>

#include "../h/rci.h"
#include "../h/cdefs.h"
#include "../h/ctree.h"
#include "../h/signalFxns.h"

char **argvec;
int interval = 10;
RCI_DESC *td1;

extern doControl();

char key_buf [256];
int *signalReq;

SIGHAN_RETURN onUsr1(sig)
int sig;
{
	printf ("\nSignal USR1 received\n\n");
	exit (1);
}

main(argc, argv)
int argc;
char **argv;
{
	char *keytree;
	TimingLog *tlog;
	TimingLog tbuf;

	argvec = argv;
	argc--; argv++;
	while (argc > 0)
	 { if (ARGUMENT ("-i"))
	    { if (argc == 1)
	       { goto usage;
	       }
	      argc--; argv++;
	      interval = atoi(*argv);
	    }
	   else
	    { goto usage;
	      break;
	    }
	   argc--; argv++;
	   continue;
usage:
	   fprintf (stderr, "Usage: %s [-i <interval>]\n", argvec[0]);
	   exit (1);
	 }

	signal (SIGUSR1, onUsr1);

	rciSetOptions (EXIT_ON_ERROR);
	td1 = rciCreate ("arb1", 0x1);
	rciRate (td1, interval, 0);
	rciScheduling (td1, ON_CLOCK, 0);

	tlog = rciTimingLogInit (td1);

	signalReq = (int*)allocMem(NULL, sizeof(int), UNDEF);
	rciControlFxns (td1, NULLF, doControl, NULLF, (int)signalReq);

	rciStart (~0);

	keytree = tree_match_parse("					\
(	quit								\
	show								\
	signal								\
	abort								\
)",0);
	if (keytree == 0)
	 { printf ("ERROR/Parse error at %.30s\n", tree_match_err_at());
	   exit (-1);
	 }
	printf ("Type 'show' to see current timing results\n");
	tree_match_debug_set(1);
	printf ("pid=%d\n", getpid());
	do
 	 { 
	   tree_match (keytree, "TimeTest> ", key_buf);

	   if (COMMAND("show"))
	    { printf ("control running at %d msec.\n", interval);
	      printf (
		"  Cnt    average  > .5   > 1   > 2   > 5  > 10   maximum\n");
	       /*XXXXXXX   X.XXXX  XXXX  XXXX  XXXX  XXXX  XXXX  XXX.XXXX */
	      readMem (tlog, &tbuf);

	      printf ("%7d   %6.4f  %4d  %4d  %4d  %4d  %4d  %8.4f\n",
		       tbuf.cycleCnt, 
		       tbuf.avgErr, 
		       tbuf.exceedCnt[TLOG_OVER_HALF],
		       tbuf.exceedCnt[TLOG_OVER_1],
		       tbuf.exceedCnt[TLOG_OVER_2],
		       tbuf.exceedCnt[TLOG_OVER_5],
		       tbuf.exceedCnt[TLOG_OVER_10],
		       tbuf.maxErr);
	    }
	   else if (COMMAND("abort"))
	    { if (signalReq)
	       { *signalReq = -1;
	       }
	    }
	   else if (COMMAND("signal"))
	    { if (signalReq)
	       { *signalReq = 1;
	       }
	    }
	 }
	while (!(COMMAND("quit")));
}
