#ifdef GL_GRAPHICS

#define DEBUG	1

#define DOUBLE_WIN	1

#define CTRLWINMARGIN	15

#include <stdio.h>
#include <errno.h>
#include <math.h>
#include <gl/gl.h>
#include <gl/qcontrol.h>
#include <device.h>

#include <sys/types.h>
#include <sys/time.h>

#include "../h/stringDefs.h"
#include "../h/cdefs.h"
#include "../h/timeval.h"
#include "../h/grf4.h"
#include "grf4Internal.h"
#include "grf4GlObject.h"

TRSF* glTransPtr;
int glPerspectiveMode;
float glNearPlane;

extern int grf4GeneralFlags; 

static TRSF Trans;

static int bgColor;
static int fgColor;

static float setLeft;
static float setRight;
static float setBottom;
static float setTop;
static float setNear;
static float setFar;

static int screenCoordsSet = 0;
static int useOrtho2 = 1;	/* default used by grf4Glwf.c */
static int useZbuffer = 0;
static int savedZbufferState;

static Matrix identMatrix = { 1.0, 0.0, 0.0, 0.0,
			      0.0, 1.0, 0.0, 0.0,
			      0.0, 0.0, 1.0, 0.0,
			      0.0, 0.0, 0.0, 1.0 };

static unsigned long getcpack();
static void glOrtho();
static void glWindow();
static void glLoadTrans();
static void setScreenCoords();
static void clearScreenCoords();
static do_lmdef();

static void doDrawButton();
static void glPxlFxn();
static void glDraw();
static void glMove();
static void glDraw2();
static void glMove2();
static void glClear();

static void decompose();
static void doDefMaterial ();
static void *glDefMaterial();
static void *glDefMaterialD();
static void *glDefLight();
static void glSetLight();
static void glResetLights();
static long glCapacity();

static void glGetWindowSize();
static void glSetWindowExtent();
static void glDrawPanel();
static void glPutText();
static void glGetTextExtent();
static void glDrawButton();
static void glGetButtonExtent();
static int glGetEventFd();
static int glEnableEvent();
static int glDisableEvent();
static int glPollEvent();
static int glReadEvent();
static void glInitDeviceMaps();

typedef struct {
	long id;	/* ID returned from winopen() */
	long x, y;	/* position of the window */
	long w, h;	/* height and width */
	int active;	/* true if the window is active */
	int eventMask;	/* events enabled for this window */
} WinDesc;

static WinDesc wdesc[NUMWINS];
static eventRef[GRF_MAX_EVENT_TYPES+1];	/* event enable reference count */

#define MAIN	wdesc[MAINWIN]
#define CTRL	wdesc[CTRLWIN]

static unsigned long getcpack()
{
	short r,g,b;
	gRGBcolor(&r,&g,&b);
	return (RGBPACK(r,g,b));
}

static void glOrtho(l,r,b,t,n,f)
float l,r,b,t,n,f;
{
	setLeft = l;
	setRight = r;
	setBottom = b;
	setTop = t;
	setNear = n;
	setFar = f;

	glPerspectiveMode = 0;
	glNearPlane = n;

	ortho (l,r,b,t,n,f);
}
	
static void glWindow(l,r,b,t,n,f)
float l,r,t,b,n,f;
{
	setLeft = l;
	setRight = r;
	setBottom = b;
	setTop = t;
	setNear = n;
	setFar = f;

	glPerspectiveMode = 1;
	glNearPlane = n;

	window (l,r,b,t,n,f);
}

#define LIGHT_GREY	RGBPACK(160,160,160)
#define DARK_GREY	RGBPACK(100,100,100)
#define DEEP_GREY	RGBPACK(60,60,60)
#undef WHITE
#define WHITE		RGBPACK(255,255,255)
#undef BLACK
#define BLACK		RGBPACK(0,0,0)

static void glDrawPanel()
{
	if (CTRL.active)
	 { winset (CTRL.id);
	   cpack (LIGHT_GREY);
	   clear();
	   winset (MAIN.id);
	 }
}

static float light_1[] = { POSITION, 1, 1, 1, 0,
		    LMNULL };

static float light_2[] = { POSITION, -1, -1, -1, 0,
			   LCOLOR, .2, .2, .2,
		    LMNULL };

static float material_1[] = { AMBIENT, .2, .2, .2,
		       DIFFUSE, 1, 1, 1,
		       EMISSION, .3, 0, 0,
		       SHININESS, 0,
		       LMNULL };

static float lmodel_1[] = { AMBIENT, 0, 0, 0,
		       LMNULL };

#define MIN_WINW	21
#define MIN_WINH	21

#ifdef _AIX

/* XXX XXX XXX AIX does not seem to support qgetfd(). So we fake it. XXX */

#include <unistd.h>
#include <fcntl.h>

typedef struct qfdDesc {
	struct qfdDesc *next;
	int fd;
} QfdDesc;

static QfdDesc *qfdList;
static int numWinFd;
static char isopen[FD_SETSIZE];

glAixQgetfdHack1()
{
	QfdDesc *qfd, *qfdNext;
	int k;

	BZERO (isopen, FD_SETSIZE);
	if (gfdList)
	 { for (qfd=qfdList; qfd; qfd=qfdNext)
	    { qfdNext = qfd->next;
	      free ((void*)qfd);
	    }
	   qfdList = NULL;
	 }
	for (k=0; k<FD_SETSIZE; k++)
	 { isopen[k] = (fcntl(k,F_GETFL,0) >= 0);
	 } 
}

glAixQgetfdHack2()
{
	int k;

	for (k=0; k<FD_SETSIZE; k++)
	 { if (!isopen[k])
	    { int flags;
	      flags = fcntl(k,F_GETFL,0);
	      if (flags >= 0 && !(flags & O_WRONLY))
	       { QfdDesc *qfd;

		 qfd = (QfdDesc*)malloc(sizeof(QfdDesc));
		 qfd->next = qfdList;
		 qfdList = qfd;
		 qfd->fd = k;
		 numWinFd++;
	       }
	    }
	 } 
	if (!qfdList)
	 { grf4Fatal ("grf4Init() -- winFd not detected\n");
	 }
}

glAixQgetfd()
{
	if (numWinFd == 1)
	 { return(qfdList->fd);		/* only choice */
	 }
	else
	 { return(qfdList->next->fd);	/* skip first desc. */
	 }
}

#endif

grf4GlInit (name, fgcolor, bgcolor, placing, flags)
char *name;
int fgcolor;
int bgcolor;
WinPlaceDesc *placing;
int flags;
{
	int doResize;
	int w, h;
	int minW, minH;

	/* set up the null window with info appropriate to the screen */

	wdesc[NULLWIN].x = 0;
	wdesc[NULLWIN].y = 0;
	wdesc[NULLWIN].w = getgdesc(GD_XPMAX);
	wdesc[NULLWIN].h = getgdesc(GD_YPMAX);

#ifdef _AIX
	glAixQgetfdHack1();
#endif
#ifndef _AIX
	foreground();
#endif	
	if (placing->placed)
	 { w = MAX(placing->w,placing->minW);
	   h = MAX(placing->h,placing->minH);
	   prefposition (placing->x, placing->x+w-1,
			 placing->y, placing->y+h-1);
	 }
#ifdef _AIX 
/* XXX XXX XXX AIX doesn't seem to support interactive placement! */
	else
	 { w = MAX(372,placing->minW);
	   h = MAX(256,placing->minH);
	   prefposition (256,256+w-1,256,256+h-1);
	 }
#endif
	if ((MAIN.id=winopen(name)) == -1)
	 { return (-1);
	 }
	MAIN.active = 1;

	buttonInit (&CTRL.w, &CTRL.h);
	getorigin (&MAIN.x, &MAIN.y);
	getsize (&MAIN.w, &MAIN.h);
	CTRL.active = (CTRL.w > 0);

	minW = MAX(MIN_WINW,placing->minW);
	minH = MAX(MIN_WINH,placing->minH);

	doResize = 0;
	if (MAIN.w < minW)
	 { MAIN.w = minW;
	   doResize = 1;
	 }
	if (MAIN.h < minH)
	 { MAIN.h = minH;
	   doResize = 1;
	 }
#ifndef _AIX
  /* XXX XXX XXX winposition on AIX is broken! Can't resize window!! */
	if (doResize)
	 { winposition (MAIN.x, MAIN.x+MAIN.w-1, MAIN.y, MAIN.y+MAIN.h-1);
	 }
#endif
	minsize (minW, minH);
	winconstraints();

	RGBmode();
	if (!(flags & GRF_DOUBLEBUF_OFF))
	 { doublebuffer();
	 }
	gconfig();
	
	mmode(MVIEWING);

	bgColor = bgcolor;
	fgColor = fgcolor;
	useOrtho2 = 0;		/* for set,clearScreenCoords */

	/* setup for receiving events */

	qdevice(REDRAW);
	qdevice(PIECECHANGE);
#if 0
	if (CTRL.active)
	 { qdevice(CURSORX);
	   qdevice(CURSORY);
	   qdevice(LEFTMOUSE);
	   qdevice(MIDDLEMOUSE);
	   qdevice(RIGHTMOUSE);
	 }
#endif
	/* turn on the z-buffer, and set up default lighting */

	if (!(flags & GRF_2DONLY))
	 { zbuffer(TRUE);
	   useZbuffer = 1;
	 }
	zfunction(ZF_LEQUAL);
	if (!(flags & GRF_2DONLY))
	 { do_lmdef(DEFLMODEL,1,lmodel_1);
	   lmbind(LMODEL,1);
	 }

	loadmatrix(identMatrix);
	
	viewport (0, MAIN.w-1, 0, MAIN.h-1);
	glOrtho (-0.5, MAIN.w-0.5, -0.5, MAIN.h-0.5, -1.0, 1.0);

	if (CTRL.active)
	 { long cx, cy;

	   foreground ();
	   CTRL.x = MAIN.x+CTRLWINMARGIN+MAIN.w;
	   CTRL.y = MAIN.y+(MAIN.h-CTRL.h);
	   prefposition (CTRL.x, CTRL.x+CTRL.w-1, CTRL.y, CTRL.y+CTRL.h-1);
	   if ((CTRL.id=winopen("control")) == -1)
	    { return (-1);
	    }
	   winset(CTRL.id);
	   RGBmode();
	   gconfig();
	   mmode(MVIEWING);
	   minsize (CTRL.w, CTRL.h);
	   maxsize (CTRL.w, CTRL.h);
	   winconstraints ();
	   viewport (0, CTRL.w-1, 0, CTRL.h-1);
	   ortho2 (-0.5,CTRL.w-0.5,-0.5,CTRL.h-0.5);
	   getorigin (&cx, &cy);
	   if (cx != CTRL.x || cy != CTRL.x+CTRL.w-1)
	    { winposition (CTRL.x, CTRL.x+CTRL.w-1, CTRL.y, CTRL.y+CTRL.h-1);
	    }
	   winset(MAIN.id);
	 }
	glDrawPanel();

#ifdef _AIX
	glAixQgetfdHack2();
#endif
	glInitDeviceMaps();
	return (1);
}

/* TODO: look after AIX incompats */

/* TODO: put in flag so panel does not stick to the main window */

static do_lmdef(type,index,props)
int type;
int index;
float *props;
{
	int np;

 	if (props)
	 { for (np=0;props[np]!=LMNULL;np++) ;
	   if (np > 0) np ++;
	 }
	else
	 { np = 0;
	 }
	lmdef(type,index,np,props);
}

static void glGetWindowSize (wp,hp, directionOfYp)
int *wp, *hp, *directionOfYp;
{
	if (wp) *wp = MAIN.w;
	if (hp) *hp = MAIN.h;
	if (directionOfYp) *directionOfYp = 1;
}

static void glPutText (str, x, y)
char *str;
int x, y;
{
	cmov2i (x, y);
	charstr (str);
}

static void glGetTextExtent (xp, yp, wp, hp, str)
int *xp, *yp, *wp, *hp;
char *str;
{
	int height;
	int descent;

	descent = getdescender();
	height = getheight();

	if (xp) *xp = 0;
#ifdef _AIX
/* AIX didn't seem to get the character position right */
	if (yp) *yp = descent/2;
#else
	if (yp) *yp = descent;
#endif
	if (wp) *wp = strwidth(str);
	if (hp) *hp = (height-descent-1);
}

#define BUTTON1			1
#define BUTTON_BORDER_WIDTH	3

static void glGetButtonExtent(b)
BUTTON_DESC *b;
{
	int x, y, w, h;
	int margin;

	glGetTextExtent (&x, &y, &w, &h, b->name);
	margin = 2 + BUTTON_BORDER_WIDTH;
	b->coffX = margin;
	b->coffY = margin + x;
	b->coffY = margin + y;
	b->minw = 2*margin + w;
	b->minh = 2*margin + h;
}

static void glDrawButton(b)
BUTTON_DESC *b;
{
	int x1,x2;
	int y1,y2;

	int baseColor;
	int charColor;
	int topEdgeColor;
	int bottomEdgeColor;

	int cbaseX;
	int cbaseY;

	winset (CTRL.id);
	x1 = b->x;
	y1 = b->y;
	x2 = x1 + b->w - 1;
	y2 = y1 + b->h - 1;

	cpack(WHITE);	/* white */

	cbaseX = b->x + b->coffX + (b->w-b->minw)/2;
	cbaseY = b->y + b->coffY + (b->h-b->minh)/2;

	if (b->flags & BF_NOTBOXED)
	 { cpack (LIGHT_GREY);
	   rectfi (x1,y1,x2,y2);
	   if (b->flags & BF_ILLUMINATED)
	    { cpack (WHITE);
	    }
	   else
	    { cpack (BLACK);
	    }
	   cmov2i (cbaseX, cbaseY);
	   charstr (b->name);
	 }
	else
	 { cpack (BLACK);
	   recti (x1,y1,x2,y2);		/* outer black box */
	   if (b->flags & BF_ILLUMINATED)
	    { baseColor = LIGHT_GREY;
	      charColor = WHITE;
	      topEdgeColor = WHITE;
	      bottomEdgeColor = DARK_GREY;
	    }
	   else
	    { baseColor = DARK_GREY;
	      charColor = BLACK;
	      topEdgeColor = DEEP_GREY;
	      bottomEdgeColor = LIGHT_GREY;
	    }
	   cpack (topEdgeColor);
	   recti (x1+1,y1+1,x2-1,y2-1);
	   cpack (bottomEdgeColor);
	   move2i (x1+1,y1+1);
	   draw2i (x2-1,y1+1);
	   cpack (baseColor);
	   rectfi (x1+2,y1+2,x2-2,y2-2);
	   cpack (charColor);
	   cmov2i (cbaseX, cbaseY);
	   charstr (b->name);
	 }
	winset(MAIN.id);
}

static void glPxlFxn(x1,y1,x2,y2)
int x1,y1,x2,y2;
{
	grf4Fatal ("Internal software error: call to pixel level function\n");
}

static void glDraw(x,y,z)
double x,y,z;
{
	draw(x,y,z);
}

static void glMove(x,y,z)
double x,y,z;
{
	move(x,y,z);
}

static void glPoint(x,y,z)
double x,y,z;
{
	pnt(x,y,z);
}

static void glDraw2(x,y)
double x,y;
{
	draw2(x,y);
}

static void glMove2(x,y)
double x,y;
{
	move2(x,y);
}

static void glPoint2(x,y)
double x,y;
{
	pnt2(x,y);
}

static void glClear()
{
	unsigned long savec;

	savec = getcpack();
	cpack(bgColor);
	clear();
	if (useZbuffer)
	 { zclear();
	 }
	cpack(savec);
}

static void glLoadTrans(T)
TRSF *T;
{
	Matrix m;

	m[0][0] = T->n.x;
	m[0][1] = T->n.y;
	m[0][2] = T->n.z;
	m[0][3] = 0.0;

	m[1][0] = T->o.x;
	m[1][1] = T->o.y;
	m[1][2] = T->o.z;
	m[1][3] = 0.0;

	m[2][0] = T->a.x;
	m[2][1] = T->a.y;
	m[2][2] = T->a.z;
	m[2][3] = 0.0;

	m[3][0] = T->p.x;
	m[3][1] = T->p.y;
	m[3][2] = T->p.z;
	m[3][3] = 1.0;

	Trans = *T;
	glTransPtr = &Trans;
	loadmatrix(m);
}


/* Lighting and material functions */

#define BLUE_PART(m)	(((m)&0xff0000)>>16)
#define GREEN_PART(m)	(((m)&0xff00)>>8)
#define RED_PART(m)	(((m)&0xff))

static void decompose(v,m)
float *v;
unsigned long m;
{
	v[0] = RED_PART(m)/255.0;
	v[1] = GREEN_PART(m)/255.0;
	v[2] = BLUE_PART(m)/255.0;
}

static void doDefMaterial (idx, d, a, e, s, shine, alpha)
int idx;
unsigned long d, a, e, s;
float shine, alpha;
{
	float proplist[21];

	proplist[0] = DIFFUSE;
	decompose(&proplist[1], d);
	proplist[4] = AMBIENT;
	decompose(&proplist[5], a);
	proplist[8] = EMISSION;
	decompose(&proplist[9], e);
	proplist[12] = SPECULAR;
	decompose(&proplist[13], s);
	proplist[16] = SHININESS;
	proplist[17] = shine;
	proplist[18] = ALPHA;
	proplist[19] = alpha;
	proplist[20] = LMNULL;
	
	lmdef(DEFMATERIAL, idx, ARRAY_SIZE(proplist), proplist);
}

static int materialIdx = 0;

static void *glDefMaterial(diffuse, ambient, emission, specular, shine, alpha)
unsigned long diffuse, ambient, emission, specular;
float shine, alpha;
{
	int *ip;

	ip = ALLOC(int);
	*ip = ++materialIdx;
	doDefMaterial (*ip, diffuse, ambient, emission, 
			    specular, shine, alpha);
	return ((void*)ip);
}

static void *glDefMaterialD(md)
MATERIAL_DEF *md;
{
	int *ip;

	ip = ALLOC(int);
	*ip = ++materialIdx;
	doDefMaterial (*ip, md->diffuse, md->ambient, md->emission, 
			    md->specular, md->shine, md->alpha);
	return ((void*)ip);
}

static void glSetMaterial(ptr)
void *ptr;
{
	if (!(grf4GeneralFlags & GRF_2DONLY))
	 { lmbind(MATERIAL, *(int*)ptr);
	 }
}

struct lampDesc {
	int code;
	int idx;
  };

static struct lampDesc lampPool[] = {
	LIGHT0, 0, LIGHT1, 0, LIGHT2, 0, LIGHT3, 0, 
	LIGHT4, 0, LIGHT5, 0, LIGHT6, 0, LIGHT7, 0,
};

static int lightIdx = 0;

static void* glDefLight(color, ambient, x,y,z,w)
unsigned long color, ambient;
float x,y,z,w;
{
	int *ip;
	float proplist[14];

	ip = ALLOC(int);
	*ip = ++lightIdx;

	proplist[0] = LCOLOR;
	decompose(&proplist[1], color);
	proplist[4] = AMBIENT;	
	decompose(&proplist[5], ambient);
	proplist[8] = POSITION;
	proplist[9] = x;
	proplist[10] = y;
	proplist[11] = z;
	proplist[12] = w;
	proplist[13] = LMNULL;

	lmdef(DEFLIGHT, *ip, ARRAY_SIZE(proplist), proplist);

	return ((void*)ip);
}

static void glSetLight(n, ptr)
int n;
void *ptr;
{
	if (n >= ARRAY_SIZE(lampPool))
	 { fprintf (stderr, "grf4SetLight(): don't have that many lamps\n");
	   return;
	 }
	if (n < 0)
	 { fprintf (stderr, "grf4SetLight(): negative lamp index\n");
	   return;
	 }
	if (ptr)
	 { lmbind (lampPool[n].code, (*(int*)ptr));
	   lampPool[n].idx = *(int*)ptr;
	 }
	else
	 { lmbind (lampPool[n].code, 0);
	   lampPool[n].idx = 0;
	 }
}	   

static void glResetLights()
{
	int i;

	for (i=0; i<ARRAY_SIZE(lampPool); i++)
	 { if (lampPool[i].idx)
	    { lmbind (lampPool[i].code, lampPool[i].idx);
	    }
	 }
}

static long glCapacity()
{
	return (GRF_HARDWARE_A | GRF_COLOR | GRF_LIGHTING);
}

grf4GlSetupFxns (ftab)
struct grf4fxnTable *ftab;
{
	ftab->capacity = glCapacity;
	ftab->getWindowSize = glGetWindowSize;
	ftab->ortho = glOrtho;
	ftab->window = glWindow;
	ftab->loadTrans = glLoadTrans;
	ftab->color = cpack;
	ftab->clear = glClear;
	ftab->flush = swapbuffers;
	ftab->move = glMove;
	ftab->draw = glDraw;
	ftab->point = glPoint;
	ftab->move2 = glMove2;
	ftab->draw2 = glDraw2;
	ftab->point2 = glPoint2;
	ftab->drawPrism = glDrawPrism;
	ftab->drawSphere = glDrawSphere;
	ftab->drawPolygon = glDrawPolygon;
	ftab->drawLineSet = glDrawLineSet;
	ftab->genDisc = glGenDisc;
	ftab->genCone = glGenCone;
	ftab->genPrism = glGenPrism;
	ftab->genSphere = glGenSphere;
	ftab->genPolygon = glGenPolygon;
	ftab->genRegPolygon = glGenRegPolygon;
	ftab->genLineSet = glGenLineSet;
	ftab->setDepth = glSetDepth;

	ftab->putText = glPutText;
	ftab->getTextExtent = glGetTextExtent;
	ftab->drawButton = glDrawButton;
	ftab->getButtonExtent = glGetButtonExtent;
	
	ftab->pxlDraw = glPxlFxn;
	ftab->pxlClear = glPxlFxn;
	ftab->pxlFlush = glPxlFxn;

	ftab->defLight = glDefLight;
	ftab->defMaterial = glDefMaterial;
	ftab->defMaterialD = glDefMaterialD;
	ftab->setLight = glSetLight;
	ftab->resetLights = glResetLights;
	ftab->setMaterial = glSetMaterial;

	ftab->getEventFd = glGetEventFd;
	ftab->enableEvent = glEnableEvent;
	ftab->disableEvent = glDisableEvent;
	ftab->pollEvent = glPollEvent;
	ftab->readEvent = glReadEvent;

	return (1);
}

/* -=-=-=-= Special Functions for the wireframe GL interface -=-=-=-= */

grf4GlwfInit(name, fgcolor, bgcolor, placing, flags)
char *name;
int fgcolor;
int bgcolor;
WinPlaceDesc *placing;
int flags;
{
	int doResize;
	int w, h;
	int minW, minH;

	/* set up information in the NULL window */

	wdesc[NULLWIN].x = 0;
	wdesc[NULLWIN].y = 0;
	wdesc[NULLWIN].w = getgdesc(GD_XPMAX);
	wdesc[NULLWIN].h = getgdesc(GD_YPMAX);

#ifdef _AIX
	glAixQgetfdHack1();
#endif
#ifndef _AIX
	foreground();
#endif	
	if (placing->placed)
	 { w = MAX(placing->w,placing->minW);
	   h = MAX(placing->h,placing->minH);
	   prefposition (placing->x, placing->x+w-1,
			 placing->y, placing->y+h-1);
	 }
#ifdef _AIX 
/* XXX XXX XXX AIX doesn't seem to support interactive placement! */
	else
	 { w = MAX(372,placing->minW);
	   h = MAX(256,placing->minH);
	   prefposition (256,256+w-1,256,256+h-1);
	 }
#endif
	if ((MAIN.id=winopen(name)) == -1)
	 { return (-1);
	 }
	MAIN.active = 1;

	buttonInit (&CTRL.w, &CTRL.h);
	getorigin (&MAIN.x, &MAIN.y);
	getsize (&MAIN.w, &MAIN.h);
	CTRL.active = (CTRL.w > 0);

	minW = MAX(MIN_WINW,placing->minW);
	minH = MAX(MIN_WINH,placing->minH);

	doResize = 0;
	if (MAIN.w < minW)
	 { MAIN.w = minW;
	   doResize = 1;
	 }
	if (MAIN.h < minH)
	 { MAIN.h = minH;
	   doResize = 1;
	 }
#ifndef _AIX
  /* XXX XXX XXX winposition on AIX is broken! Can't resize window!! */
	if (doResize)
	 { winposition (MAIN.x, MAIN.x+MAIN.w-1, MAIN.y, MAIN.y+MAIN.h-1);
	 }
#endif
	minsize (minW, minH);
	winconstraints();

	RGBmode();
	if (!(flags & GRF_DOUBLEBUF_OFF))
	 { doublebuffer();
	 }
	gconfig();
	
	bgColor = bgcolor;
	fgColor = fgcolor;

	/* setup for receiving events */

	qdevice(REDRAW);
	qdevice(PIECECHANGE);
#if 0
	if (CTRL.active)
	 { qdevice(CURSORX);
	   qdevice(CURSORY);
	   qdevice(LEFTMOUSE);
	   qdevice(MIDDLEMOUSE);
	   qdevice(RIGHTMOUSE);
	 }
#endif
	loadmatrix(identMatrix);

	viewport (0, MAIN.w-1, 0, MAIN.h-1);
	ortho2 (-0.5,MAIN.w-0.5,-0.5,MAIN.h-0.5);
	if (CTRL.active)
	 { long cx, cy;

	   foreground ();
	   CTRL.x = MAIN.x+CTRLWINMARGIN+MAIN.w;
	   CTRL.y = MAIN.y+(MAIN.h-CTRL.h);
	   prefposition (CTRL.x, CTRL.x+CTRL.w-1, CTRL.y, CTRL.y+CTRL.h-1);
	   if ((CTRL.id=winopen("panel")) == -1)
	    { return (-1);
	    }
	   winset(CTRL.id);
	   RGBmode();
	   gconfig();
	   mmode(MVIEWING);
	   minsize (CTRL.w, CTRL.h);
	   maxsize (CTRL.w, CTRL.h);
	   winconstraints ();
	   viewport (0, CTRL.w-1, 0, CTRL.h-1);
	   ortho2 (-0.5,CTRL.w-0.5,-0.5,CTRL.h-0.5);
	   getorigin (&cx, &cy);
	   if (cx != CTRL.x || cy != CTRL.x+CTRL.w-1)
	    { winposition (CTRL.x, CTRL.x+CTRL.w-1, CTRL.y, CTRL.y+CTRL.h-1);
	    }
	   winset(MAIN.id);
	 }
	glDrawPanel();

#ifdef _AIX
	glAixQgetfdHack2();
#endif
	glInitDeviceMaps();
	cpack(fgColor);
	return (1);
}

static void glwfClear(x,y,w,h)
int x, y, w, h;
{
	unsigned long savec;
	savec = getcpack();
	cpack(bgColor);
	clear();
	cpack(savec);
}

static void glwfColor(color)
int color;
{
#if  0
	fgColor = color;
	cpack(color);
#endif
}

static void glwfFlush(x,y,w,h)
int x,y,w,h;
{
	swapbuffers();
}
	 
static void glwfDraw(x1,y1,x2,y2)
int x1, y1, x2, y2;
{
	move2 (x1, y1);
	draw2 (x2, y2);
}

static void glwfPoint (x1, y1)
int x1, y1;
{
	pnt2 (x1, y1);
}

grf4GlwfSetupFxns (ftab)
struct grf4fxnTable *ftab;
{
	ftab->getWindowSize = glGetWindowSize;
	ftab->color = glwfColor;

	ftab->drawButton = glDrawButton;
	ftab->getButtonExtent = glGetButtonExtent;

	ftab->pxlDraw = glwfDraw;
	ftab->pxlPoint = glwfPoint;
	ftab->pxlFlush = glwfFlush;
	ftab->pxlClear = glwfClear;

	ftab->getEventFd = glGetEventFd;
	ftab->enableEvent = glEnableEvent;
	ftab->disableEvent = glDisableEvent;
	ftab->pollEvent = glPollEvent;
	ftab->readEvent = glReadEvent;

	return (1);
}

/* -=-=-=-= Event handling stuff -=-=-=-=-=-= */

static AlistElem alistBase[] = {
	GRF_KB_A,		AKEY,
	GRF_KB_B,		BKEY,
	GRF_KB_C,		CKEY,
	GRF_KB_D,		DKEY,
	GRF_KB_E,		EKEY,
	GRF_KB_F,		FKEY,
	GRF_KB_G,		GKEY,
	GRF_KB_H,		HKEY,
	GRF_KB_I,		IKEY,
	GRF_KB_J,		JKEY,
	GRF_KB_K,		KKEY,
	GRF_KB_L,		LKEY,
	GRF_KB_M,		MKEY,
	GRF_KB_N,		NKEY,
	GRF_KB_O,		OKEY,
	GRF_KB_P,		PKEY,
	GRF_KB_Q,		QKEY,
	GRF_KB_R,		RKEY,
	GRF_KB_S,		SKEY,
	GRF_KB_T,		TKEY,
	GRF_KB_U,		UKEY,
	GRF_KB_V,		VKEY,
	GRF_KB_W,		WKEY,
	GRF_KB_X,		XKEY,
	GRF_KB_Y,		YKEY,
	GRF_KB_Z,		ZKEY,
	GRF_KB_0,		ZEROKEY,
	GRF_KB_1,		ONEKEY,
	GRF_KB_2,		TWOKEY,
	GRF_KB_3,		THREEKEY,
	GRF_KB_4,		FOURKEY,
	GRF_KB_5,		FIVEKEY,
	GRF_KB_6,		SIXKEY,
	GRF_KB_7,		SEVENKEY,
	GRF_KB_8,		EIGHTKEY,
	GRF_KB_9,		NINEKEY,
	GRF_KB_Break,		BREAKKEY,
	GRF_KB_Ctrl,		CTRLKEY,
	GRF_KB_CapsLock,	CAPSLOCKKEY,
	GRF_KB_RightShift,	RIGHTSHIFTKEY,
	GRF_KB_LeftShift,	LEFTSHIFTKEY,
	GRF_KB_Esc,		ESCKEY,
	GRF_KB_Tab,		TABKEY,
	GRF_KB_Ret,		RETKEY,
	GRF_KB_Space,		SPACEKEY,
	GRF_KB_Linefeed,	LINEFEEDKEY,
	GRF_KB_Backspace,	BACKSPACEKEY,
	GRF_KB_Del,		DELKEY,
	GRF_KB_Semicolon,	SEMICOLONKEY,
	GRF_KB_Period,		PERIODKEY,
	GRF_KB_Comma,		COMMAKEY,
	GRF_KB_Quote,		QUOTEKEY,
	GRF_KB_Backquote,	ACCENTGRAVEKEY,
	GRF_KB_Minus,		MINUSKEY,
	GRF_KB_Slash,		VIRGULEKEY,
	GRF_KB_Backslash,	BACKSLASHKEY,
	GRF_KB_Equal,		EQUALKEY,
	GRF_KB_LeftBracket,	LEFTBRACKETKEY,
	GRF_KB_RightBracket,	RIGHTBRACKETKEY,
	GRF_KB_LeftArrow,	LEFTARROWKEY,
	GRF_KB_DownArrow,	DOWNARROWKEY,
	GRF_KB_RightArrow,	RIGHTARROWKEY,
	GRF_KB_UpArrow,		UPARROWKEY,
	GRF_KB_Pad0,		PAD0,
	GRF_KB_Pad1,		PAD1,
	GRF_KB_Pad2,		PAD2,
	GRF_KB_Pad3,		PAD3,
	GRF_KB_Pad4,		PAD4,
	GRF_KB_Pad5,		PAD5,
	GRF_KB_Pad6,		PAD6,
	GRF_KB_Pad7,		PAD7,
	GRF_KB_Pad8,		PAD8,
	GRF_KB_Pad9,		PAD9,
	GRF_KB_PadF1,		PADPF1,
	GRF_KB_PadF2,		PADPF2,
	GRF_KB_PadF3,		PADPF3,
	GRF_KB_PadF4,		PADPF4,
	GRF_KB_PadPeriod,	PADPERIOD,
	GRF_KB_PadMinus,	PADMINUS,
	GRF_KB_PadComma,	PADCOMMA,
	GRF_KB_PadEnter,	PADENTER,
	GRF_KB_LeftAlt,		LEFTALTKEY,
	GRF_KB_RightAlt,	RIGHTALTKEY,
	GRF_KB_RightCtrl,	RIGHTCTRLKEY,
	GRF_KB_F1,		F1KEY,
	GRF_KB_F2,		F2KEY,
	GRF_KB_F3,		F3KEY,
	GRF_KB_F4,		F4KEY,
	GRF_KB_F5,		F5KEY,
	GRF_KB_F6,		F6KEY,
	GRF_KB_F7,		F7KEY,
	GRF_KB_F8,		F8KEY,
	GRF_KB_F9,		F9KEY,
	GRF_KB_F10,		F10KEY,
	GRF_KB_F11,		F11KEY,
	GRF_KB_F12,		F12KEY,
	GRF_KB_PrintScreen,	PRINTSCREENKEY,
	GRF_KB_ScrollLock,	SCROLLLOCKKEY,
	GRF_KB_Pause,		PAUSEKEY,
	GRF_KB_Insert,		INSERTKEY,
	GRF_KB_Home,		HOMEKEY,
	GRF_KB_PageUp,		PAGEUPKEY,
	GRF_KB_End,		ENDKEY,
	GRF_KB_PageDown,	PAGEDOWNKEY,
	GRF_KB_NumLock,		NUMLOCKKEY,
	GRF_KB_PadDivide,	PADVIRGULEKEY,
	GRF_KB_PadMultiply,	PADASTERKEY,
	GRF_KB_PadPlus,		PADPLUSKEY,

/* XXX LEFTMOUSE def must follow last key def */

	GRF_LEFT_MOUSE,		LEFTMOUSE,
	GRF_MIDDLE_MOUSE,	MIDDLEMOUSE,
	GRF_RIGHT_MOUSE,	RIGHTMOUSE,

	GRF_FOCUS_CHANGE,	INPUTCHANGE,
};

static AlistElem *alistFwd;
static AlistElem *alistRev;

static int cursorX = -1;
static int cursorY = -1;
static int curWin = -1;

static int ctrlWinSticky = 1;

static void glInitDeviceMaps()
{
	alistBuild (&alistFwd, &alistRev, alistBase, ARRAY_SIZE(alistBase));
}

static int glToGrfDevice(gld)
int gld;
{
	int grfd;

	if (alistSearch (&grfd, gld, alistRev, ARRAY_SIZE(alistBase)) == 1)
	 { return (grfd);
	 }
	else
	 { return (-1);
	 }
}

static int grfToGlDevice(grfd)
int grfd;
{
	int gld;

	if (alistSearch (&gld, grfd, alistFwd, ARRAY_SIZE(alistBase)) == 1)
	 { return (gld);
	 }
	else
	 { return (-1);
	 }
}

static int glGetEventFd()
{
#ifdef _AIX
	return (glAixQgetfd());
#else
	return (qgetfd());
#endif
}

static int connectGlDev (gldev, enable)
int gldev;
int enable;
{
	if (enable)
	 { qdevice (gldev);
	 }
	else
	 { unqdevice (gldev);
	 }
}

static int connectEvent (win, type, enable)
int win;
long type;
int enable;
{
	int glDev;

	switch (type)
	 { case GRF_KEY_EVENT:
	    { AlistElem *keys;

	      /* connectGlDev (KEYBD, enable); */
	      /* XXX there doesn't seem to be a general key enable */
	      keys=alistBase;
	      do
	       { connectGlDev (keys->right, enable);
		 keys++;
	       }
	      while (keys->right != LEFTMOUSE);	/* XXX */
	      break;
	    }
	   case GRF_BUTTON_EVENT:
	    { connectGlDev (LEFTMOUSE, enable);
	      connectGlDev (MIDDLEMOUSE, enable);
	      connectGlDev (RIGHTMOUSE, enable);
	      break;
	    }
	   case GRF_MOTION_EVENT:
	    { connectGlDev (CURSORX, enable);
	      connectGlDev (CURSORY, enable);
	      break;
	    }
	   case GRF_FOCUS_EVENT:
	    { connectGlDev (INPUTCHANGE, enable);
	      break;
	    }
	   case GRF_RESIZE_EVENT:
	   case GRF_REDRAW_EVENT:
	    { /* do nothing; these are always enabled */
	      break;
	    }
	   default:
	    { grf4Fatal ("GL: connectEvent, bogus type %d\n", type);
	    }
	 }
}

static int glEnableEvent (win, type)
int win;
long type;
{
	if (wdesc[win].eventMask & (1<<type))		/* is already enabled */
	 { return (1);
	 }
	else
	 { eventRef[type]++;
	   if (eventRef[type] == 1)
	    { connectEvent (win, type, 1);
	    }
	   wdesc[win].eventMask |= (1<<type);
	   return (0);
	 }
}

static int glDisableEvent (win, type)
int win;
long type;
{
	if (!(wdesc[win].eventMask & (1<<type)))	/* is not enabled */
	 { return (0);
	 }
	else
	 { eventRef[type]--;
	   if (eventRef[type] == 0)
	    { connectEvent (win, type, 0);
	    }
	   if (eventRef[type] < 0)
	    { grf4Fatal ("glDisablebEvent: ref count for type %d < 0\n", type);
	    }
	   wdesc[win].eventMask &= ~(1<<type);
	   return (1);
	 }
}

static int winIdToNum (wid)
int wid;
{
	if (wid == MAIN.id)
	 { return (MAINWIN);
	 }
	else if (wid == CTRL.id)
	 { return (CTRLWIN);
	 }
	else
	 { return (NULLWIN);
	 }
}

static int glPollEvent (win, code, ev)
int win;
int code;
Grf4Event *ev;
{
	int val1 = -1;
	int val2 = -1;
	int glDev;

	if (win > 0 && !wdesc[win].active)
	 { return (-1);
	 }
	switch (code)
	 { default:
	    { glDev = grfToGlDevice(code);
	      if (glDev == -1)
	       { return (-1);
	       }
	      if (GRF_IS_VALUATOR(code))
	       { val1 = getvaluator(glDev);
	       }
	      else
	       { val1 = getbutton(glDev);
	       }
	      break;
	    }
	   case GRF_CURSOR:
	    { val1 = getvaluator(CURSORX) - wdesc[win].x;
	      val2 = getvaluator(CURSORY) - wdesc[win].y;
	      break;
	    }
	   case GRF_REDRAW:
	    { val1 = wdesc[win].x;
	      val2 = wdesc[win].y;
	      break;
	    }
	   case GRF_RESIZE:
	    { val1 = wdesc[win].w;
	      val2 = wdesc[win].h;
	      break;
	    }
	   case GRF_FOCUS_CHANGE:
	    { val1 = winIdToNum (getvaluator(INPUTCHANGE));
	      break;
	    }
	 }
	ev->code = code;
	ev->val1 = val1;
	ev->val2 = val2;
	return (1);
}

typedef struct {
	long dev;
	short val;
} GlEvent;

GlEvent qbuf;
int qbufSet = 0;

static int getGlEvent (gev)
GlEvent *gev;
{
	if (qbufSet)
	 { *gev = qbuf;
	   qbufSet = 0;
	   return (1);
	 }
	else if (qtest())
	 { qbuf.dev = qread(&qbuf.val);
	   *gev = qbuf;
	   return (1);
	 }
	else
	 { return (0);
	 }
}

static int ungetGlEvent ()
{
	qbufSet = 1;
}

/* The event stack is used to insert new events where necessary */

#define EVSTACKSIZE	10

static Grf4Event evStack[EVSTACKSIZE];
static int evStackp = 0;

static int pushEvent (code, val1, val2, win)
int code, val1, val2, win;
{
	if (evStackp < EVSTACKSIZE)
	 { evStack[evStackp].code = code;
	   evStack[evStackp].val1 = val1;
	   evStack[evStackp].val2 = val2;
	   evStack[evStackp].win  = win;
	   evStackp++;
	   return (1);
	 }
	else
	 { return (0);
	 }
}

static int popEvent (ev)
Grf4Event *ev;
{
	if (evStackp)
	 { *ev = evStack[--evStackp];
	   return (1);
	 }
	else
	 { return (0);
	 }
}

static int glReadEvent (ev)
Grf4Event *ev;
{
	GlEvent gev;
	int code = GRF_NULLEVENT;
	int val1 = -1;
	int val2 = -1;
	int win;

	if (popEvent (ev))
	 { return (1);
	 }
	if (!getGlEvent(&gev))
	 { return (0);
	 }

	win = curWin; /* changed below if necessary */
	switch (gev.dev)
	 { default:
	    { if (ISKEYBD(gev.dev))
	       { if (wdesc[win].eventMask & GRF_KEY_MASK)
		  { val1 = gev.val;
		    code = glToGrfDevice (gev.dev);
		  }
	       }
	      break;
	    }
	   case LEFTMOUSE:
	   case MIDDLEMOUSE:
	   case RIGHTMOUSE:
	    { if (wdesc[win].eventMask & GRF_BUTTON_MASK)
	       { val1 = gev.val;
		 code = glToGrfDevice (gev.dev);
	       }
	      break;
	    }
	   case PIECECHANGE:
	   case REDRAW:
	    { long w, h, x, y;
	      int doCtrlRepos = 0;
	      int more;
	      int winNext;
	      GlEvent gevNext;

	      win = winIdToNum(gev.val);

	      /* hack to absorb multiple redraw events */

	      do
	       { more = getGlEvent(&gevNext);
		 winNext = winIdToNum(gevNext.val);
	       }
	      while (more && winNext==win &&
		     (gevNext.dev==REDRAW||gevNext.dev==PIECECHANGE));
	      if (more)
	       { ungetGlEvent();
	       }

	      code = GRF_REDRAW; /* changed below if necessary */
	      switch (win)
	       { case MAINWIN:
		  { getsize(&w, &h);
		    getorigin(&x, &y);
		    if (w != MAIN.w || h != MAIN.h)
		     { /* resize event */
		       MAIN.w = val1 = w;
		       MAIN.h = val2 = h;
		       if (useOrtho2)
			{ ortho2 (-0.5,MAIN.w-0.5,-0.5,MAIN.h-0.5);
			}
		       if (MAIN.w-1 >= 0)
			{ viewport (0, MAIN.w-1, 0, MAIN.h-1);
			}
		       if (ctrlWinSticky)
			{ doCtrlRepos = 1;
			}
		       code = GRF_RESIZE;
		       /* insert a redraw event */
		       MAIN.x = x;
		       MAIN.y = y;
		       pushEvent (GRF_REDRAW, x, y, win);
		     }
		    else
		     { if (x != MAIN.x || y != MAIN.y)
		       	{ /* window position has changed. */
			  if (ctrlWinSticky)
			   { doCtrlRepos = 1;
			   }
			}
		       MAIN.x = val1 = x;
		       MAIN.y = val2 = y;
		     }
		    break;
		  }
		 case CTRLWIN:
		  { winset(CTRL.id);
		    getsize(&w,&h);
		    getorigin(&x,&y);
		    if (w != CTRL.w || h != CTRL.h)
		     { /* resize event. */
		       /* NOTE: this shouldn't happen for the control window,
			  so for the moment just bury it. */
		       doCtrlRepos = 1;
		     }
		    if (x != CTRL.x || y != CTRL.y)
		     { if (ctrlWinSticky)
			{ doCtrlRepos = 1;
			}
		       else
			{ CTRL.x = val1 = x;
			  CTRL.y = val2 = y;
			}
		     }
		    if (doCtrlRepos)
		     { /* winposition() will create new events */
		       code = GRF_NULLEVENT;
		     }
		    winset(MAIN.id);
		    break;
		  }
	       }
	      if (doCtrlRepos)
	       { if (CTRL.active)
		  { winset(CTRL.id);
		    if (ctrlWinSticky)
		     { CTRL.x = MAIN.x+CTRLWINMARGIN+MAIN.w;
		       CTRL.y = MAIN.y+(MAIN.h-CTRL.h);
		     }
		    winposition (CTRL.x, CTRL.x+CTRL.w-1, CTRL.y, CTRL.y+CTRL.h-1);
		    winset(MAIN.id);
		  }
	       }
	      if (win == CTRLWIN && code == GRF_REDRAW)
	       { glDrawPanel();
	       }
	      break;
	    }
	   case CURSORX:
	   case CURSORY:
	    { int cx = -1;
	      int cy = -1;
	      int more;

	      do
	       { if (gev.dev==CURSORX)
		  { cx = gev.val;
		  }
		 else
		  { cy = gev.val;
		  }
		 more = getGlEvent(&gev);
	       }
	      while (more && (gev.dev==CURSORX||gev.dev==CURSORY));
	      if (more)
	       { ungetGlEvent();
	       }
	     /* It seems that in practice, X and Y cursor events always come in
		pairs, so that neither cx or cy should be -1. */
	      if (cx == -1)
	       { if (cursorX == -1)
		  { cursorX = getvaluator(CURSORX);	/* punt */
		  }
		 cx = cursorX;
	       }
	      if (cy == -1)
	       { if (cursorY == -1)
		  { cursorY = getvaluator(CURSORY);	/* punt */
		  }
		 cy = cursorY;
	       }
	      cursorX = cx;
	      cursorY = cy;
	      if (wdesc[win].eventMask & GRF_MOTION_MASK)
	       { val1 = cursorX-wdesc[win].x;
		 val2 = cursorY-wdesc[win].y;
		 code = GRF_CURSOR;
	       }
	      break;
	    }
	   case INPUTCHANGE:
	    { int newWin;

	      newWin = winIdToNum(gev.val);
	      if (newWin != curWin)
	       { if (wdesc[newWin].eventMask & GRF_FOCUS_MASK)
		  { pushEvent (GRF_FOCUS_CHANGE, newWin, 0, newWin);
		  }
		 if (wdesc[curWin].eventMask & GRF_FOCUS_MASK)
		  { code = GRF_FOCUS_CHANGE;
		    val1 = NULLWIN;
		  }
	       }
	      curWin = newWin;
	      break;
	    }
	 }
	ev->code = code;
	ev->val1 = val1;
	ev->val2 = val2;
	ev->win = win;
	return (1);
}

#else

#include <stdio.h>

grf4GlInit()
{
	return(-1);	
}

grf4GlSetupFxns()
{
	fprintf (stderr, "grf4Init(): GL support not available\n");
	return (-1);
}

grf4GlwfInit()
{
	return(-1);	
}

grf4GlwfSetupFxns()
{
	fprintf (stderr, "grf4Init(): GL support not available\n");
	return (-1);
}

#endif	/* GL_GRAPHICS */
