/*-=-=-=-=-=-=-=-=-=-=-=- RCCL/RCI Notice of Copyright -=-=-=-=-=-=-=-=-=-=-*
 *									    *
 *	Copyright (C) 1991 by John E. Lloyd.				    *
 *									    *
 *	Information on copyright permissions, as well as a complete	    *
 *	disclaimer of warranty, is given in the file COPYRIGHT.		    *
 *									    *
 *-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=*/

#ifdef GL_GRAPHICS

#include <stdio.h>

#include <gl/gl.h>
#include <device.h>

#include "../h/stringDefs.h"
#include "../h/cdefs.h"
#include "../h/timeval.h"
#include "../h/grf4.h"
#include "grf4Internal.h"

static Colorindex fgColor;	/* init, color */
static Colorindex bgColor;	/* drawDisplay, init, clear */

static unsigned long getcpack();

extern void glGetButtonExtent();
extern void glDrawButton();
extern void glDrawDisplay();
extern void glGetWindowSize();
extern void glSetWindowExtent();
extern int glGetEventFd();
extern int glEnableEvent();
extern int glDisableEvent();
extern int glPollEvent();
extern int glReadEvent();

static int drawAreaVisible;
static Matrix identMatrix = { 1.0, 0.0, 0.0, 0.0,
			      0.0, 1.0, 0.0, 0.0,
			      0.0, 0.0, 1.0, 0.0,
			      0.0, 0.0, 0.0, 1.0 };

static unsigned long getcpack()
{
	short r,g,b;
	gRGBcolor(&r,&g,&b);
	return (RGBPACK(r,g,b));
}

#define MIN_WINW	21
#define MIN_WINH	21

grf4GlwfInit(name, fgcolor, bgcolor, x,y,w,h)
char *name;
int fgcolor;
int bgcolor;
int x,y,w,h;
{
	long winx, winy, winw, winh;
	int panw, panh;

#ifdef _AIX
	glAixQgetfdHack1();
#endif

#ifndef _AIX
	foreground();
#endif	
	if (x != -1)
	 { prefposition (x,x+w-1,y,y+h-1);
	 }
#ifdef _AIX 
/* XXX XXX XXX AIX doesn't seem to support interactive placement! */
	else
	 { prefposition (256,628,256,511);
	 }
#endif
	if (winopen (name) == -1)
	 { return (-1);
	 }
	buttonInit (&panw, &panh);
	getorigin (&winx, &winy);
	getsize (&winw, &winh);
	if (panw)
	 { /* Make sure the new window is big enough to include the
	      panel. */
#ifndef _AIX		
   /* XXX XXX XXX winposition on AIX is broken! Can't resize window!! */
	   winw += panw;
	   winh = MAX(panh,winh);
	   winposition (winx, winx+winw-1, winy, winy+winh-1);
#endif
	   minsize (panw+MIN_WINW, MAX(MIN_WINH,panh));
	   winconstraints();
	 }
	RGBmode();
	if (!(flags & GRF_DOUBLEBUF_OFF))
	 { doublebuffer();
	 }
	gconfig();
	
	bgColor = bgcolor;
	fgColor = fgcolor;

	/* setup for receiving events */

	qdevice(CURSORX);
	qdevice(CURSORY);
	qdevice(LEFTMOUSE);
	qdevice(MIDDLEMOUSE);
	qdevice(REDRAW);
	qdevice(PIECECHANGE);

	loadmatrix(identMatrix);
	
	if (winw-panw-1 >= 0)
	 { viewport ((Screencoord)0, (Screencoord)(winw-panw-1), 
		     (Screencoord)0, (Screencoord)(winh-1));
	   drawAreaVisible = 1;
	 }
	else
	 { drawAreaVisible = 0;
	 }
	ortho2 (-0.5,winw-panw-0.5,-0.5,winh-0.5);
	glSetWindowExtent (winx, winy, winw, winh, panw, panh);

	glDrawDisplay(winw, winh, panw);

#ifdef _AIX
	glAixQgetfdHack2();
#endif
	glInitDeviceMaps();
	cpack(fgColor);
	return (1);
}

void glwfClear(x,y,w,h)
int x,y,w,h;
{
	unsigned long savec;
	savec = getcpack();
	cpack(bgColor);
	clear();
	cpack(savec);
}

void glwfColor(color)
int color;
{
#if  0
	fgColor = color;
	cpack(color);
#endif
}

void glwfFlush(x,y,w,h)
int x,y,w,h;
{
	swapbuffers();
}
	 
void glwfDraw(x1,y1,x2,y2)
int x1, y1, x2, y2;
{
	move2 (x1, y1);
	draw2 (x2, y2);
}


grf4GlwfSetupFxns (ftab)
struct grf4fxnTable *ftab;
{
	ftab->getWindowSize = glGetWindowSize;
#if DEAD
	ftab->cycle = glwfCycle;
	ftab->wait = glwfWait;
	ftab->select = glwfSelect;
	ftab->eventHandler = glwfEventHandler;
#endif
	ftab->color = glwfColor;

	ftab->drawButton = glDrawButton;
	ftab->getButtonExtent = glGetButtonExtent;

	ftab->pxlDraw = glwfDraw;
	ftab->pxlFlush = glwfFlush;
	ftab->pxlClear = glwfClear;

	ftab->getEventFd = glGetEventFd;
	ftab->enableEvent = glEnableEvent;
	ftab->disableEvent = glDisableEvent;
	ftab->pollEvent = glPollEvent;
	ftab->readEvent = glReadEvent;

	return (1);
}

#else

#include <stdio.h>

grf4GlwfInit ()
{
	return(-1);
}

grf4GlwfSetupFxns()
{
	fprintf (stderr, "grf4Init(): GL wire-frame support not available\n");
	return (-1);
}

#endif
