#include <stdio.h>
#include <math.h>
#include "../h/cdefs.h"
#include "../h/grf4.h"
#include "argDef.h"

#ifndef M_PI
#define M_PI    3.14159265358979323846
#endif
#define RTOD    (180/M_PI)
#define DTOR    (M_PI/180)

int winw;
int winh;
float xmin = -100.0;
float xmax =  100.0;
float ymin = -100.0;
float ymax =  100.0;
float xwidth;
float ywidth;

float aaa;
float bbb;

drawSeg (x0, y0, x1, y1)
float x0, y0, x1, y1;
{
	grf4Move2 (x0, y0);
	grf4Draw2 (x1, y1);
}

drawIt()
{
	grf4Clear();

	grf4Move2 (xmin, ymin);
	grf4Draw2 (xmax, ymax);
	grf4Move2 (xmin, (ymax+ymin)/2.0);
	grf4Draw2 (xmax, (ymax+ymin)/2.0);

	grf4Flush();
}

onQuit()
{
	exit (0);
}

int aaaIncButton;
int aaaDispButton;
int aaaDecButton;
int bbbIncButton;
int bbbDispButton;
int bbbDecButton;

onDelta_aaa(n, code)
int n, code;
{
	char buf[256];

	if (code == BF_PRESSED)
	 { if (n == aaaIncButton)
	    { aaa += 1.0;
	    }
	   else
	    { aaa -= 1.0;
	    }
	   sprintf (buf, "aaa=%5.2f", aaa);
	   grf4SetButtonName (aaaDispButton, buf);
	   drawIt();
	 }
}

onDelta_bbb(n, code)
int n, code;
{
	char buf[256];

	if (code == BF_PRESSED)
	 { if (n == bbbIncButton)
	    { bbb += 1.0;
	    }
	   else
	    { bbb -= 1.0;
	    }
	   sprintf (buf, "bbb=%5.2f", bbb);
	   grf4SetButtonName (bbbDispButton, buf);
	   drawIt();
	 }
}

onSavePlot(n, code)
int n, code;
{
	FILE *fp;
	char fileName[256];

	if (code == BF_RELEASED)
	 { printf ("File name> ");
	   scanf ("%s", fileName);
	   if ((fp = fopen(fileName, "w")) == NULL)
	    { printf ("Can't open file %s\n", fileName);
	    }
	   grf4PostScriptInit (fp, "plot.ps", /*flags=*/0,
				   /*width=*/ 72*6.0, /*height=*/ 72*9.0);
	   grf4HideLines(1);
	   drawIt();
	   grf4HideLines(0);
	   grf4PostScriptFinish();
	   fclose(fp);
	 }
}

int printHelp;
char **argvec;

main(argc, argv)
int argc;
char **argv;
{
	ArgDefList *dlist;
	int n;
	FILE *fp;

	argvec = argv;
	dlist = argDefCreate();
	argDefAdd (dlist, 
	   "-help %v, 1,\"print out 'help' information\"", &printHelp,
	   0);

	argc--; argv++;
	while (argc)
	 { n = argDefMatch (dlist, argc, argv);
	   if (n < 0)
	    { fprintf (stderr, "%s\n", argDefErrorStr());
	      exit (1);
	    }
	   else if (n > 0)
	    { argc -= n;
	      argv += n;
	    }
	   else
	    { fprintf (stderr, "Usage: %s\n", argvec[0]);
	      argDefPrintUsage (stderr, dlist, 80);
	      exit (1);
	    } 
	 }
	if (printHelp)
	 { argDefPrintHelp(stdout, dlist);
	   exit (0);
	 }

	grf4AddButton ("quit", 0, 1, onQuit, 0);
	aaaIncButton = grf4AddButton ("+", 0, 1, onDelta_aaa, 0);
	aaaDispButton = grf4AddButton ("aaa=xxxxxx", BF_INACTIVE|BF_SAMEROW, 
					       2, onDelta_aaa, 0);	
	aaaDecButton = grf4AddButton ("-", BF_SAMEROW, 1, onDelta_aaa, 0);
	bbbIncButton = grf4AddButton ("+", 0, 1, onDelta_bbb, 0);
	bbbDispButton = grf4AddButton ("bbb=xxxxx", BF_INACTIVE|BF_SAMEROW, 
						2, onDelta_bbb, 0);
	bbbDecButton = grf4AddButton ("-", BF_SAMEROW, 1, onDelta_bbb, 0);
	grf4AddButton ("save plot", 0, 1, onSavePlot, 0);
	grf4PlaceWindow (300, 300, 600, 400);
	if (grf4Init ("foo", GRF_WHITE, GRF_BLACK, GRF_X11) < 0)
	 { printf ("Can't init graphics\n");
	   exit (-1);
	 }
	grf4SetDrawFxn (drawIt, 0);

	grf4Color (0xffffff);
	grf4GetWindowSize (&winw, &winh);

/*
	xmin = -4/3.0*l;
	xmax =  4/3.0*l;
	ymin = -2*M_PI;
	ymax =  2*M_PI;
*/
	xwidth = xmax-xmin;
	ywidth = ymax-ymin;
	
	grf4Ortho (xmin, xmax, ymin, ymax, -200.0, 200.0, 0.0);

	drawIt();

	grf4Wait (-1);
}
