/*-=-=-=-=-=-=-=-=-=-=-=- RCCL/RCI Notice of Copyright -=-=-=-=-=-=-=-=-=-=-*
 *									    *
 *	Copyright (C) 1991 by John E. Lloyd.				    *
 *									    *
 *	Information on copyright permissions, as well as a complete	    *
 *	disclaimer of warranty, is given in the file COPYRIGHT.		    *
 *									    *
 *-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=*/
#include <stdio.h>
#include <math.h>

#include "../h/cdefs.h"
#include "../h/rccl.h"	/* XXX MUST include rccl.h before grf4.h */
#include "../h/linear_kynvar.h"

#include "../h/grf4.h"

#define NUM_CIRC_EDGES		12

#define J1	0

static int capacity;

#define BEIGE		RGBPACK(225,160, 90)
#define DARK_BROWN	RGBPACK( 85, 55, 35)
#define DARK_GREY	RGBPACK( 50, 50, 50)
#define BLACK		RGBPACK(  0,  0,  0)
#define WHITE		RGBPACK(255,255,255)
#define SILVER		RGBPACK(200,200,220)
#define AMBER		RGBPACK(255,125,  0)

/* Stage sim stuff */

#define LSIM_TRACK_DEPTH		 1.2
#define LSIM_TRACK_WIDTH		15.24
#define LSIM_TRACK_X_LEN		45.72
#define LSIM_PEDESTAL_X_DEPTH		 6.35

static multA1();

Grf4Object *linearBaseX;
Grf4Object *linearPedestalX;

defineLinearSimLinks()
{
	void *bodyMat;
	void *rotorMat;

	rotorMat = grf4DefMaterial(SILVER, SILVER, 0, 0, 0.0, 1.0);
	bodyMat = grf4DefMaterial(DARK_GREY, DARK_GREY, 0,0x808080,50.0,1.0);

	linearBaseX = grf4GenBox (LSIM_TRACK_X_LEN, LSIM_TRACK_WIDTH, 
				 LSIM_TRACK_DEPTH, bodyMat);
	linearPedestalX = grf4GenBox (LSIM_TRACK_WIDTH, LSIM_TRACK_WIDTH,
				 LSIM_PEDESTAL_X_DEPTH, bodyMat);

  	capacity = grf4Capacity();
}

drawLinearSim (T, kyn, var, drawHand, handpos)
register TRSF *T;
KYN *kyn;
void *var;
int drawHand;
int handpos;
{
	/* clear the display and draw a robot, given a set of joint angles */

	LINEAR_KYN *evar = (LINEAR_KYN*)var;
	GEN_KYN *gen = kyn->gen;

	TRSF trans;

	grf4GetTrans (&trans);
	T[0] = trans;
	multTrsfXyz (&trans, -LSIM_TRACK_X_LEN/2, -LSIM_TRACK_WIDTH/2, 
			     -LSIM_TRACK_DEPTH-LSIM_PEDESTAL_X_DEPTH);
	grf4SetTrans (&trans);
	grf4DrawObject (linearBaseX);

	multA1 (&trans, &T[0], kyn, evar);
	T[1] = trans;
	multTrsfRpy (&trans, 0, -90, 0);
	multTrsfXyz (&trans, -LSIM_TRACK_WIDTH/2, -LSIM_TRACK_WIDTH/2, 
			     -LSIM_PEDESTAL_X_DEPTH);
	grf4SetTrans (&trans);
	grf4DrawObject (linearPedestalX);
}

linearSimViewSize (r, b, h, kyn)
float *r, *b, *h;
KYN *kyn;
{
	GEN_KYN *gen = kyn->gen;

	*r = LSIM_TRACK_X_LEN;
	*b = 0;
	*h = *r;
}

linearSimLinkFrames (T, B, kyn, var)
TRSF *T, *B;
KYN *kyn;
void *var;
{
	LINEAR_KYN *evar = (LINEAR_KYN*)var;

	T[0] = *B;
	multA1 (&T[1], &T[0], kyn, evar);
}

/* Routines for calculating the frame matrices for LINEAR robots */

static multA1(tr, t1, kyn, var)
register TRSF *tr;
register TRSF *t1;
KYN *kyn;
LINEAR_VAR *var;
{
	register GEN_KYN *gen = kyn->gen;
	float D;

	D = var->d[J1]/10;

	tr->n.x = -t1->a.x;
	tr->o.x = -t1->n.x;
	tr->a.x =  t1->o.x;
	tr->p.x =  t1->n.x*D + t1->p.x;

	tr->n.y = -t1->a.y;
	tr->o.y = -t1->n.y;
	tr->a.y =  t1->o.y;
	tr->p.y =  t1->n.y*D + t1->p.y;

	tr->n.z = -t1->a.z;
	tr->o.z = -t1->n.z;
	tr->a.z =  t1->o.z;
	tr->p.z =  t1->n.z*D + t1->p.z;
}
