
/*#define SIM*/
/*#define DEBUG1*/


/*
#define TIMER_1
#define TIMER_2
*/

/* Command executed by a KIN process */

#define KIN_DEPART_FN         1
#define KIN_WTOJ_FN           2
#define KIN_OL_WTOJ_FN        3
#define KIN_OL_WMAT_TOJ_FN    4
#define KIN_JTOW_FN           5
#define KIN_OL_DEPART_FN      6
#define KIN_CFG_FN            7
#define KIN_TST			8
#define KIN_MTOW_FN			9
#define KIN_WTOM_FN			10
#define KIN_FORCECON_FN		11
#define KIN_FORCESET_FN		12
#define KIN_B430_FN		13
#define KIN_FN				15  /* Msg to Kin are between [0..14] */

/* Command executed by the KIN InterFace (IF) process */

#define IF_SMALL_DBASE_LD	  0xF9
#define IF_DBASE_LOAD	  	  0xFA
#define IF_TASK_START		  0xFB
#define IF_TASK_OUTPUT		  0xFC
#define IF_TASK_EXEC_REG	  0xFD
#define IF_TASK_ALLOC_REG	  0xFE
#define IF_TASK_STATUS_REG	  0xFF


/*
* 	Messages received by the xputer are denoted by (xxxx_IN);
* 	Messages sent by the xputer to the host are denoted by (xxx_OUT)
* 
* NOTE:
* 
* 	The IN message protocol which requires an action by one of the KIN
* 	processes is as follows:
* 
* 	1) 	The host selects the Task Identification Number by setting the
* 		TaskIdReg;
* 	2)  The host sends the TaskIdReg (1 byte) to the xputer;
* 	3)  The host sends the message length (1 int) + parameters 
		(x quadruples) to the Xputer;
* 	4)  If a KIN task selected by the Host is already busy, a bit of the
* 		Status Register indicating a load Task error is set by the 
		Xputer.
* 	5)  The Xputer replies by sending a status byte. If a Task Error 
		occured, the Task must be reloaded.
* 
* 		ex:
* 		1) TaskIdReg = 0x81 (Task Id 7  and Task Id 1 selected)
* 		2) Host to Xputer >> TaskIdReg;
* 		3) Host to Xputer >> MsgLen of Task 1 + Task 1 Parameters
* 		   Host to Xputer >> MsgLen of Task 7 + Task 7 Parameters
* 		4) Xputer to Host >> Status byte = 0x80 (Task 7 already busy,
* 			Task 1 loaded and executing). Task 7 to be reloaded.
* 
* 	The OUT message protocol to download the result of a KIN process is
* 	as follows:
* 
* 	1) 	The host selects the Task Identification Number by setting the
* 		TaskIdReg;
* 	2)  The host sends the TaskIdReg (1 byte) to the xputer;
* 	3)  The Xputer sends the message length (1 int) + parameters 
*		(x quadruples) to the host;
* 	4)  The Xputer sends a Status byte. If the requested Task Output was
* 		not done or a bad Task Id was selected, the Status byte is set
* 		and the KIN output data previously sent is not reliable.
* 
* 
* 	All parameter data types in the messages are quadruple to ensure
* 	quadruple byte alignment at the Kin interface process.
* 
* 	The following describe the content of the parameter list followed by
* 	their size and type. (1i = 1 integer, 8f = 8 floats);
* 
* 	The KIN_DBASE_LOAD data is done during initialization and is not
* 	considered a time critical function. For clarity, values of the
* 	database are received by the Xputer using multiple ChanIn calls
* 	instead of reading a single string.
* 
* DEPART_MSG_IN 	 	Function(1i)+Config(i1)+World(8f)+Depart(8f)+OldJt(8f)
* DEPART_MSG_OUT 	 	Function(1i)+RetCode(1i)+Joint(8f)
* OL_DEPART_MSG_IN 		Function(1i)+Config(i1)+World(8f)+Depart(8f)+OldJt(8f)
* OL_DEPART_MSG_OUT 	Function(1i)+RetCode(1i)+Joint(8f)+World(8f)
* OL_WTOJ_MSG_IN  		Function(1i)+Config(1i)+World(8f)+OldJt(8f)
* OL_WTOJ_MSG_OUT 		Function(1i)+RetCode(1i)+Joint(8f)
* WTOJ_MSG_IN  		Function(1i)+Config(1i)+World(8f)+OldJt(8f)
* WTOJ_MSG_OUT 		Function(1i)+RetCode(1i)+Joint(8f)
* OL_WMAT_TOJ_MSG_IN  	Function(1i)+Config(1i)+World(12f)+OldJt(8f)
* OL_WMAT_TOJ_MSG_OUT 	Function(1i)+RetCode(1i)+Joint(8f)
* JTOW_MSG_IN         	Function(1i)+Config(1i)+Joint(8f)
* JTOW_MSG_OUT        	Function(1i)+ReCode(1i)+World(8f)
* CFG_MSG_IN          	Function(1i)+Config(1i)+Pos(8f)
* CFG_MSG_OUT         	Function(1i)+RetCode(1i)+NewConfig(1i)
* SMALL_DB_LD_MSG_IN  	Function(1i)+ToolTransform(8f)+BaseShift(8f)
* SMALL_DB_LD_MSG_OUT	Function(1i)+RetCode(1i)
* KIN_TST_MSG_IN      	Function(1i)+TestValue(1i)
* KIN_TST_MSG_OUT     	Function(1i)+RetCode(1i)+TestValue(1i)
* FORCECON_MSG_IN		Function(1i)+ForceData(6f)+ForceGains(6f)
* FORCECON_MSG_OUT		Function(1i)+RetCode(1i)+Deltas(8f)
* B430_MSG_IN		Function(1i)+GripReq(2i)
* B430_MSG_OUT		Function(1i)+RetCode(1i)+GripRep(9i)
*/

#define DEPART_MSG_IN         104
#define DEPART_MSG_OUT        40
#define OL_DEPART_MSG_IN      104
#define OL_DEPART_MSG_OUT     72
#define OL_WTOJ_MSG_IN		88
#define OL_WTOJ_MSG_OUT       40
#define WTOJ_MSG_IN			72
#define WTOJ_MSG_OUT         	40
#define WTOM_MSG_IN			72
#define WTOM_MSG_OUT         	40
#define OL_WMAT_TOJ_MSG_IN    72
#define OL_WMAT_TOJ_MSG_OUT   40
#define JTOW_MSG_IN           40
#define JTOW_MSG_OUT          40
#define MTOW_MSG_IN           40
#define MTOW_MSG_OUT          40
#define CFG_MSG_IN            40
#define CFG_MSG_OUT           12
#define SMALL_DB_LD_MSG_IN    68
#define SMALL_DB_LD_MSG_OUT	8
#define KIN_TST_MSG_IN        8
#define KIN_TST_MSG_OUT      	12
#define FORCECON_MSG_OUT		(4+4+32+32)
#define FORCECON_MSG_IN    	(4+24+24)
#define B430_MSG_OUT		(4+4+36)
#define B430_MSG_IN		(4+8)

#define KIN_MAX_MSG_LEN         120

/* kin transputer process ID */

#define KIN_MAIN			0
#define KIN_HOST_ID			1
#define KINIF_ID			2
#define KIN_ID_BASE			3   /* (3 to 11) */

#define WS_KIN				4096 	/* Kin process task workspace */
#define WS_KINIF			4096    /* KinIF process task workspace */
#define WS_SERVOIF			4096
#define NUM_OF_KIN			8       /* Number of KIN processes */
