#include <stdio.h>
#include <ctype.h>
#include <errno.h>
#include "../h/stringDefs.h"
#include "../h/utilFxns.h"
#include "../h/errorStack.h"
#include "../h/errorCodes.h"

FILE *pathSearch (fullnamep, pathStr, name)
char **fullnamep;
const char *pathStr;
const char *name;
{
	/* Returns a FILE * to the first occurance of the file
	   'name' in the directories 'pathStr' which can be opened
	   read-only. If no file can be openned, NULL is returned. */

	const char *pathPtr;
	char *nameBuf;
	char *namePtr;

	FILE *fp = NULL;

	nameBuf = (char*)malloc(strlen(name)+strlen(pathStr)+2);

	pathPtr = pathStr;
	do
	 { namePtr = nameBuf;
	   while (*pathPtr != ':' && *pathPtr != '\0')
	    { *namePtr++ = *pathPtr++;
	    }
	   if (namePtr != nameBuf)
	    { *namePtr++ = '/';
	      *namePtr = '\0';
	      strcat (nameBuf, name);
	      fp = fopen (nameBuf, "r");
	    }
	   else
	    { if (*pathPtr == '\0')
	       { free (nameBuf);
		 return (NULL);
	       }
	    }
	   if (*pathPtr == ':')
	    { pathPtr++;
	    }
	 }
	while (fp == NULL);

	if (fullnamep)
	 { *fullnamep = copyofstr(nameBuf);
	 }
	free (nameBuf);
	return (fp);	
}

FILE *locateFile (fullnamep, name)
char **fullnamep;
const char *name;
{
	/* Driver routine for pathSearch. Checks environment variables,
	   and prints appropriate error messages. */

	char *fxnName = "locateFile()";
	FILE *fp;

	if (getenvPathTarget() == NULL)
	 { errorMsg (EEnviron, 
"locateFile() -- environment variable %s not set\n", getPathTargetName());
	   return (NULL);
	 }
	fp = pathSearch (fullnamep, getenvPathTarget(), name);
	if (fp == NULL)
	 { errorMsg (EBadAccess, 
"locateFile() -- can't locate or open %s in the path\n\
\tdescribed by the environment variable %s;\n\
\tcheck the path and/or permissions\n", name, getPathTargetName());
	 }
	else
	 { if (rciGetenv ("RCCL_PATH_PRINT"))
	    { printf ("reading %s ...\n", *fullnamep);
	    }
	 }
	return (fp);
}
