/*-=-=-=-=-=-=-=-=-=-=-=- RCCL/RCI Notice of Copyright -=-=-=-=-=-=-=-=-=-=-*
 *									    *
 *	Copyright (C) 1987, 1991 by John E. Lloyd.			    *
 *									    *
 *	Information on copyright permissions, as well as a complete	    *
 *	disclaimer of warranty, is given in the file COPYRIGHT.		    *
 *									    *
 *-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=*/

/*=~=~=~=~ McGill-Computer-Vision-and-Robotics-Laboratory--C-Module ~=~=~=~=
 *		
 * MODULE NAME:	vector.c
 *
 * DESCRIPTION:	A library of vector support routines for floats and doubles.
 *
 *		The object is to use these to replace the myriad of different
 *		routines that exist for manipulating identical data 
 *		structures.
 * 
 * AUTHOR/DATE:	John Lloyd, Spring 87
 *
 *~=~=~=~=~=~= McGill-Computer-Vision-and-Robotics-Laboratory ~=~=~=~=~=~=*/

#include <stdio.h>
#include <math.h>

#include "../h/cdefs.h"
#include "../h/vector.h"
#include "../h/robotMath.h"

static int fprintV();

/* Assignment routines */

float *copyVf (r, v, size)
register float *r, *v;
int size;
{
	register int j;

	for (j=0; j<size; j++)
	 { r[j] = v[j];
	 }
	return (r);
}

double *copyVd (r, v, size)
register double *r, *v;
int size;
{
	register int j;

	for (j=0; j<size; j++)
	 { r[j] = v[j];
	 }
	return (r);
}

int *copyVi (r, v, size)
register int *r, *v;
int size;
{
	register int j;

	for (j=0; j<size; j++)
	 { r[j] = v[j];
	 }
	return (r);
}

float *copyVdf (r, v, size)
register float *r;
register double *v;
int size;
{
	register int j;

	for (j=0; j<size; j++)
	 { r[j] = v[j];
	 }
	return (r);
}

double *copyVfd (r, v, size)
register double *r;
register float *v;
int size;
{
	register int j;

	for (j=0; j<size; j++)
	 { r[j] = v[j];
	 }
	return (r);
}

float *copyVif (r, v, size)
register float *r;
register int *v;
int size;
{
	register int j;

	for (j=0; j<size; j++)
	 { r[j] = v[j];
	 }
	return (r);
}

int *copyVfi (r, v, size)
register int *r;
register float *v;
int size;
{
	register int j;

	for (j=0; j<size; j++)
	 { r[j] = v[j];
	 }
	return (r);
}

double *copyVid (r, v, size)
register double *r;
register int *v;
int size;
{
	register int j;

	for (j=0; j<size; j++)
	 { r[j] = v[j];
	 }
	return (r);
}

int *copyVdi (r, v, size)
register int *r;
register double *v;
int size;
{
	register int j;

	for (j=0; j<size; j++)
	 { r[j] = v[j];
	 }
	return (r);
}

/* routines for zeroing a vector */

float *zeroVf (r, size)
register float *r;
int size;
{
	register int j;
	for (j=0; j<size; j++)
	 { r[j] = 0.0;
	 }
	return (r);	
}

double *zeroVd (r, size)
register double *r;
int size;
{
	register int j;
	for (j=0; j<size; j++)
	 { r[j] = 0.0;
	 }
	return (r);	
}

int *zeroVi (r, size)
register int *r;
int size;
{
	register int j;
	for (j=0; j<size; j++)
	 { r[j] = 0.0;
	 }
	return (r);	
}

/* single precision arithmetic routines */

float *subtractVf (r, v1, v2, size)
register float *r, *v1, *v2;
int size;
{
	register int j;

	for (j=0; j<size; j++)
	 { r[j] = v1[j] - v2[j];
	 }
	return (r);
}

float *addVf (r, v1, v2, size)
register float *r, *v1, *v2;
int size;
{
	register int j;

	for (j=0; j<size; j++)
	 { r[j] = v1[j] + v2[j];
	 }
	return (r);
}

float *scaleVf (r, a, v, size)
register float *r, *v;
register double a;
int size;
{
	register int j;

	for (j=0; j<size; j++)
	 { r[j] = a * v[j];
	 }
	return (r);
}

float *combineVf (r, a, v1, b, v2, size)
register float *r, *v1, *v2;
register double a, b;
int size;
{
	register int j;

	for (j=0; j<size; j++)
	 { r[j] = a * v1[j] + b * v2[j];
	 }
	return (r);
}

float norm2Vf (v, size)
register float *v;
int size;
{
	register int j;
	register float x = 0.0;

	for (j=0; j<size; j++)
	 { x += v[j]*v[j];
	 }
	return (sqrt(x));
}

float normInfVf (v, size)
register float *v;
int size;
{
	register int j;
	register float x = 0.0;
	register float val;

	for (j=0; j<size; j++)
	 { val = v[j];
	   if (ABS(val) > x)
	    { x = ABS(val);
	    }
	 }
	return (x);
}

float norm1Vf (v, size)
register float *v;
int size;
{
	register int j;
	register float x = 0.0;

	for (j=0; j<size; j++)
	 { x += ABS(v[j]);
	 }
	return (x);
}

float normVf (v, size)
register float *v;
int size;
{
	return (norm2Vf (v, size));
}

float dotVf (v1, v2, size)
register float *v1, *v2;
int size;
{
	register int j;
	register float x = 0.0;

	for (j=0; j<size; j++)
	 { x += v1[j]*v2[j];
	 }
	return (x);
}

/* double precision arithmetic routines */

double *subtractVd (r, v1, v2, size)
register double *r, *v1, *v2;
int size;
{
	register int j;

	for (j=0; j<size; j++)
	 { r[j] = v1[j] - v2[j];
	 }
	return (r);
}

double *addVd (r, v1, v2, size)
register double *r, *v1, *v2;
int size;
{
	register int j;

	for (j=0; j<size; j++)
	 { r[j] = v1[j] + v2[j];
	 }
	return (r);
}

double *scaleVd (r, a, v, size)
register double *r, *v;
register double a;
int size;
{
	register int j;

	for (j=0; j<size; j++)
	 { r[j] = a * v[j];
	 }
	return (r);
}

double *combineVd (r, a, v1, b, v2, size)
register double *r, *v1, *v2;
register double a, b;
int size;
{
	register int j;

	for (j=0; j<size; j++)
	 { r[j] = a * v1[j] + b * v2[j];
	 }
	return (r);
}

double norm2Vd (v, size)
register double *v;
int size;
{
	register int j;
	register double x = 0.0;

	for (j=0; j<size; j++)
	 { x += v[j]*v[j];
	 }
	return (sqrt(x));
}

double normInfVd (v, size)
register double *v;
int size;
{
	register int j;
	register double x = 0.0;
	register double val;

	for (j=0; j<size; j++)
	 { val = v[j];
	   if (ABS(val) > x)
	    { x = ABS(val);
	    }
	 }
	return (x);
}

double norm1Vd (v, size)
register double *v;
int size;
{
	register int j;
	register double x = 0.0;

	for (j=0; j<size; j++)
	 { x += ABS(v[j]);
	 }
	return (x);
}

double normVd (v, size)
register double *v;
int size;
{
	return (norm2Vd (v, size));
}

double dotVd (v1, v2, size)
register double *v1, *v2;
int size;
{
	register int j;
	register double x = 0.0;

	for (j=0; j<size; j++)
	 { x += v1[j]*v2[j];
	 }
	return (x);
}
 
/* integer arithmetic routines */

int *subtractVi (r, v1, v2, size)
register int *r, *v1, *v2;
int size;
{
	register int j;

	for (j=0; j<size; j++)
	 { r[j] = v1[j] - v2[j];
	 }
	return (r);
}

int *addVi (r, v1, v2, size)
register int *r, *v1, *v2;
int size;
{
	register int j;

	for (j=0; j<size; j++)
	 { r[j] = v1[j] + v2[j];
	 }
	return (r);
}

int *scaleVi (r, a, v, size)
register int *r, *v;
register double a;
int size;
{
	register int j;

	for (j=0; j<size; j++)
	 { r[j] = a * v[j];
	 }
	return (r);
}

int *combineVi (r, a, v1, b, v2, size)
register int *r, *v1, *v2;
register double a, b;
int size;
{
	register int j;

	for (j=0; j<size; j++)
	 { r[j] = a * v1[j] + b * v2[j];
	 }
	return (r);
}

double norm2Vi (v, size)
register int *v;
int size;
{
	register int j;
	register double x = 0.0;

	for (j=0; j<size; j++)
	 { x += v[j]*v[j];
	 }
	return (sqrt(x));
}

double normInfVi (v, size)
register int *v;
int size;
{
	register int j;
	register double x = 0.0;
	register double val;

	for (j=0; j<size; j++)
	 { val = v[j];
	   if (ABS(val) > x)
	    { x = ABS(val);
	    }
	 }
	return (x);
}

double norm1Vi (v, size)
register int *v;
int size;
{
	register int j;
	register double x = 0.0;

	for (j=0; j<size; j++)
	 { x += ABS(v[j]);
	 }
	return (x);
}

double normVi (v, size)
register int *v;
int size;
{
	return (norm2Vi (v, size));
}

double dotVi (v1, v2, size)
register int *v1, *v2;
int size;
{
	register int j;
	register double x = 0.0;

	for (j=0; j<size; j++)
	 { x += v1[j]*v2[j];
	 }
	return (x);
}

/* IO routines */

fscanVd (fp, v, size)
FILE *fp;
double *v;
int size;
{
	register int j;
	int status;

	for (j=0; j<size; j++)
	 { if ((status = fscanf (fp, "%lf", &v[j])) != 1)
	    { break;
	    }
	 }
	return (j == size ? 1 : status );
}

fscanVf (fp, v, size)
FILE *fp;
float *v;
int size;
{
	register int j;
	int status;

	for (j=0; j<size; j++)
	 { if ((status = fscanf (fp, "%f", &v[j])) != 1)
	    { break;
	    }
	 }
	return (j == size ? 1 : status );
}

fscanVi (fp, v, size)
FILE *fp;
int *v;
int size;
{
	register int j;
	int status;

	for (j=0; j<size; j++)
	 { if ((status = fscanf (fp, "%d", &v[j])) != 1)
	    { break;
	    }
	 }
	return (j == size ? 1 : status );
}

scanVf (v, size)
float *v;
int size;
{
	return (fscanVf (stdin, v, size));
}

scanVd (v, size)
double *v;
int size;
{
	return (fscanVd (stdin, v, size));
}

scanVi (v, size)
int *v;
int size;
{
	return (fscanVi (stdin, v, size));
}

#define FLOAT_TYPE	0
#define DOUBLE_TYPE	1
#define INT_TYPE	2

int fprintVd (fp, s, v, size)
FILE *fp;
const char *s;
double *v;
{
	return (fprintV (fp, s, (void*)v, size, DOUBLE_TYPE));
}

int fprintVf (fp, s, v, size)
FILE *fp;
const char *s;
float *v;
{
	return (fprintV (fp, s, (void*)v, size, FLOAT_TYPE));
}

int fprintVi (fp, s, v, size)
FILE *fp;
const char *s;
int *v;
{
	return (fprintV (fp, s, (void*)v, size, INT_TYPE));
}

static int fprintV (fp, s, v, size, type)
FILE *fp;
const char *s;
void *v;
int size;
int type;
{
#define MAXF	256
	char buf[MAXF+2];
	char *format;
	register char *cp;
	register int c, i;
	int nc;
	int ncTotal = 0;
	
	float *vf;
	double *vd;
	int *vi;

	switch (type)
	 { case INT_TYPE:
	    { vi = (int*)v;
	      break;
	    }
	   case FLOAT_TYPE:
	    { vf = (float*)v;
	      break;
	    }
	   case DOUBLE_TYPE:
	    { vd = (double*)v;
	      break;
	    }
	 }
	strncpy (buf, s, MAXF);
	buf[MAXF] = '\0';

	cp = format = buf;		/* locate format specifier */
	while (*cp != '\0')
	 { if (*cp == '%')
	    { if (*(cp+1) == '%')
	       { cp++;
	       }
	      else
	       { break;
	       }
	    }
	   cp++;
	 }
	
	if (*cp == '\0')
	 { nc = fprintf (fp, format);
	   if (nc < 0)
	    { return (nc);
	    }
	   else
	    { ncTotal += nc;
	    }
	 }
	else
	 { *cp = '\0';
	   nc = fprintf (fp, format);
	   if (nc < 0)
	    { return (nc);
	    }
	   else
	    { ncTotal += nc;
	    }

	   *cp = '%';
	   format = cp;
	   while (*cp != '\0')
	    { c = *cp;
	      if (c == 'd' || c == 'o' || c == 'x' ||
		  c == 'f' || c == 'g' || c == 'e' ||
		  c == 'F' || c == 'G' ||
		  c == 'c' || c == 's' || c == 'u' )
	       { break;
	       }
	      cp++;
	    }
	   if (*cp != '\0')
 	    { char c1, c2;

	      c1 = cp[1];		/* massage format string to add */
	      c2 = cp[2];		/* extra <space> after number */
	      cp[1] = ' ';
	      cp[2] = '\0';

	      for (i=0; i<size; i++)
	       { 
		 switch (type)
		  { case INT_TYPE:
		     { nc = fprintf (fp, format, vi[i]);
		       if (nc < 0)
			{ return (nc);
			}
		       else
			{ ncTotal += nc;
			}
		       break;
		     }
		    case FLOAT_TYPE:
		     { nc = fprintf (fp, format, (double)vf[i]);
		       if (nc < 0)
			{ return (nc);
			}
		       else
			{ ncTotal += nc;
			}
		       break;
		     }
		    case DOUBLE_TYPE:
		     { nc = fprintf (fp, format, vd[i]);
		       if (nc < 0)
			{ return (nc);
			}
		       else
			{ ncTotal += nc;
			}
		       break;
		     }
		  }
	       }
	      cp[1] = c1;		/* restore format string */
	      cp[2] = c2;

	      format = cp+1;
	      nc = fprintf (fp, format);
	      if (nc < 0)
	       { return (nc);
	       }
	      else
	       { ncTotal += nc;
	       }
	    }
	 }
	return (ncTotal);
}
	
int printVd (s, v, size)
const char *s;
double *v;
int size;
{
	return (fprintVd (stdout, s, v, size));
}

int printVf (s, v, size)
const char *s;
float *v;
int size;
{
	return (fprintVf (stdout, s, v, size));
}

int printVi (s, v, size)
const char *s;
int *v;
int size;
{
	return (fprintVi (stdout, s, v, size));
}

int printJf (fmt, v, n)
const char *fmt;
float *v;
int n;
{
#define MAXN	12
	float deg[MAXN];

	n = MIN(n, MAXN);
	scaleVf (deg, RADTODEG, v, n);
	return (printVf (fmt, deg, n));
#undef MAXN
}

int fprintJf (fp, fmt, v, n)
FILE *fp;
const char *fmt;
float *v;
int n;
{
#define MAXN	12
	float deg[MAXN];

	n = MIN(n, MAXN);
	scaleVf (deg, RADTODEG, v, n);
	return (fprintVf (fp, fmt, deg, n));
#undef MAXN
}
