#include <sys/types.h>
#include <sys/file.h>
#include <sys/errno.h>
#include <sys/open.h>
#include <sys/cred.h>
#include <sys/uio.h>
#include <sys/conf.h>
#include <sys/ddi.h>
#include <sys/poll.h>
#include <sys/kmem.h>
#include <sys/cmn_err.h>
#include <sys/edt.h>
#include <sys/vmereg.h>
#include <sys/pio.h>
#include <sys/eisa.h>

#include "rciioctl.h"
#include "eppreg.h"

int eppdevflag = 0;
extern int epp_internal_major;
extern EppSoftc eppi[];

#define IRQ_MASK	0x0018

void eppintr(unit)
int unit;
{
	/* Called in response to an interrupt from the EPP */

	EppSoftc *info;
	int pending;
	int s;

	info = &eppi[unit];
#if 0
	XXX fix up interrupt code XXX
	while (pending = info->epp->intpend)
	 { 	   
	   info->epp->intclear = pending;

	   if (info->flags & EPP_BUSY)	/* is device open? */
	    { if (pending & EPP_INTRMASK)
	       { 
		 s = splhi();
		 if (info->intrCnt < info->qsize)
		  { info->intrCnt++;		/* bump the counter */
		    pollwakeup (info->pollhead, POLLIN|POLLRDNORM);
		  }
		 splx(s);
		 pending &= ~(EPP_INTRMASK);
	       }
	    }
	   if (pending)
	    { /* declare a stray interrupt */
	      cmn_err (CE_CONT, "epp%d: stray interrupt(s); mask=0x%x\n",
				 unit, pending);
	    }
	 }
#endif
}

eppedtinit(e)
edt_t *e;
{
	/* Initialization routine. Verify that the board is there, and
	   initialize its registers to something reasonable. */

	int i, vec;
	volatile Epp *epp;
	piomap_t *pmap;

	pmap = pio_mapalloc(e->e_bus_type, e->e_adap, &e->e_space[0],
			    PIOMAP_FIXED, "epp");
	if (pmap == 0)
	 { return;
	 }
	epp = (Epp*)pio_mapaddr (pmap, e->e_space[0].ios_iopaddr);

#if 0
	if (! goodid(epp))
	 { pio_mapfree (pmap);
	   return;
	 }
#endif
	eppi[e->e_ctlr].epp = 0;

	vec = eisa_ivec_alloc (e->e_adap, IRQ_MASK, EISA_EDGE_IRQ);
	if (vec == -1)
	 { cmn_err (CE_WARN, "epp: unit %d, no irq vector\n", e->e_ctlr);
	   pio_mapfree (pmap);
	   return;
	 }

	eisa_ivec_set (e->e_adap, vec, eppintr, e->e_ctlr);

#if 0
	XXX init device registers here XXX

	epp->csr = EPP_RESET;
	epp->csr = 0;
	epp->csr = EPP_RED_OFF | EPP_GREEN_OFF;
	epp->dirs = EPP_DIRMASK;
	epp->vector = vec;
#endif
	eppi[e->e_ctlr].epp = epp;
	cmn_err (CE_CONT, "eppedtinit: unit %d vector %x ipl %d\n",
		e->e_ctlr, vec, ((vme_intrs_t*)(e->e_bus_info))->v_brl);
}

eppopen(dev,flag,otyp,crp)
dev_t *dev;
int flag;
int otyp;
cred_t *crp;
{
	/* This device is an exclusive open device: only one process
	   may open it at a time. This prevents contention problems
	   and makes the driver software easier to implement. */

	EppSoftc *info;
	volatile Epp *epp;
	int oldflags;
	int s;

	info = &eppi[geteminor(*dev)];

	oldflags = info->flags;

	if (! info->epp)
	 { return (ENXIO);
	 }
	epp = info->epp;
	s = splhi();
	if (info->flags & EPP_BUSY)
	 { splx(s);
	   return (EBUSY);
	 }
	else
	 { info->flags |= EPP_BUSY;
	 }
	splx(s);

	info->qsize = TQ_DEFAULT_SIZE;
	info->intrCnt = 0;
	info->pollhead = phalloc (KM_SLEEP);
#if 0
	XXX init device registers XXX
	epp->dirs = EPP_DIRMASK;
	epp->intclear = EPP_INTRCLEAR;
#endif
	return (0);
}

static doclose(info)
EppSoftc *info;
{
	volatile Epp *epp;
	int s;

	epp = info->epp;

#if 0
	XXX clear device registers XXX
	epp->csr = EPP_RED_OFF | EPP_GREEN_OFF;
	epp->intmask = 0;
	epp->intclear = 0xff;
#endif
	if (info->pollhead)
	 { phfree (info->pollhead);
	 }
	else
	 { cmn_err (CE_WARN, "eppclose: pollhead is NULL\n");
	 }
	info->flags = 0;
}

eppclose(dev,flag,otyp,crp)
dev_t dev;
int flag;
int otyp;
cred_t *crp;
{
	/* When closing the device, reset the registers to what
	   they were when the device was openned. */

	EppSoftc *info;

	info = &eppi[geteminor(dev)];

	/* check for soft-close */
	if (! (info->flags & EPP_BUSY) )
	 { /* don't return an error, since this may happen normally */
/*DBG*/cmn_err (CE_CONT, "hard close returned, nothing to do\n");
	   return (0);
	 }
	doclose (info);
/*DBG*/cmn_err (CE_CONT, "hard close called\n");
	return (0);
}

eppread(dev,uiop,crp)
dev_t dev;
uio_t *uiop;
cred_t *crp;
{
	/* The read routine returns a single byte (set to 1) if an interrupt
	   arrived. */

	EppSoftc *info;
	int cnt, s;
	int ack = 1;

	info = &eppi[geteminor(dev)];

	/* check for soft-close */
	if (! (info->flags & EPP_BUSY) )
	 { return (EBADF);
	 }

	s = splhi();
	cnt = info->intrCnt;
	if (cnt)
	 { info->intrCnt--;
	 }
	splx(s);
	if (cnt)
	 { if (uread (ack, uiop) != 0)
	    { return (EFAULT);
	    }
	 }
	return (0);
}

eppwrite(dev,uiop,crp)
dev_t dev;
uio_t *uiop;
cred_t *crp;
{
	return (ENXIO);
}

eppchpoll(dev,events,anyyet,reventsp,phpp)
dev_t dev;
short events;
int anyyet;
short *reventsp;
struct pollhead **phpp;
{
	/* Check to see if any interrupt events have arrived. If not,
	   then queue the process for a possible select wakeup. */

	EppSoftc *info;
	int cnt, s;

	info = &eppi[geteminor(dev)];

	/* check for soft-close */
	if (! (info->flags & EPP_BUSY) )
	 { return (EBADF);
	 }

	/* What about races? */
	if ((events & (POLLIN|POLLRDNORM)) && info->intrCnt > 0)
	 { *reventsp = events & (POLLIN|POLLRDNORM);
	 }
	else
	 { *reventsp = 0;
	   if (!anyyet)
	    { *phpp = info->pollhead;
	    }
	 }
	return (0);
}

eppioctl(dev,cmd,arg,mode,crp,rvalp)
dev_t dev;
int cmd;
void *arg;
int mode;
cred_t *crp;
int *rvalp;
{
	EppSoftc *info;

	info = &eppi[geteminor(dev)];

	/* check for soft-close */
	if (! (info->flags & EPP_BUSY) )
	 { return (EBADF);
	 }

	switch (cmd)
	 { default:
	    { return (ENOTTY);
	      break;
	    }
	   case RCIIOC_CONFIRM:
	    { break;
	    }
	   case RCIIOC_INTR_ENABLE:
	    { int mask = (int)arg;
	      volatile Epp *epp = info->epp;
	      int s;

	      s = splhi(); /* is this really necessary? */
	      if (mask)
	       { 
#if 0
		 XXX enable interrupts XXX
		 epp->intmask |= EPP_INTRMASK;
		 epp->intclear = EPP_INTRCLEAR;
		 if (! (epp->csr & EPP_INTENB))
		  { epp->csr |= EPP_INTENB;
		  }
#endif
	       }
	      else
	       { 
#if 0
		 XXX clear interrupts XXX
		 epp->csr &= ~EPP_INTENB;
		 epp->intmask = 0;
		 epp->intclear = 0xff;
#endif
	       }
	      splx(s);
	      break;
	    }
	   case RCIIOC_SET_INTRQ:
	    { int size, s;

	      size = (int)arg;
	      if (size <= 0)
	       { return (EINVAL);
	       }
	      else
	       { s = splhi();
		 info->qsize = size;
		 info->intrCnt = 0;
		 splx(s);
	       }
	      break;
	    }
	   case RCIIOC_GET_INTRQ:
	    { int size;

	      size = info->qsize;
	      if (copyout ((caddr_t)&size, (caddr_t)arg, sizeof(int)))
	       { return (EFAULT);
		 break;
	       }
	      break;
	    }
/* RCIIOC_SOFT_CLOSE is a special hack just for the SGI machines, since
   it seems there is a bug: calling close(2) does not always result
   in the driver close routine being called right away. So we fake
   it with an ioctl, just to be sure. What a botch! */
	   case RCIIOC_SOFT_CLOSE:
	    { doclose (info);      
/*DBG*/cmn_err (CE_CONT, "soft close called\n");
	      break;
	    }
	 }
	return (0);
}
