#ifndef _EPPREG_H_
#define _EPPREG_H_

#ifdef KERNEL
#include "sys/rci.h"
#include "sys/poll.h"
#else
#include <sys/poll.h>
#include "rci.h"
#endif

/* control masks and codes */

/* Cfg1 register. Controls group 1, consisting of ports A and B */

#define EPP_DMAEN1	0x8000	/* enable DMA for group 1 */
#define EPP_INTEN1	0x4000	/* enable interrupts on DRDY1 */
#define EPP_T1SMASK	0x3800	/* transmission delay control bits */
#define EPP_DIOBEN	0x0400	/* Port B handshaking enable */
#define EPP_DIOAEN	0x0200	/* Port A handshaking enable */
#define EPP_LRESET1	0x0100	/* Local reset for group 1 */
#define EPP_INVREQ1	0x0040	/* interpret REQ1 as active low */
#define EPP_DBLBUFA	0x0020	/* double buffer port A */
#define EPP_PULSE1	0x0010	/* use pulse handshaking for group 1 */
#define EPP_EDGE1	0x0008	/* if pulse handshaking, use trailing edge */
#define EPP_INVACK1	0x0004	/* configure ACK1 to be active low */
#define EPP_SETACK1	0x0002	/* control ACK1 on the I/O connector */
#define EPP_OUT1	0x0001	/* control OUT1 on the I/O connector */

/* Cfg2 register. Controls group 2, consisting of ports C and D */

#define EPP_DMAEN2	0x8000	/* enable DMA for group 2 */
#define EPP_INTEN2	0x4000	/* enable interrupts on DRDY2 */
#define EPP_T2SMASK	0x3800	/* transmission delay control bits */
#define EPP_DIODEN	0x0400	/* Port D handshaking enable */
#define EPP_DIOCEN	0x0200	/* Port C handshaking enable */
#define EPP_LRESET2	0x0100	/* Local reset for group 2 */
#define EPP_INVREQ2	0x0040	/* interpret REQ2 as active low */
#define EPP_DBLBUFC	0x0020	/* double buffer port C */
#define EPP_PULSE2	0x0010	/* use pulse handshaking for group 2 */
#define EPP_EDGE2	0x0008	/* if pulse handshaking, use trailing edge */
#define EPP_INVACK2	0x0004	/* configure ACK2 to be active low */
#define EPP_SETACK2	0x0002	/* control ACK2 on the I/O connector */
#define EPP_OUT2	0x0001	/* control OUT2 on the I/O connector */

/* Cfg3 register */

#define EPP_DBLBUFB	0x8000	/* double buffer port B */
#define EPP_WRITED	0x4000	/* configure port D for writing */
#define EPP_WRITEB	0x2000	/* configure port B for writing */
#define EPP_TRANS32	0x1000	/* enable 32-bit transfer mode */
#define EPP_WRITEC	0x0800	/* configure port C for writing */
#define EPP_WRITEA	0x0400	/* configure port A for writing */
/* Other Cfg3 bits control the counters and so are omitted here */

/* Stat register */
#define EPP_DMACH	0x8000	/* Current DMA channel bit */
#define EPP_CNTINT	0x4000	/* counter interrupt */
#define EPP_TRANS32	0x1000	/* 32-bit transfers enabled */
#define EPP_IN1		0x0800	/* extra input line 1 */
#define EPP_DMATC2	0x0400	/* end of group 2 DMA interrupt */
#define EPP_DMATC1	0x0200	/* end of group 1 DMA interrupt */
#define EPP_DRDY1	0x0100	/* group 1 DRDY1 interrupt */
#define EPP_REQ1	0x0080	/* REQ1 as seen at the I/O connector */
#define EPP_ACK1	0x0040	/* group 1 ACK1 value */
#define EPP_IN2		0x0020	/* extra input line 2 */
#define EPP_REVC_ID	0x0010	/* Set for revision B board */
#define EPP_DRDY2	0x0004	/* group 2 DRDY2 interrupt */
#define EPP_REQ2	0x0002	/* REQ2 as seen at the I/O connector */
#define EPP_ACK2	0x0001	/* group 2 ACK2 value */

typedef struct {
	volatile unsigned short cfg1stat;	/* 00 */
	volatile unsigned short cfg2;		/* 02 */
	volatile unsigned short cfg3;		/* 04 */
	volatile unsigned short portAB;		/* 06 */
	volatile unsigned short portCD;		/* 08 */
	volatile unsigned short cntintclr;	/* 0a */
	volatile unsigned short dmaclr1;	/* 0c */
	volatile unsigned short dmaclr2;	/* 0e */
	volatile unsigned char rtsishft;	/* 10 */
	volatile char unused2[1];		
	volatile unsigned char rtsistrb;	/* 12 */
	volatile char unused3[1];
	volatile unsigned short cfg4;		/* 14 */
	volatile char unused4[2];
	volatile unsigned char cntr1;		/* 18 */
	volatile char unused5[1];
	volatile unsigned char cntr2;		/* 1a */
	volatile char unused6[1];
	volatile unsigned char cntr3;		/* 1c */
	volatile char unused7[1];
	volatile unsigned char cntrcmd;		/* 1e */
	volatile char unused8[1];
  } Epp;

typedef struct {
	volatile int intrCnt;
	volatile int qsize;
	/*volatile*/ struct pollhead *pollhead;
#define EPP_BUSY	0x01
	volatile int flags;
	volatile Epp *epp;
} EppSoftc;

#endif
