#include <sys/types.h>
#include <sys/stat.h>
#include <sys/time.h>
#include <fcntl.h>
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <errno.h>

#include <atiioctl.h>
#include <atireg.h>
#include <tval.h>
#include <assert.h>
#include <inout.h>

char *devname = "/dev/ati0";

void nap(sec)
double sec;
{
	struct timeval tv;
	struct timeval now;
	struct timeval then;
	fd_set fds;
	int n;
	extern int errno;
	
	gettimeofday(&now,0);
	tv.tv_sec = (int) sec;
	tv.tv_usec = (int) (1000000 * (sec - tv.tv_sec));
	then.tv_sec = now.tv_sec + tv.tv_sec;
	then.tv_usec = now.tv_usec + tv.tv_usec;
	while (then.tv_usec >= 1000000)
	 { then.tv_usec -= 1000000;
	   then.tv_sec ++;
	 }
	while (1)
	 { FD_ZERO (&fds);
	   errno = 0;
	   n = select(32,&fds,0,0,&tv);
	   gettimeofday(&now,0);
	   tv.tv_sec = then.tv_sec - now.tv_sec;
	   tv.tv_usec = then.tv_usec - now.tv_usec;
	   while (tv.tv_usec < 0)
	    { tv.tv_usec += 1000000;
	      tv.tv_sec --;
	    }
	   if (tv.tv_sec < 0)
	    { return;
	    }
	 }
}

#define NUMSAMPS	1000

typedef struct {
	  int iobase;
	  void *regs;
	  int fhead;
	  int ftail;
	  AtiForceRec *fbuf;
	  int fsize;
	  AtiCardDataRec cdata;
	  } AtiSoftc;

#define ATI_SPIN_TIMEOUT 10000
#define copyout(a,b,n) (memcpy (b,a,n), 0)
#define copyin(a,b,n) (memcpy (b,a,n), 0)

static int atiioctl (int fd, int cmd, void *arg);
static void atiReadData (AtiForceRec *rec, AtiSoftc *info);
static int atiReadCom (int com, int mode, void *ptr, int n,
			AtiSoftc *info);
static int atiWriteCom (int com, int mode, void *ptr, int n,
			AtiSoftc *info);
static int atiHardwareReset (AtiSoftc *info);
static int atiHardwareClose(AtiSoftc *info);
static int atiHardwareOpen (AtiSoftc *info);
static int atiHardwareIntrEnable (AtiSoftc *info, int enable);

AtiSoftc atiinfo;
AtiSoftc *info = &atiinfo;

main()
{
	Tval tv0, tv1, tvd;
	AtiForceRec fbuf[1000];
	int fd, i, k, n, ntotal;
	AtiSensorRec srec;
	AtiForceRec frec;
	AtiBiasRec brec0;
	AtiBiasRec brec;
	AtiTestRec trec;
	AtiCalMatrixRec crec;
	AtiCalMatrixRec crec0;
	AtiPeakValueRec prec;
	int fsize;
	int oflags0;
	int oflags;
	int crate;

	if ((fd = open (devname, O_RDWR)) == -1)
	 { perror ("open");
	   exit (1);
	 }
	if (ioctl (fd, ATIIOC_ENABLE_IOPORTS, 1) == -1 ||
	    ioctl (fd, ATIIOC_ENABLE_IOPORTS, 0) == -1)
	 { perror ("ioctl ATIIOC_ENABLE_IOPORTS");
	   exit (1);
	 }
	info->iobase = 0x280;
	if (atiReadCom (ATI_CARD_DATA_CMD, ATI_READ, &info->cdata,
			sizeof(info->cdata)/2, info) < 0)
	 { fprintf (stderr, "atiReadCom failed\n");
	   exit (1);
	 }

	if (ioctl (fd, ATIIOC_GET_SENSOR_INFO, &srec) < 0)
	 { perror ("ioctl  ATIIOC_GET_SENSOR_INFO");
	   exit (1);
	 }
	printf ("srec.cntsPerForce = %d\n", srec.cntsPerForce);
	printf ("srec.cntsPerTorque = %d\n", srec.cntsPerTorque);
	printf ("srec.units = %d\n", srec.units);

	if (ioctl (fd, ATIIOC_GET_BIAS, &brec0) < 0)
	 { perror ("ioctl ATIIOC_GET_BIAS");
	   exit (1);
	 }
	printf ("bias: ");
	for (i=0; i<8; i++)
	 { printf ("%d \n", brec0.bias[i]);
	   brec.bias[i] = i;
	 }
	printf ("\n");

	if (ioctl (fd, ATIIOC_SET_BIAS, &brec) < 0)
	 { perror ("ioctl ATIIOC_SET_BIAS");
	   exit (1);
	 }
	if (ioctl (fd, ATIIOC_GET_BIAS, &brec) < 0)
	 { perror ("ioctl ATIIOC_GET_BIAS");
	   exit (1);
	 }
	printf ("btest:\n");
	for (i=0; i<8; i++)
	 { printf ("%d ", brec.bias[i]);
	 }
	printf ("\n");
	for (i=0; i<8; i++)
	 { assert (i==brec.bias[i]);
	 }
	if (ioctl (fd, ATIIOC_SET_BIAS, &brec0) < 0)
	 { perror ("ioctl ATIIOC_SET_BIAS");
	   exit (1);
	 }

	if (ioctl (fd, ATIIOC_TEST_RESET, &trec) < 0)
	 { perror ("ioctl ATIIOC_TEST_RESET");
	   exit (1);
	 }
	printf ("trec.ecode = 0x%x\n", trec.ecode);
	printf ("trec.ramCheck = 0x%x\n", trec.ramCheck);
	printf ("trec.gageCheck = 0x%x\n", trec.gageCheck);
	if (ioctl (fd, ATIIOC_SET_CAPBUF_SIZE, 1234) < 0)
	 { perror ("ioctl ATIIOC_SET_CAPBUF_SIZE");
	   exit (1);
	 }
	if (ioctl (fd, ATIIOC_GET_CAPBUF_SIZE, &fsize) < 0)
	 { perror ("ioctl ATIIOC_GET_CAPBUF_SIZE");
	   exit (1);
	 }
	assert (fsize == 1234);
	if (ioctl (fd, ATIIOC_GET_FORCE_DATA, &frec) < 0)
	 { perror ("ioctl ATIIOC_GET_FORCE_DATA");
	   exit (1);
	 }
	printf ("force: ");
	for (i=0; i<6; i++)
	 { printf ("%d ", frec.fdata[i]);
	 }
	printf ("\n");
	if (ioctl (fd, ATIIOC_GET_CALMATRIX, &crec0) < 0)
	 { perror ("ioctl ATIIOC_GET_CALMATRIX");  
	   exit (1);
	 }
	for (i=0; i<48; i++)
	 { crec.calmat[i] = crec0.calmat[i] + i;
	 }
	if (ioctl (fd, ATIIOC_SET_CALMATRIX, &crec) < 0)
	 { perror ("ioctl ATIIOC_SET_CALMATRIX");  
	   exit (1);
	 }
	if (ioctl (fd, ATIIOC_GET_CALMATRIX, &crec) < 0)
	 { perror ("ioctl ATIIOC_GET_CALMATRIX");  
	   exit (1);
	 }
	for (i=0; i<48; i++)
	 { assert (crec.calmat[i] == crec0.calmat[i] + i);
	 }
	if (ioctl (fd, ATIIOC_TEST_RESET, &trec) < 0)
	 { perror ("ioctl ATIIOC_TEST_RESET");
	   exit (1);
	 }
	if (ioctl (fd, ATIIOC_GET_CALMATRIX, &crec) < 0)
	 { perror ("ioctl ATIIOC_GET_CALMATRIX");  
	   exit (1);
	 }
	for (i=0; i<48; i++)
	 { assert (crec.calmat[i] == crec0.calmat[i]);
	 }
	if (ioctl (fd, ATIIOC_GET_PEAK_DATA, &prec) < 0)
	 { perror ("ioctl ATIIOC_GET_PEAK_DATA");
	   exit (1);
	 }
	printf ("min peak: ");
	for (i=0; i<6; i++)
	 { printf ("%d ", prec.min[i]);
	 }
	printf ("\n");
	printf ("max peak: ");
	for (i=0; i<6; i++)
	 { printf ("%d ", prec.max[i]);
	 }
	printf ("\n");
	if (ioctl (fd, ATIIOC_GET_OPFLAGS, &oflags0) < 0)
	 { perror ("ioctl ATIIOC_GET_OPFLAGS");
	   exit (1);
	 }
	printf ("oflags=0x%x\n", oflags0);
	printf ("oflags = 0x%x\n", oflags);
	if (ioctl (fd, ATIIOC_SET_OPFLAGS, 0) < 0)
	 { perror ("ioctl ATIIOC_SET_OPFLAGS");
	   exit (1);
	 }
	if (ioctl (fd, ATIIOC_GET_OPFLAGS, &oflags) < 0)
	 { perror ("ioctl ATIIOC_GET_OPFLAGS");
	   exit (1);
	 }
	assert (oflags == 0);
	if (ioctl (fd, ATIIOC_SET_OPFLAGS, oflags0) < 0)
	 { perror ("ioctl ATIIOC_SET_OPFLAGS");
	   exit (1);
	 }
	if (ioctl (fd, ATIIOC_GET_CAPTURE_RATE, &crate) < 0)
	 { perror ("ioctl ATIIOC_GET_CAPTURE_RATE");
	   exit (1);
	 }
	printf ("capture rate: %d\n", crate);

	if (ioctl (fd, ATIIOC_SET_CAPTURE, 200) < 0)
	 { perror ("ioctl ATIIOC_SET_CAPTURE");
	   exit (1);
	 }
#if 0
	gettimeofday (&tv0, (struct timezone*)0);
	for (ntotal=0; ntotal<NUMSAMPS; ntotal += n/sizeof(AtiForceRec))
	 { 
	   if ((n = read (fd, &fbuf, sizeof(fbuf))) < 0)
	    { perror ("read");
	      exit (1);
	    }
	   nap (0.01);
	 }
	gettimeofday (&tv1, (struct timezone*)0);
	tvd = subTval (tv1, tv0);
	printf ("%d samples in %g seconds\n", NUMSAMPS, tvalToDouble(tvd));
	printf ("%g samps/sec\n", NUMSAMPS/tvalToDouble(tvd));
	for (i=0; i<n/sizeof(AtiForceRec); i++)
	 { for (k=0; k<6; k++)
	    { double cntsper;

	      cntsper = k<3 ? 250 : 10000;
	      printf ("%8.3f ", 10*fbuf[i].fdata[k]/cntsper);
	    }
	   printf ("\n");
	 }
#endif
	for (ntotal=0; ntotal<50; ntotal += n/sizeof(AtiForceRec))
	 { 
	   if ((n = read (fd, &fbuf[ntotal], sizeof(fbuf))) < 0)
	    { perror ("read");
	      exit (1);
	    }
	   nap (0.01);
	 }
	for (i=0; i<ntotal; i++)
	 { for (k=0; k<6; k++)
	    { double cntsper;

	      cntsper = k<3 ? 250 : 10000;
	      printf ("%8.3f ", 10*fbuf[i].fdata[k]/cntsper);
	    }
	   printf ("\n");
	 }
	return (0);
}

static int atiioctl (int fd, int cmd, void *arg)
{
	info->iobase = 0x280;

	switch (cmd)
	 { default:
	    { errno = ENOTTY;
	      return -1;
	      break;
	    }
	   case ATIIOC_GET_SENSOR_INFO:
	    { AtiSensorRec rec;

	      if (atiReadCom (ATI_SENSOR_DATA_CMD, ATI_READ, &rec,
			      sizeof(rec)/2, info) < 0)
	       { errno = ETIME;
		 return -1;
	       }
	      if (copyout((caddr_t)&rec, (caddr_t)arg, sizeof(AtiSensorRec)))
	       { errno = EFAULT;
		 return -1;
	       }
	      break;
	    }
	   case ATIIOC_GET_BIAS:
	    { AtiBiasRec rec;
	      
	      if (atiReadCom (ATI_BIAS_ARRAY_CMD, ATI_READ, &rec,
			      sizeof(rec)/2, info) < 0)
	       { errno = ETIME; return -1;
	       }
	      if (copyout((caddr_t)&rec, (caddr_t)arg, sizeof(AtiBiasRec)))
	       { errno = EFAULT; return -1;
	       }
	      break;
	    }
	   case ATIIOC_SET_BIAS:
	    { AtiBiasRec rec;
	      
	      if (copyin((caddr_t)arg, (caddr_t)&rec, sizeof(AtiBiasRec)))
	       { errno = EFAULT; return -1;
	       }
printf ("set %d %d %d\n", rec.bias[0], rec.bias[1], rec.bias[2]);
	      if (atiWriteCom (ATI_BIAS_ARRAY_CMD, 2, &rec,
			       sizeof(rec)/2, info) < 0)
	       { errno = ETIME; return -1;
	       }
	      break;
	    }
	   case ATIIOC_TEST_RESET:
	    { AtiTestRec rec;
	      
	      if (atiReadCom (ATI_RESET_CMD, ATI_NO_MODE, &rec,
			      sizeof(rec)/2, info) < 0)
	       { errno = ETIME; return -1;
	       }
	      if (copyout((caddr_t)&rec, (caddr_t)arg, sizeof(AtiTestRec)))
	       { errno = EFAULT; return -1;
	       }
	      break;
	    }
	   case ATIIOC_SET_CAPBUF_SIZE:
	    { if (arg <= 0)
	       { errno = EINVAL; return -1;
	       }
	      if (info->fsize != arg+1)
	       { if (info->fsize)
		  { free (info->fbuf);
		    info->fsize = 0;
		  }
		 info->fbuf = (AtiForceRec*)malloc (
			       ((int)arg+1)*sizeof(AtiForceRec));
		 if (info->fbuf == NULL)
		  { errno = ENOMEM; return -1;
		  }
		 info->fsize = (arg+1);
	       }
	      break;
	    }
	   case ATIIOC_GET_CAPBUF_SIZE:
	    { int size = info->fsize-1;
	      if (copyout((caddr_t)&size, (caddr_t)arg, sizeof(int)))
	       { errno = EFAULT; return -1;
	       }
	      break;
	    }
	   case ATIIOC_GET_FORCE_DATA:
	    { AtiForceRec rec;

	      atiReadData (&rec, info);
	      if (copyout((caddr_t)&rec, (caddr_t)arg, sizeof(AtiForceRec)))
	       { errno = EFAULT; return -1;
	       }
	      break;
	    }
	   case ATIIOC_SET_CALMATRIX:
	    { AtiCalMatrixRec rec;

	      if (copyin((caddr_t)arg, (caddr_t)&rec, sizeof(AtiCalMatrixRec)))
	       { errno = EFAULT; return -1;
	       }
	      if (atiWriteCom (ATI_CALMATRIX_CMD, ATI_WRITE, &rec,
			       sizeof(rec)/2, info) < 0)
	       { errno = ETIME; return -1;
	       }
	      break;
	    }
	   case ATIIOC_GET_CALMATRIX:
	    { AtiCalMatrixRec rec;

	      if (atiReadCom (ATI_CALMATRIX_CMD, ATI_READ, &rec,
			      sizeof(rec)/2, info) < 0)
	       { errno = ETIME; return -1;
	       }
	      if (copyout((caddr_t)&rec, (caddr_t)arg,sizeof(AtiCalMatrixRec)))
	       { errno = EFAULT; return -1;
	       }
	      break;
	    }
	   case ATIIOC_GET_PEAK_DATA:
	    { AtiPeakValueRec rec;

	      if (atiReadCom (ATI_GET_PEAK_CMD, ATI_READ, &rec,
			      sizeof(rec)/2, info) < 0)
	       { errno = ETIME; return -1;
	       }
	      if (copyout((caddr_t)&rec, (caddr_t)arg,sizeof(AtiPeakValueRec)))
	       { errno = EFAULT; return -1;
	       }
	      break;
	    }
	   case ATIIOC_GET_PEAK_RESET:
	    { AtiPeakValueRec rec;

	      /* ATI_WRITE here means to reset the data after returning it */
	      if (atiReadCom (ATI_GET_PEAK_CMD, ATI_WRITE, &rec,
			      sizeof(rec)/2, info) < 0)
	       { errno = ETIME; return -1;
	       }
	      if (copyout((caddr_t)&rec, (caddr_t)arg,sizeof(AtiPeakValueRec)))
	       { errno = EFAULT; return -1;
	       }
	      break;
	    }
	   case ATIIOC_GET_OPFLAGS:
	    { int oflags;

	      oflags = info->cdata.oflags;
	      if (copyout((caddr_t)&oflags, (caddr_t)arg,sizeof(int)))
	       { errno = EFAULT; return -1;
	       }
	      break;
	    }
	   case ATIIOC_SET_OPFLAGS:
	    { unsigned short oflags = (unsigned short)arg;
	      if (oflags != info->cdata.oflags)
	       { info->cdata.oflags = oflags;
		 if (atiWriteCom (ATI_CARD_DATA_CMD, ATI_WRITE, &info->cdata,
				  sizeof(info->cdata)/2, info) < 0)
		  { errno = ETIME; return -1;
		  }
	       }
	      break;
	    }
	   case ATIIOC_GET_CAPTURE_RATE:
	    { int rate;

	      rate = info->cdata.sampleRate;
	      if (copyout((caddr_t)&rate, (caddr_t)arg,sizeof(int)))
	       { errno = EFAULT; return -1;
	       }
	      break;
	    }
	   case ATIIOC_SET_CAPTURE:
	    {
	      if (arg > 4000)
	       { errno = EINVAL; return -1;
	       }
	      if (arg <= 0)
	       { arg = 0;
	       }
	      if (arg != 0 && arg != info->cdata.sampleRate)
	       { info->cdata.sampleRate = arg;
		 if (atiWriteCom (ATI_CARD_DATA_CMD,ATI_WRITE, &info->cdata,
				  sizeof(info->cdata)/2, info) < 0)
		  { errno = ETIME; return -1;
		  }
	       }
	      if (info->fsize == 0)
	       { info->fbuf = (AtiForceRec*)malloc (
		       (ATI_DEFAULT_FBUF_SIZE+1)*sizeof(AtiForceRec));
		 if (info->fbuf == NULL)
		  { errno = ENOMEM; return -1;
		  }
		 info->fsize = (ATI_DEFAULT_FBUF_SIZE+1);
	       }
	      if (info->cdata.intrEnable == 0)
	       { info->fhead = 0;
		 info->ftail = 0;
	       }	      
	      if (atiHardwareIntrEnable (info, arg != 0) < 0)
	       { errno = ETIME; return -1;
	       }
	      break;
	    }
	 }
	return (0);
}


#if MAPPED_IO
#define OUTB(p,v)	regs->p = (v)
#define OUTW(p,v)	regs->p = (v)
#define INB(p)		regs->p
#define INW(p)		regs->p
#else
#define OUTB(p,v)	outb(info->iobase+ATI_##p,v)
#define OUTW(p,v)	outw(info->iobase+ATI_##p,v)
#define INB(p)		inb(info->iobase+ATI_##p)
#define INW(p)		inw(info->iobase+ATI_##p)
#endif

static void atiReadData (AtiForceRec *rec, AtiSoftc *info)
{
#if MAPPED_IO
	AtiRegs *regs = info->regs;
#endif
	int i;

	OUTW (addr, 0x101);
	rec->dword = INW (data);
	if (rec->dword)
	 { OUTW (addr, 0x101);
	   OUTW (data, 0);	/* clear monitor condition */
	 }
	for (i=0; i<6; i++)
	 { rec->fdata[i] = INW (data);
	 }
}

static int atiReadCom (int com, int mode, void *ptr, int n,
			AtiSoftc *info)
{
#if MAPPED_IO
	AtiRegs *regs = info->regs;
#endif
	unsigned short *dbuf = (unsigned short *)ptr;
	int retcode;

	if (atiWriteCom (com, mode, NULL, 0, info) < 0)
	 { return (-1);
	 }
	OUTW(addr,0);	  /* reset address */
	retcode = INW(data);
	if (mode != ATI_NO_MODE)
	 { INW(data);
	 }
	while (n--)
	 { *dbuf++ = INW(data);
	 }
	return (retcode);
}

static int atiWriteCom (int com, int mode, void *ptr, int n,
			AtiSoftc *info)
{
#if MAPPED_IO
	AtiRegs *regs = info->regs;
#endif
	unsigned short *dbuf = (unsigned short *)ptr;
	unsigned dword, retcode;
	int spin;

	/*
	   write command block to DPRAM at offset 0x000 
	 */
	OUTW(addr,0); 		/* set address register to command_block */
	OUTW(data,com);
	if (mode != ATI_NO_MODE)
	 { OUTW(data,mode);
	 }
	if (dbuf)
	 { while (n--)
	    { OUTW(data,*dbuf++);
	    }
	 }
	/*
	   read MailBox at DPRAM 0x3FE to clear host interrupt flag
	 */
	OUTW(addr,0x3FE);	     	/* set mailbox address */
	INW(data);		   	/* reset interrupt flag */
	OUTW(addr,0x3FE) ;	    	/* set mailbox address */
	OUTW(data,com);		    	/* clear acknowledgment */

	/*
	   write command word to Mailbox at DPRAM 0x3FF to set DSP intr flag
	 */
	OUTW(addr,0x3ff) ;
	OUTW(data,com) ;

	/*
	   waits for command completion (interrupt flag) from DSP
	 */
	retcode = 1;
	for (spin = ATI_SPIN_TIMEOUT; spin; spin--)
	 {
	   dword = INW(status);
	   // wait for interrupt flag
	   if((dword & 0x0010)==0)     		/* interrupt set */
	    {
	      /* read mailbox location at 0x3FE to reset interrupt flag */
	      OUTW(addr,0x3FE) ;	 	/* write dpram address */
	      retcode = INW(data) ;	 	/* reset interrupt flag */

	      /* test mailbox for command acknowledgment */
	      if(retcode & 0x00ff)		/* test for ack */
	       { break;
	       }
	    }
	 }
	if (spin)
	 { return retcode;
	 }
	else
	 { return -1;
	 }
}

static int atiHardwareReset (AtiSoftc *info)
{
	/* make the device quiescent. Called at attach and detach time */

	/* get card info */
	if (atiReadCom (ATI_CARD_DATA_CMD, ATI_READ, &info->cdata,
			sizeof(info->cdata)/2, info) < 0)
	 { return (-1);
	 }
	return (atiHardwareIntrEnable (info, 0));
}

static int atiHardwareClose(AtiSoftc *info)
{
	/* make the device quiet when it is closed */

	/* disable interrupts */
	return (atiHardwareIntrEnable (info, 0));
}

static int atiHardwareOpen (AtiSoftc *info)
{
	/* initialize the device when it is openned */	

	if (atiReadCom (ATI_CARD_DATA_CMD, ATI_READ, &info->cdata,
			sizeof(info->cdata)/2, info) < 0)
	 { return (-1);
	 }
	return (0);
}

static int atiHardwareIntrEnable (AtiSoftc *info, int enable)
{
	/* enable/disable device interrupts */

	if (enable != info->cdata.intrEnable)
	 { info->cdata.intrEnable = enable;
	   if (atiWriteCom (ATI_CARD_DATA_CMD, ATI_WRITE, &info->cdata,
			    sizeof(info->cdata)/2, info) < 0)
	    { return (-1);
	    }
	 }
	return (1);
}
